@@ -1,8 +1,8 @@
 /* ldapmodrdn.c - generic program to modify an entry's RDN using LDAP */
-/* $OpenLDAP: pkg/ldap/clients/tools/ldapmodrdn.c,v 1.116.2.2 2007/02/13 20:22:02 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2007 The OpenLDAP Foundation.
+ * Copyright 1998-2020 The OpenLDAP Foundation.
  * Portions Copyright 1998-2003 Kurt D. Zeilenga.
  * Portions Copyright 1998-2001 Net Boolean Incorporated.
  * Portions Copyright 2001-2003 IBM Corporation.
@@ -38,8 +38,8 @@
  * This work was originally developed by the University of Michigan
  * (as part of U-MICH LDAP).  Additional significant contributors
  * include:
- *    Kurt D. Zeilenga
- *    Juan C Gomez
+ *	Kurt D. Zeilenga
+ *	Juan C Gomez
  */
 
 
@@ -68,11 +68,11 @@ static int   remove_old_RDN = 0;
 
 
 static int domodrdn(
-    LDAP	*ld,
-    char	*dn,
-    char	*rdn,
-    char	*newSuperior,
-    int		remove );	/* flag: remove old RDN */
+	LDAP	*ld,
+	char	*dn,
+	char	*rdn,
+	char	*newSuperior,
+	int		remove );	/* flag: remove old RDN */
 
 void
 usage( void )
@@ -83,7 +83,11 @@ usage( void )
 	fprintf( stderr, _("		If not given, the list of modifications is read from stdin or\n"));
 	fprintf( stderr, _("		from the file specified by \"-f file\" (see man page).\n"));
 	fprintf( stderr, _("Rename options:\n"));
-	fprintf( stderr, _("  -r         remove old RDN\n"));
+ 	fprintf( stderr, _("  -c         continuous operation mode (do not stop on errors)\n"));
+ 	fprintf( stderr, _("  -f file    read operations from `file'\n"));
+ 	fprintf( stderr, _("  -M         enable Manage DSA IT control (-MM to make critical)\n"));
+ 	fprintf( stderr, _("  -P version protocol version (default: 3)\n"));
+	fprintf( stderr, _("  -r		 remove old RDN\n"));
 	fprintf( stderr, _("  -s newsup  new superior entry\n"));
 	tool_common_usage();
 	exit( EXIT_FAILURE );
@@ -91,7 +95,7 @@ usage( void )
 
 
 const char options[] = "rs:"
-	"cd:D:e:f:h:H:IMnO:o:p:P:QR:U:vVw:WxX:y:Y:Z";
+	"cd:D:e:f:h:H:IMnNO:o:p:P:QR:U:vVw:WxX:y:Y:Z";
 
 int
 handle_private_option( int i )
@@ -118,7 +122,7 @@ handle_private_option( int i )
 			optarg++;
 		}
 
-		control = strdup( optarg );
+		control = optarg;
 		if ( (cvalue = strchr( control, '=' )) != NULL ) {
 			*cvalue++ = '\0';
 		}
@@ -127,8 +131,8 @@ handle_private_option( int i )
 #endif
 
 	case 'r':	/* remove old RDN */
-	    remove_old_RDN++;
-	    break;
+		remove_old_RDN++;
+		break;
 
 	case 's':	/* newSuperior */
 		if( protocol == LDAP_VERSION2 ) {
@@ -136,9 +140,9 @@ handle_private_option( int i )
 				prog, protocol );
 			exit( EXIT_FAILURE );
 		}
-	    newSuperior = strdup( optarg );
-	    protocol = LDAP_VERSION3;
-	    break;
+		newSuperior = optarg;
+		protocol = LDAP_VERSION3;
+		break;
 
 	default:
 		return 0;
@@ -150,100 +154,99 @@ handle_private_option( int i )
 int
 main(int argc, char **argv)
 {
-    char		*entrydn = NULL, *rdn = NULL, buf[ 4096 ];
-    FILE		*fp;
-    LDAP		*ld;
+	char		*entrydn = NULL, *rdn = NULL, buf[ 4096 ];
+	FILE		*fp = NULL;
+	LDAP		*ld = NULL;
 	int		rc, retval, havedn;
 
-    tool_init( TOOL_MODRDN );
-    prog = lutil_progname( "ldapmodrdn", argc, argv );
+	tool_init( TOOL_MODRDN );
+	prog = lutil_progname( "ldapmodrdn", argc, argv );
 
 	tool_args( argc, argv );
 
-    havedn = 0;
-    if (argc - optind == 2) {
-	if (( rdn = strdup( argv[argc - 1] )) == NULL ) {
-	    perror( "strdup" );
-	    return( EXIT_FAILURE );
+	havedn = 0;
+	if (argc - optind == 2) {
+		if (( rdn = strdup( argv[argc - 1] )) == NULL ) {
+			perror( "strdup" );
+			retval = EXIT_FAILURE;
+			goto fail;
+		}
+		if (( entrydn = strdup( argv[argc - 2] )) == NULL ) {
+			perror( "strdup" );
+			retval = EXIT_FAILURE;
+			goto fail;
+		}
+		++havedn;
+	} else if ( argc - optind != 0 ) {
+		fprintf( stderr, _("%s: invalid number of arguments (%d), only two allowed\n"), prog, argc-optind );
+		usage();
 	}
-        if (( entrydn = strdup( argv[argc - 2] )) == NULL ) {
-	    perror( "strdup" );
-	    return( EXIT_FAILURE );
-        }
-	++havedn;
-    } else if ( argc - optind != 0 ) {
-	fprintf( stderr, _("%s: invalid number of arguments (%d), only two allowed\n"), prog, argc-optind );
-	usage();
-    }
 
-    if ( infile != NULL ) {
-	if (( fp = fopen( infile, "r" )) == NULL ) {
-	    perror( infile );
-	    return( EXIT_FAILURE );
+	if ( infile != NULL ) {
+		if (( fp = fopen( infile, "r" )) == NULL ) {
+			perror( infile );
+			retval = EXIT_FAILURE;
+			goto fail;
+		}
+	} else {
+		fp = stdin;
 	}
-    } else {
-	fp = stdin;
-    }
 
 	ld = tool_conn_setup( 0, 0 );
 
-	if ( pw_file || want_bindpw ) {
-		if ( pw_file ) {
-			rc = lutil_get_filed_password( pw_file, &passwd );
-			if( rc ) return EXIT_FAILURE;
-		} else {
-			passwd.bv_val = getpassphrase( _("Enter LDAP Password: ") );
-			passwd.bv_len = passwd.bv_val ? strlen( passwd.bv_val ) : 0;
-		}
-	}
-
 	tool_bind( ld );
 
 	tool_server_controls( ld, NULL, 0 );
 
-    retval = rc = 0;
-    if (havedn)
-	retval = domodrdn( ld, entrydn, rdn, newSuperior, remove_old_RDN );
-    else while ((rc == 0 || contoper) && fgets(buf, sizeof(buf), fp) != NULL) {
-	if ( *buf != '\n' ) {	/* blank lines optional, skip */
-	    buf[ strlen( buf ) - 1 ] = '\0';	/* remove nl */
+	retval = rc = 0;
+	if (havedn)
+		retval = domodrdn( ld, entrydn, rdn, newSuperior, remove_old_RDN );
+	else while ((rc == 0 || contoper) && fgets(buf, sizeof(buf), fp) != NULL) {
+		if ( *buf != '\n' ) {	/* blank lines optional, skip */
+			buf[ strlen( buf ) - 1 ] = '\0';	/* remove nl */
 
-	    if ( havedn ) {	/* have DN, get RDN */
-		if (( rdn = strdup( buf )) == NULL ) {
-                    perror( "strdup" );
-                    return( EXIT_FAILURE );
+			if ( havedn ) {	/* have DN, get RDN */
+				if (( rdn = strdup( buf )) == NULL ) {
+					perror( "strdup" );
+					retval = EXIT_FAILURE;
+					goto fail;
+				}
+				rc = domodrdn(ld, entrydn, rdn, newSuperior, remove_old_RDN );
+				if ( rc != 0 )
+					retval = rc;
+				havedn = 0;
+				free( rdn ); rdn = NULL;
+				free( entrydn ); entrydn = NULL;
+			} else if ( !havedn ) {	/* don't have DN yet */
+				if (( entrydn = strdup( buf )) == NULL ) {
+					retval = EXIT_FAILURE;
+					goto fail;
+				}
+				++havedn;
+			}
 		}
-		rc = domodrdn(ld, entrydn, rdn, newSuperior, remove_old_RDN );
-		if ( rc != 0 )
-			retval = rc;
-		havedn = 0;
-	    } else if ( !havedn ) {	/* don't have DN yet */
-	        if (( entrydn = strdup( buf )) == NULL ) {
-		    perror( "strdup" );
-		    return( EXIT_FAILURE );
-	        }
-		++havedn;
-	    }
 	}
-    }
 
-	tool_unbind( ld );
-	tool_destroy();
-    return( retval );
+fail:
+	if ( fp && fp != stdin ) fclose( fp );
+	if ( entrydn ) free( entrydn );
+	if ( rdn ) free( rdn );
+	tool_exit( ld, retval );
 }
 
 static int domodrdn(
-    LDAP	*ld,
-    char	*dn,
-    char	*rdn,
-    char	*newSuperior,
-    int		remove ) /* flag: remove old RDN */
+	LDAP	*ld,
+	char	*dn,
+	char	*rdn,
+	char	*newSuperior,
+	int		remove ) /* flag: remove old RDN */
 {
 	int rc, code, id;
 	char *matcheddn=NULL, *text=NULL, **refs=NULL;
+	LDAPControl **ctrls = NULL;
 	LDAPMessage *res;
 
-    if ( verbose ) {
+	if ( verbose ) {
 		printf( _("Renaming \"%s\"\n"), dn );
 		printf( _("\tnew rdn=\"%s\" (%s old rdn)\n"),
 			rdn, remove ? _("delete") : _("keep") );
@@ -284,7 +287,7 @@ static int domodrdn(
 		}
 	}
 
-	rc = ldap_parse_result( ld, res, &code, &matcheddn, &text, &refs, NULL, 1 );
+	rc = ldap_parse_result( ld, res, &code, &matcheddn, &text, &refs, &ctrls, 1 );
 
 	if( rc != LDAP_SUCCESS ) {
 		fprintf( stderr, "%s: ldap_parse_result: %s (%d)\n",
@@ -314,6 +317,11 @@ static int domodrdn(
 		}
 	}
 
+	if (ctrls) {
+		tool_print_ctrls( ld, ctrls );
+		ldap_controls_free( ctrls );
+	}
+
 	ber_memfree( text );
 	ber_memfree( matcheddn );
 	ber_memvfree( (void **) refs );
