@@ -1,8 +1,8 @@
 /* allop.c - returns all operational attributes when appropriate */
-/* $OpenLDAP: pkg/ldap/contrib/slapd-modules/allop/allop.c,v 1.3.2.1 2007/02/13 20:22:03 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2005-2007 The OpenLDAP Foundation.
+ * Copyright 2005-2020 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -39,6 +39,7 @@ allop-URI	<ldapURI>
 #include <ac/string.h>
 
 #include "slap.h"
+#include "config.h"
 
 #define	SLAP_OVER_VERSION_REQUIRE(major,minor,patch) \
 	( \
@@ -140,7 +141,7 @@ allop_db_config(
 }
 
 static int
-allop_db_destroy( BackendDB *be )
+allop_db_destroy( BackendDB *be, ConfigReply *cr )
 {
 	slap_overinst	*on = (slap_overinst *)be->bd_info;
 	allop_t		*ao = (allop_t *)on->on_bi.bi_private;
@@ -244,6 +245,7 @@ allop_init()
 {
 	allop.on_bi.bi_type = "allop";
 
+	allop.on_bi.bi_flags = SLAPO_BFLAG_SINGLE;
 	allop.on_bi.bi_db_config = allop_db_config;
 	allop.on_bi.bi_db_destroy = allop_db_destroy;
 
