@@ -1,8 +1,8 @@
 /* lastmod.c - returns last modification info */
-/* $OpenLDAP: pkg/ldap/contrib/slapd-modules/lastmod/lastmod.c,v 1.2.2.1 2007/02/13 20:22:03 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2004-2007 The OpenLDAP Foundation.
+ * Copyright 2004-2020 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -367,7 +367,7 @@ best_guess( Operation *op,
 		struct berval *bv_modifiersName, struct berval *bv_nmodifiersName )
 {
 	if ( bv_entryCSN ) {
-		char		csnbuf[ LDAP_LUTIL_CSNSTR_BUFSIZE ];
+		char		csnbuf[ LDAP_PVT_CSNSTR_BUFSIZE ];
 		struct berval	entryCSN;
 	
 		entryCSN.bv_val = csnbuf;
@@ -736,9 +736,7 @@ lastmod_response( Operation *op, SlapReply *rs )
 }
 
 static int
-lastmod_db_init(
-	BackendDB *be
-)
+lastmod_db_init( BackendDB *be, ConfigReply *cr )
 {
 	slap_overinst		*on = (slap_overinst *)be->bd_info;
 	lastmod_info_t		*lmi;
@@ -757,7 +755,7 @@ lastmod_db_init(
 			code = register_at( mat[i].schema, ad, 0 );
 			if ( code ) {
 				Debug( LDAP_DEBUG_ANY,
-					"lastmod_init: register_at failed\n", 0, 0, 0 );
+					"lastmod_init: register_at failed\n" );
 				return -1;
 			}
 			(*ad)->ad_type->sat_flags |= mat[i].flags;
@@ -771,7 +769,7 @@ lastmod_db_init(
 			code = register_oc( moc[i].schema, Oc, 0 );
 			if ( code ) {
 				Debug( LDAP_DEBUG_ANY,
-					"lastmod_init: register_oc failed\n", 0, 0, 0 );
+					"lastmod_init: register_oc failed\n" );
 				return -1;
 			}
 			(*Oc)->soc_flags |= moc[i].flags;
@@ -827,16 +825,14 @@ lastmod_db_config(
 }
 
 static int
-lastmod_db_open(
-	BackendDB *be
-)
+lastmod_db_open( BackendDB *be, ConfigReply *cr )
 {
 	slap_overinst	*on = (slap_overinst *) be->bd_info;
 	lastmod_info_t	*lmi = (lastmod_info_t *)on->on_bi.bi_private;
 	char		buf[ 8192 ];
 	static char		tmbuf[ LDAP_LUTIL_GENTIME_BUFSIZE ];
 
-	char			csnbuf[ LDAP_LUTIL_CSNSTR_BUFSIZE ];
+	char			csnbuf[ LDAP_PVT_CSNSTR_BUFSIZE ];
 	struct berval		entryCSN;
 	struct berval timestamp;
 
@@ -904,9 +900,7 @@ lastmod_db_open(
 }
 
 static int
-lastmod_db_destroy(
-	BackendDB *be
-)
+lastmod_db_destroy( BackendDB *be, ConfigReply *cr )
 {
 	slap_overinst	*on = (slap_overinst *)be->bd_info;
 	lastmod_info_t	*lmi = (lastmod_info_t *)on->on_bi.bi_private;
@@ -939,6 +933,7 @@ int
 lastmod_initialize()
 {
 	lastmod.on_bi.bi_type = "lastmod";
+	lastmod.on_bi.bi_flags = SLAPO_BFLAG_SINGLE;
 	lastmod.on_bi.bi_db_init = lastmod_db_init;
 	lastmod.on_bi.bi_db_config = lastmod_db_config;
 	lastmod.on_bi.bi_db_destroy = lastmod_db_destroy;
