@@ -1,8 +1,8 @@
 /* decode.c - ber input decoding routines */
-/* $OpenLDAP: pkg/ldap/libraries/liblber/decode.c,v 1.105.2.1 2007/02/13 20:22:10 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2007 The OpenLDAP Foundation.
+ * Copyright 1998-2020 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -33,133 +33,222 @@
 #include <stdio.h>
 
 #include <ac/stdlib.h>
-
 #include <ac/stdarg.h>
 #include <ac/string.h>
 #include <ac/socket.h>
 
 #include "lber-int.h"
 
-static ber_len_t ber_getnint LDAP_P((
-	BerElement *ber,
-	ber_int_t *num,
-	ber_len_t len ));
 
-/* return the tag - LBER_DEFAULT returned means trouble */
-ber_tag_t
-ber_get_tag( BerElement *ber )
+/* out->bv_len should be the buffer size on input */
+int
+ber_decode_oid( BerValue *in, BerValue *out )
+{
+	const unsigned char *der;
+	unsigned long val;
+	unsigned val1;
+	ber_len_t i;
+	char *ptr;
+
+	assert( in != NULL );
+	assert( out != NULL );
+
+	/* need 4 chars/inbyte + \0 for input={7f 7f 7f...} */
+	if ( !out->bv_val || (out->bv_len+3)/4 <= in->bv_len )
+		return -1;
+
+	ptr = NULL;
+	der = (unsigned char *) in->bv_val;
+	val = 0;
+	for ( i=0; i < in->bv_len; i++ ) {
+		val |= der[i] & 0x7f;
+		if ( !( der[i] & 0x80 )) {
+			if ( ptr == NULL ) {
+				/* Initial "x.y": val=x*40+y, x<=2, y<40 if x<2 */
+				ptr = out->bv_val;
+				val1 = (val < 80 ? val/40 : 2);
+				val -= val1*40;
+				ptr += sprintf( ptr, "%u", val1 );
+			}
+			ptr += sprintf( ptr, ".%lu", val );
+			val = 0;
+		} else if ( val - 1UL < LBER_OID_COMPONENT_MAX >> 7 ) {
+			val <<= 7;
+		} else {
+			/* val would overflow, or is 0 from invalid initial 0x80 octet */
+			return -1;
+		}
+	}
+	if ( ptr == NULL || val != 0 )
+		return -1;
+
+	out->bv_len = ptr - out->bv_val;
+	return 0;
+}
+
+/* Return tag, with *bv = rest of element (starting at length octets) */
+static ber_tag_t
+ber_tag_and_rest( const BerElement *ber, struct berval *bv )
 {
-	unsigned char	xbyte;
 	ber_tag_t	tag;
-	unsigned int	i;
+	ptrdiff_t	rest;
+	unsigned char	*ptr;
 
 	assert( ber != NULL );
 	assert( LBER_VALID( ber ) );
 
-	if ( ber_pvt_ber_remaining( ber ) < 1 ) {
-		return LBER_DEFAULT;
+	ptr = (unsigned char *) ber->ber_ptr;
+	rest = (unsigned char *) ber->ber_end - ptr;
+	if ( rest <= 0 ) {
+		goto fail;
 	}
 
-	if ( ber->ber_ptr == ber->ber_buf ) {
-		tag = *(unsigned char *)ber->ber_ptr;
-	} else {
-		tag = ber->ber_tag;
+	tag = ber->ber_tag;
+	if ( (char *) ptr == ber->ber_buf ) {
+		tag = *ptr;
 	}
-	ber->ber_ptr++;
-
+	ptr++;
+	rest--;
 	if ( (tag & LBER_BIG_TAG_MASK) != LBER_BIG_TAG_MASK ) {
-		return tag;
+		goto done;
 	}
 
-	for ( i = 1; i < sizeof(ber_tag_t); i++ ) {
-		if ( ber_read( ber, (char *) &xbyte, 1 ) != 1 ) {
-			return LBER_DEFAULT;
-		}
-
-		tag <<= 8;
-		tag |= 0x00ffUL & (ber_tag_t) xbyte;
-
-		if ( ! (xbyte & LBER_MORE_TAG_MASK) ) {
+	do {
+		if ( rest <= 0 ) {
 			break;
 		}
-	}
+		tag <<= 8;
+		tag |= *ptr++ & 0xffU;
+		rest--;
 
-	/* tag too big! */
-	if ( i == sizeof(ber_tag_t) ) {
-		return LBER_DEFAULT;
-	}
+		if ( ! (tag & LBER_MORE_TAG_MASK) ) {
+			goto done;
+		}
+	} while ( tag <= (ber_tag_t)-1 / 256 );
 
+ fail:
+	/* Error or unsupported tag size */
+	tag = LBER_DEFAULT;
+
+ done:
+	bv->bv_len = rest;
+	bv->bv_val = (char *) ptr;
 	return tag;
 }
 
+/* Return the tag - LBER_DEFAULT returned means trouble */
 ber_tag_t
-ber_skip_tag( BerElement *ber, ber_len_t *len )
+ber_get_tag( BerElement *ber )
+{
+	struct berval bv;
+	ber_tag_t tag = ber_tag_and_rest( ber, &bv );
+
+	ber->ber_ptr = bv.bv_val;
+	return tag;
+}
+
+/* Return next element's tag and point *bv at its contents in-place */
+ber_tag_t
+ber_peek_element( const BerElement *ber, struct berval *bv )
 {
 	ber_tag_t	tag;
-	unsigned char	lc;
-	ber_len_t	i, noctets;
-	unsigned char netlen[sizeof(ber_len_t)];
+	ber_len_t	len, rest;
+	unsigned	i;
+	unsigned char *ptr;
 
-	assert( ber != NULL );
-	assert( len != NULL );
-	assert( LBER_VALID( ber ) );
+	assert( bv != NULL );
 
 	/*
 	 * Any ber element looks like this: tag length contents.
-	 * Assuming everything's ok, we return the tag byte (we
-	 * can assume a single byte), and return the length in len.
+	 * Assuming everything's ok, we return the tag, and point
+	 * bv at the contents.
 	 *
 	 * Assumptions:
 	 *	1) definite lengths
 	 *	2) primitive encodings used whenever possible
 	 */
 
-	*len = 0;
+	len = 0;
 
 	/*
 	 * First, we read the tag.
 	 */
+	tag = ber_tag_and_rest( ber, bv );
 
-	if ( (tag = ber_get_tag( ber )) == LBER_DEFAULT ) {
-		return LBER_DEFAULT;
+	rest = bv->bv_len;
+	ptr = (unsigned char *) bv->bv_val;
+	if ( tag == LBER_DEFAULT || rest == 0 ) {
+		goto fail;
 	}
 
 	/*
-	 * Next, read the length.  The first byte contains the length of
-	 * the length.	If bit 8 is set, the length is the long form,
-	 * otherwise it's the short form.  We don't allow a length that's
-	 * greater than what we can hold in a ber_len_t.
+	 * Next, read the length.  The first octet determines the length
+	 * of the length.	If bit 8 is 0, the length is the short form,
+	 * otherwise if the octet != 0x80 it's the long form, otherwise
+	 * the ber element has the unsupported indefinite-length format.
+	 * Lengths that do not fit in a ber_len_t are not accepted.
 	 */
 
-	if ( ber_read( ber, (char *) &lc, 1 ) != 1 ) {
-		return LBER_DEFAULT;
-	}
+	len = *ptr++;
+	rest--;
 
-	if ( lc & 0x80U ) {
-		noctets = (lc & 0x7fU);
-
-		if ( noctets > sizeof(ber_len_t) ) {
-			return LBER_DEFAULT;
+	if ( len & 0x80U ) {
+		len &= 0x7fU;
+		if ( len - 1U > sizeof(ber_len_t) - 1U || rest < len ) {
+			/* Indefinite-length/too long length/not enough data */
+			goto fail;
 		}
 
-		if( (unsigned) ber_read( ber, (char *) netlen, noctets ) != noctets ) {
-			return LBER_DEFAULT;
+		rest -= len;
+		i = len;
+		for( len = *ptr++ & 0xffU; --i; len |= *ptr++ & 0xffU ) {
+			len <<= 8;
 		}
-
-		for( i = 0; i < noctets; i++ ) {
-			*len <<= 8;
-			*len |= netlen[i];
-		}
-
-	} else {
-		*len = lc;
 	}
 
 	/* BER element should have enough data left */
-	if( *len > (ber_len_t) ber_pvt_ber_remaining( ber ) ) {
-		return LBER_DEFAULT;
+	if( len > rest ) {
+	fail:
+		tag = LBER_DEFAULT;
+	}
+
+	bv->bv_len = len;
+	bv->bv_val = (char *) ptr;
+	return tag;
+}
+
+/* Move past next element, point *bv at it in-place, and return its tag.
+ * The caller may \0-terminate *bv, as next octet is saved in ber->ber_tag.
+ * Similar to ber_get_stringbv(ber, bv, LBER_BV_NOTERM) except on error.
+ */
+ber_tag_t
+ber_skip_element( BerElement *ber, struct berval *bv )
+{
+	ber_tag_t tag = ber_peek_element( ber, bv );
+
+	if ( tag != LBER_DEFAULT ) {
+		ber->ber_ptr = bv->bv_val + bv->bv_len;
+		ber->ber_tag = *(unsigned char *) ber->ber_ptr;
+	}
+
+	return tag;
+}
+
+/* Move past next element, point *bv at the complete element in-place, and
+ * return its tag. The caller may \0-terminate *bv, as next octet is saved in
+ * ber->ber_tag. Similar to ber_skip_element(ber, bv) except the tag+length
+ * header is also included in *bv.
+ */
+ber_tag_t
+ber_skip_raw( BerElement *ber, struct berval *bv )
+{
+	char		*val = ber->ber_ptr;
+	ber_tag_t	tag = ber_skip_element( ber, bv );
+
+	if ( tag != LBER_DEFAULT ) {
+		bv->bv_len += bv->bv_val - val;
+		bv->bv_val = val;
 	}
-	ber->ber_tag = *(unsigned char *)ber->ber_ptr;
 
 	return tag;
 }
@@ -169,58 +258,61 @@ ber_peek_tag(
 	BerElement *ber,
 	ber_len_t *len )
 {
-	/*
-	 * This implementation assumes ber_skip_tag() only
-	 * modifies ber_ptr field of the BerElement.
-	 */
-
-	char *save;
-	ber_tag_t	tag, old;
-
-	old = ber->ber_tag;
-	save = ber->ber_ptr;
-	tag = ber_skip_tag( ber, len );
-	ber->ber_ptr = save;
-	ber->ber_tag = old;
+	struct berval bv;
+	ber_tag_t tag = ber_peek_element( ber, &bv );
 
+	*len = bv.bv_len;
 	return tag;
 }
 
-static ber_len_t
-ber_getnint(
-	BerElement *ber,
-	ber_int_t *num,
-	ber_len_t len )
+ber_tag_t
+ber_skip_tag( BerElement *ber, ber_len_t *lenp )
 {
-	unsigned char buf[sizeof(ber_int_t)];
+	struct berval bv;
+	ber_tag_t tag = ber_peek_element( ber, &bv );
+
+	ber->ber_ptr = bv.bv_val;
+	ber->ber_tag = *(unsigned char *) ber->ber_ptr;
+
+	*lenp = bv.bv_len;
+	return tag;
+}
+
+ber_tag_t
+ber_get_int(
+	BerElement *ber,
+	ber_int_t *num )
+{
+	struct berval bv;
+	ber_tag_t	tag = ber_skip_element( ber, &bv );
+
+	if ( tag == LBER_DEFAULT ) {
+		return tag;
+	}
+
+	return ber_decode_int( &bv, num ) ? LBER_DEFAULT : tag;
+}
+
+int
+ber_decode_int( const struct berval *bv, ber_int_t *num )
+{
+	ber_len_t	len = bv->bv_len;
+	if ( len > sizeof(ber_int_t) )
+		return -1;
 
-	assert( ber != NULL );
 	assert( num != NULL );
-	assert( LBER_VALID( ber ) );
-
-	/*
-	 * The tag and length have already been stripped off.  We should
-	 * be sitting right before len bytes of 2's complement integer,
-	 * ready to be read straight into an int.  We may have to sign
-	 * extend after we read it in.
-	 */
-
-	if ( len > sizeof(ber_int_t) ) {
-		return -1;
-	}
-
-	/* read into the low-order bytes of our buffer */
-	if ( (ber_len_t) ber_read( ber, (char *) buf, len ) != len ) {
-		return -1;
-	}
 
+	/* parse two's complement integer */
 	if( len ) {
-		/* sign extend if necessary */
+		unsigned char *buf = (unsigned char *) bv->bv_val;
 		ber_len_t i;
-		ber_int_t netnum = 0x80 & buf[0] ? -1 : 0;
+		ber_int_t netnum = buf[0] & 0xff;
+
+		/* sign extend */
+		netnum = (netnum ^ 0x80) - 0x80;
 
 		/* shift in the bytes */
-		for( i=0 ; i<len; i++ ) {
+		for( i = 1; i < len; i++ ) {
 			netnum = (netnum << 8 ) | buf[i];
 		}
 
@@ -229,31 +321,8 @@ ber_getnint(
 	} else {
 		*num = 0;
 	}
-	ber->ber_tag = *(unsigned char *)ber->ber_ptr;
 
-	return len;
-}
-
-ber_tag_t
-ber_get_int(
-	BerElement *ber,
-	ber_int_t *num )
-{
-	ber_tag_t	tag;
-	ber_len_t	len;
-
-	assert( ber != NULL );
-	assert( LBER_VALID( ber ) );
-
-	if ( (tag = ber_skip_tag( ber, &len )) == LBER_DEFAULT ) {
-		return LBER_DEFAULT;
-	}
-
-	if ( ber_getnint( ber, num, len ) != len ) {
-		return LBER_DEFAULT;
-	}
-	
-	return tag;
+	return 0;
 }
 
 ber_tag_t
@@ -270,29 +339,22 @@ ber_get_stringb(
 	char *buf,
 	ber_len_t *len )
 {
-	ber_len_t	datalen;
+	struct berval bv;
 	ber_tag_t	tag;
 
-	assert( ber != NULL );
-	assert( LBER_VALID( ber ) );
-
-	if ( (tag = ber_skip_tag( ber, &datalen )) == LBER_DEFAULT ) {
+	if ( (tag = ber_skip_element( ber, &bv )) == LBER_DEFAULT ) {
 		return LBER_DEFAULT;
 	}
 
 	/* must fit within allocated space with termination */
-	if ( datalen >= *len ) {
+	if ( bv.bv_len >= *len ) {
 		return LBER_DEFAULT;
 	}
 
-	if ( (ber_len_t) ber_read( ber, buf, datalen ) != datalen ) {
-		return LBER_DEFAULT;
-	}
-	ber->ber_tag = *(unsigned char *)ber->ber_ptr;
+	memcpy( buf, bv.bv_val, bv.bv_len );
+	buf[bv.bv_len] = '\0';
 
-	buf[datalen] = '\0';
-
-	*len = datalen;
+	*len = bv.bv_len;
 	return tag;
 }
 
@@ -308,222 +370,201 @@ enum bgbvc { ChArray, BvArray, BvVec, BvOff };
  * stack use to the absolute minimum.
  */
 typedef struct bgbvr {
-	enum bgbvc choice;
-	BerElement *ber;
-	int alloc;
-	ber_len_t siz;
-	ber_len_t off;
-	union {
-		char ***c;
-		BerVarray *ba;
-		struct berval ***bv;
-	} res;
+	const enum bgbvc choice;
+	const int option;	/* (ALLOC unless BvOff) | (STRING if ChArray) */
+	ber_len_t siz;		/* input array element size, output count */
+	ber_len_t off;		/* BvOff offset to the struct berval */
+	void *result;
 } bgbvr;
 
 static ber_tag_t
-ber_get_stringbvl( bgbvr *b, ber_len_t *rlen )
+ber_get_stringbvl( BerElement *ber, bgbvr *b )
 {
 	int i = 0, n;
 	ber_tag_t tag;
-	ber_len_t len;
+	ber_len_t tot_size = 0, siz = b->siz;
 	char *last, *orig;
 	struct berval bv, *bvp = NULL;
+	union stringbvl_u {
+		char **ca;				/* ChArray */
+		BerVarray ba;			/* BvArray */
+		struct berval **bv;		/* BvVec */
+		char *bo;				/* BvOff */
+	} res;
 
-	/* For rewinding, just like ber_peek_tag() */
-	orig = b->ber->ber_ptr;
-	tag = b->ber->ber_tag;
+	tag = ber_skip_tag( ber, &bv.bv_len );
 
-	if ( ber_first_element( b->ber, &len, &last ) != LBER_DEFAULT ) {
-		for ( ; b->ber->ber_ptr < last; i++ ) {
-			if (ber_skip_tag( b->ber, &len ) == LBER_DEFAULT) break;
-			b->ber->ber_ptr += len;
-			b->ber->ber_tag = *(unsigned char *)b->ber->ber_ptr;
+	if ( tag != LBER_DEFAULT ) {
+		tag = 0;
+		orig = ber->ber_ptr;
+		last = orig + bv.bv_len;
+
+		for ( ; ber->ber_ptr < last; i++, tot_size += siz ) {
+			if ( ber_skip_element( ber, &bv ) == LBER_DEFAULT )
+				break;
 		}
+		if ( ber->ber_ptr != last ) {
+			i = 0;
+			tag = LBER_DEFAULT;
+		}
+
+		ber->ber_ptr = orig;
+		ber->ber_tag = *(unsigned char *) orig;
 	}
 
-	if ( rlen ) *rlen = i;
-
+	b->siz = i;
 	if ( i == 0 ) {
-		*b->res.c = NULL;
-		return 0;
+		return tag;
 	}
 
-	n = i;
-
-	/* Allocate the result vector */
+	/* Allocate and NULL-terminate the result vector */
+	b->result = ber_memalloc_x( tot_size + siz, ber->ber_memctx );
+	if ( b->result == NULL ) {
+		return LBER_DEFAULT;
+	}
 	switch (b->choice) {
 	case ChArray:
-		*b->res.c = ber_memalloc_x( (n+1)*sizeof( char * ),
-			b->ber->ber_memctx);
-		if ( *b->res.c == NULL ) return LBER_DEFAULT;
-		(*b->res.c)[n] = NULL;
+		res.ca = b->result;
+		res.ca[i] = NULL;
 		break;
 	case BvArray:
-		*b->res.ba = ber_memalloc_x( (n+1)*sizeof( struct berval ),
-			b->ber->ber_memctx);
-		if ( *b->res.ba == NULL ) return LBER_DEFAULT;
-		(*b->res.ba)[n].bv_val = NULL;
+		res.ba = b->result;
+		res.ba[i].bv_val = NULL;
 		break;
 	case BvVec:
-		*b->res.bv = ber_memalloc_x( (n+1)*sizeof( struct berval *),
-			b->ber->ber_memctx);
-		if ( *b->res.bv == NULL ) return LBER_DEFAULT;
-		(*b->res.bv)[n] = NULL;
+		res.bv = b->result;
+		res.bv[i] = NULL;
 		break;
 	case BvOff:
-		*b->res.ba = ber_memalloc_x( (n+1) * b->siz, b->ber->ber_memctx );
-		if ( *b->res.ba == NULL ) return LBER_DEFAULT;
-		((struct berval *)((long)(*b->res.ba) + n*b->siz +
-			b->off))->bv_val = NULL;
+		res.bo = (char *) b->result + b->off;
+		((struct berval *) (res.bo + tot_size))->bv_val = NULL;
+		tot_size = 0;
 		break;
 	}
-	b->ber->ber_ptr = orig;
-	b->ber->ber_tag = tag;
-	ber_skip_tag( b->ber, &len );
-	
-	for (n=0; n<i; n++)
-	{
-		tag = ber_next_element( b->ber, &len, last );
-		if ( ber_get_stringbv( b->ber, &bv, b->alloc ) == LBER_DEFAULT ) {
-			goto nomem;
+
+	n = 0;
+	do {
+		tag = ber_get_stringbv( ber, &bv, b->option );
+		if ( tag == LBER_DEFAULT ) {
+			goto failed;
 		}
 
 		/* store my result */
 		switch (b->choice) {
 		case ChArray:
-			(*b->res.c)[n] = bv.bv_val;
+			res.ca[n] = bv.bv_val;
 			break;
 		case BvArray:
-			(*b->res.ba)[n] = bv;
+			res.ba[n] = bv;
 			break;
 		case BvVec:
-			bvp = ber_memalloc_x( sizeof( struct berval ), b->ber->ber_memctx);
+			bvp = ber_memalloc_x( sizeof( struct berval ),
+				ber->ber_memctx );
 			if ( !bvp ) {
-				LBER_FREE(bv.bv_val);
-				goto nomem;
+				ber_memfree_x( bv.bv_val, ber->ber_memctx );
+				goto failed;
 			}
-			(*b->res.bv)[n] = bvp;
+			res.bv[n] = bvp;
 			*bvp = bv;
 			break;
 		case BvOff:
-			*(BerVarray)((long)(*b->res.ba)+n*b->siz+b->off) = bv;
+			*(struct berval *)(res.bo + tot_size) = bv;
+			tot_size += siz;
 			break;
 		}
-	}
+	} while (++n < i);
 	return tag;
 
-nomem:
-	if (b->alloc || b->choice == BvVec) {
-		for (--n; n>=0; n--) {
+failed:
+	if (b->choice != BvOff) { /* BvOff does not have LBER_BV_ALLOC set */
+		while (--n >= 0) {
 			switch(b->choice) {
 			case ChArray:
-				LBER_FREE((*b->res.c)[n]);
+				ber_memfree_x(res.ca[n], ber->ber_memctx);
 				break;
 			case BvArray:
-				LBER_FREE((*b->res.ba)[n].bv_val);
+				ber_memfree_x(res.ba[n].bv_val, ber->ber_memctx);
 				break;
 			case BvVec:
-				LBER_FREE((*b->res.bv)[n]->bv_val);
-				LBER_FREE((*b->res.bv)[n]);
+				ber_memfree_x(res.bv[n]->bv_val, ber->ber_memctx);
+				ber_memfree_x(res.bv[n], ber->ber_memctx);
 				break;
 			default:
 				break;
 			}
 		}
 	}
-	LBER_FREE(*b->res.c);
-	*b->res.c = NULL;
+	ber_memfree_x(b->result, ber->ber_memctx);
+	b->result = NULL;
 	return LBER_DEFAULT;
 }
 
 ber_tag_t
-ber_get_stringbv( BerElement *ber, struct berval *bv, int alloc )
+ber_get_stringbv( BerElement *ber, struct berval *bv, int option )
 {
 	ber_tag_t	tag;
+	char		*data;
 
-	assert( ber != NULL );
-	assert( bv != NULL );
-
-	assert( LBER_VALID( ber ) );
-
-	if ( (tag = ber_skip_tag( ber, &bv->bv_len )) == LBER_DEFAULT ) {
+	tag = ber_skip_element( ber, bv );
+	if ( tag == LBER_DEFAULT ||
+		(( option & LBER_BV_STRING ) &&
+		 bv->bv_len && memchr( bv->bv_val, 0, bv->bv_len - 1 )))
+	{
 		bv->bv_val = NULL;
 		return LBER_DEFAULT;
 	}
 
-	if ( (ber_len_t) ber_pvt_ber_remaining( ber ) < bv->bv_len ) {
-		return LBER_DEFAULT;
-	}
-
-	if ( alloc ) {
+	data = bv->bv_val;
+	if ( option & LBER_BV_ALLOC ) {
 		bv->bv_val = (char *) ber_memalloc_x( bv->bv_len + 1,
 			ber->ber_memctx );
 		if ( bv->bv_val == NULL ) {
 			return LBER_DEFAULT;
 		}
 
-		if ( bv->bv_len > 0 && (ber_len_t) ber_read( ber, bv->bv_val,
-			bv->bv_len ) != bv->bv_len )
-		{
-			LBER_FREE( bv->bv_val );
-			bv->bv_val = NULL;
-			return LBER_DEFAULT;
+		if ( bv->bv_len != 0 ) {
+			memcpy( bv->bv_val, data, bv->bv_len );
 		}
-	} else {
-		bv->bv_val = ber->ber_ptr;
-		ber->ber_ptr += bv->bv_len;
+		data = bv->bv_val;
 	}
-	ber->ber_tag = *(unsigned char *)ber->ber_ptr;
-	bv->bv_val[bv->bv_len] = '\0';
+	if ( !( option & LBER_BV_NOTERM ))
+		data[bv->bv_len] = '\0';
 
 	return tag;
 }
 
 ber_tag_t
-ber_get_stringbv_null( BerElement *ber, struct berval *bv, int alloc )
+ber_get_stringbv_null( BerElement *ber, struct berval *bv, int option )
 {
 	ber_tag_t	tag;
+	char		*data;
 
-	assert( ber != NULL );
-	assert( bv != NULL );
-
-	assert( LBER_VALID( ber ) );
-
-	if ( (tag = ber_skip_tag( ber, &bv->bv_len )) == LBER_DEFAULT ) {
+	tag = ber_skip_element( ber, bv );
+	if ( tag == LBER_DEFAULT || bv->bv_len == 0 ) {
 		bv->bv_val = NULL;
-		return LBER_DEFAULT;
-	}
-
-	if ( (ber_len_t) ber_pvt_ber_remaining( ber ) < bv->bv_len ) {
-		return LBER_DEFAULT;
-	}
-
-	if ( bv->bv_len == 0 ) {
-		bv->bv_val = NULL;
-		ber->ber_tag = *(unsigned char *)ber->ber_ptr;
 		return tag;
 	}
 
-	if ( alloc ) {
+	if (( option & LBER_BV_STRING ) &&
+		memchr( bv->bv_val, 0, bv->bv_len - 1 ))
+	{
+		bv->bv_val = NULL;
+		return LBER_DEFAULT;
+	}
+
+	data = bv->bv_val;
+	if ( option & LBER_BV_ALLOC ) {
 		bv->bv_val = (char *) ber_memalloc_x( bv->bv_len + 1,
 			ber->ber_memctx );
 		if ( bv->bv_val == NULL ) {
 			return LBER_DEFAULT;
 		}
 
-		if ( bv->bv_len > 0 && (ber_len_t) ber_read( ber, bv->bv_val,
-			bv->bv_len ) != bv->bv_len )
-		{
-			LBER_FREE( bv->bv_val );
-			bv->bv_val = NULL;
-			return LBER_DEFAULT;
-		}
-	} else {
-		bv->bv_val = ber->ber_ptr;
-		ber->ber_ptr += bv->bv_len;
+		memcpy( bv->bv_val, data, bv->bv_len );
+		data = bv->bv_val;
 	}
-	ber->ber_tag = *(unsigned char *)ber->ber_ptr;
-	bv->bv_val[bv->bv_len] = '\0';
+	if ( !( option & LBER_BV_NOTERM ))
+		data[bv->bv_len] = '\0';
 
 	return tag;
 }
@@ -536,7 +577,7 @@ ber_get_stringa( BerElement *ber, char **buf )
 
 	assert( buf != NULL );
 
-	tag = ber_get_stringbv( ber, &bv, 1 );
+	tag = ber_get_stringbv( ber, &bv, LBER_BV_ALLOC | LBER_BV_STRING );
 	*buf = bv.bv_val;
 
 	return tag;
@@ -550,7 +591,7 @@ ber_get_stringa_null( BerElement *ber, char **buf )
 
 	assert( buf != NULL );
 
-	tag = ber_get_stringbv_null( ber, &bv, 1 );
+	tag = ber_get_stringbv_null( ber, &bv, LBER_BV_ALLOC | LBER_BV_STRING );
 	*buf = bv.bv_val;
 
 	return tag;
@@ -570,9 +611,9 @@ ber_get_stringal( BerElement *ber, struct berval **bv )
 		return LBER_DEFAULT;
 	}
 
-	tag = ber_get_stringbv( ber, *bv, 1 );
+	tag = ber_get_stringbv( ber, *bv, LBER_BV_ALLOC );
 	if ( tag == LBER_DEFAULT ) {
-		LBER_FREE( *bv );
+		ber_memfree_x( *bv, ber->ber_memctx );
 		*bv = NULL;
 	}
 	return tag;
@@ -584,63 +625,50 @@ ber_get_bitstringa(
 	char **buf,
 	ber_len_t *blen )
 {
-	ber_len_t	datalen;
 	ber_tag_t	tag;
+	struct berval	data;
 	unsigned char	unusedbits;
 
-	assert( ber != NULL );
 	assert( buf != NULL );
 	assert( blen != NULL );
 
-	assert( LBER_VALID( ber ) );
-
-	if ( (tag = ber_skip_tag( ber, &datalen )) == LBER_DEFAULT ) {
-		*buf = NULL;
-		return LBER_DEFAULT;
+	if ( (tag = ber_skip_element( ber, &data )) == LBER_DEFAULT ) {
+		goto fail;
 	}
-	--datalen;
 
-	*buf = (char *) ber_memalloc_x( datalen, ber->ber_memctx );
+	if ( --data.bv_len > (ber_len_t)-1 / 8 ) {
+		goto fail;
+	}
+	unusedbits = *(unsigned char *) data.bv_val++;
+	if ( unusedbits > 7 ) {
+		goto fail;
+	}
+
+	if ( memchr( data.bv_val, 0, data.bv_len )) {
+		goto fail;
+	}
+
+	*buf = (char *) ber_memalloc_x( data.bv_len, ber->ber_memctx );
 	if ( *buf == NULL ) {
 		return LBER_DEFAULT;
 	}
+	memcpy( *buf, data.bv_val, data.bv_len );
 
-	if ( ber_read( ber, (char *)&unusedbits, 1 ) != 1 ) {
-		LBER_FREE( buf );
-		*buf = NULL;
-		return LBER_DEFAULT;
-	}
-
-	if ( (ber_len_t) ber_read( ber, *buf, datalen ) != datalen ) {
-		LBER_FREE( buf );
-		*buf = NULL;
-		return LBER_DEFAULT;
-	}
-	ber->ber_tag = *(unsigned char *)ber->ber_ptr;
-
-	*blen = datalen * 8 - unusedbits;
+	*blen = data.bv_len * 8 - unusedbits;
 	return tag;
+
+ fail:
+	*buf = NULL;
+	return LBER_DEFAULT;
 }
 
 ber_tag_t
 ber_get_null( BerElement *ber )
 {
 	ber_len_t	len;
-	ber_tag_t	tag;
+	ber_tag_t	tag = ber_skip_tag( ber, &len );
 
-	assert( ber != NULL );
-	assert( LBER_VALID( ber ) );
-
-	if ( (tag = ber_skip_tag( ber, &len )) == LBER_DEFAULT ) {
-		return LBER_DEFAULT;
-	}
-
-	if ( len != 0 ) {
-		return LBER_DEFAULT;
-	}
-	ber->ber_tag = *(unsigned char *)ber->ber_ptr;
-
-	return( tag );
+	return( len == 0 ? tag : LBER_DEFAULT );
 }
 
 ber_tag_t
@@ -648,18 +676,7 @@ ber_get_boolean(
 	BerElement *ber,
 	ber_int_t *boolval )
 {
-	ber_int_t	longbool;
-	ber_tag_t	rc;
-
-	assert( ber != NULL );
-	assert( boolval != NULL );
-
-	assert( LBER_VALID( ber ) );
-
-	rc = ber_get_int( ber, &longbool );
-	*boolval = longbool;
-
-	return rc;
+	return ber_get_int( ber, boolval );
 }
 
 ber_tag_t
@@ -668,8 +685,6 @@ ber_first_element(
 	ber_len_t *len,
 	char **last )
 {
-	assert( ber != NULL );
-	assert( len != NULL );
 	assert( last != NULL );
 
 	/* skip the sequence header, use the len to mark where to stop */
@@ -677,11 +692,10 @@ ber_first_element(
 		*last = NULL;
 		return LBER_DEFAULT;
 	}
-	ber->ber_tag = *(unsigned char *)ber->ber_ptr;
 
 	*last = ber->ber_ptr + *len;
 
-	if ( *last == ber->ber_ptr ) {
+	if ( *len == 0 ) {
 		return LBER_DEFAULT;
 	}
 
@@ -695,9 +709,7 @@ ber_next_element(
 	LDAP_CONST char *last )
 {
 	assert( ber != NULL );
-	assert( len != NULL );
 	assert( last != NULL );
-
 	assert( LBER_VALID( ber ) );
 
 	if ( ber->ber_ptr >= last ) {
@@ -715,8 +727,8 @@ ber_scanf ( BerElement *ber,
 {
 	va_list		ap;
 	LDAP_CONST char		*fmt_reset;
-	char		*s, **ss;
-	struct berval	**bvp, *bval;
+	char		*s, **ss, ***sss;
+	struct berval	data, *bval, **bvp, ***bvpp;
 	ber_int_t	*i;
 	ber_len_t	*l;
 	ber_tag_t	*t;
@@ -727,14 +739,15 @@ ber_scanf ( BerElement *ber,
 
 	assert( ber != NULL );
 	assert( fmt != NULL );
-
 	assert( LBER_VALID( ber ) );
 
 	fmt_reset = fmt;
 
-	ber_log_printf( LDAP_DEBUG_TRACE, ber->ber_debug,
-		"ber_scanf fmt (%s) ber:\n", fmt );
-	ber_log_dump( LDAP_DEBUG_BER, ber->ber_debug, ber, 1 );
+	if ( ber->ber_debug & (LDAP_DEBUG_TRACE|LDAP_DEBUG_BER)) {
+		ber_log_printf( LDAP_DEBUG_TRACE, ber->ber_debug,
+			"ber_scanf fmt (%s) ber:\n", fmt );
+		ber_log_dump( LDAP_DEBUG_BER, ber->ber_debug, ber, 1 );
+	}
 
 	for ( rc = 0; *fmt && rc != LBER_DEFAULT; fmt++ ) {
 		/* When this is modified, remember to update
@@ -773,7 +786,7 @@ ber_scanf ( BerElement *ber,
 			break;
 
 		case 'e':	/* enumerated */
-		case 'i':	/* int */
+		case 'i':	/* integer */
 			i = va_arg( ap, ber_int_t * );
 			rc = ber_get_int( ber, i );
 			break;
@@ -794,14 +807,14 @@ ber_scanf ( BerElement *ber,
 				 * len ptr on finish. parsed in-place.
 				 */
 		{
-			bgbvr cookie = { BvOff };
-			cookie.ber = ber;
-			cookie.res.ba = va_arg( ap, struct berval ** );
-			cookie.alloc = 0;
+			bgbvr cookie = { BvOff, 0 };
+			bvp = va_arg( ap, struct berval ** );
 			l = va_arg( ap, ber_len_t * );
 			cookie.siz = *l;
 			cookie.off = va_arg( ap, ber_len_t );
-			rc = ber_get_stringbvl( &cookie, l );
+			rc = ber_get_stringbvl( ber, &cookie );
+			*bvp = cookie.result;
+			*l = cookie.siz;
 			break;
 		}
 
@@ -811,7 +824,7 @@ ber_scanf ( BerElement *ber,
 
 		case 'o':	/* octet string in a supplied berval */
 			bval = va_arg( ap, struct berval * );
-			rc = ber_get_stringbv( ber, bval, 1 );
+			rc = ber_get_stringbv( ber, bval, LBER_BV_ALLOC );
 			break;
 
 		case 'O':	/* octet string - allocate & include length */
@@ -837,46 +850,47 @@ ber_scanf ( BerElement *ber,
 
 		case 'v':	/* sequence of strings */
 		{
-			bgbvr cookie = { ChArray };
-			cookie.ber = ber;
-			cookie.res.c = va_arg( ap, char *** );
-			cookie.alloc = 1;
-			rc = ber_get_stringbvl( &cookie, NULL );
+			bgbvr cookie = {
+				ChArray, LBER_BV_ALLOC | LBER_BV_STRING, sizeof( char * )
+			};
+			rc = ber_get_stringbvl( ber, &cookie );
+			*(va_arg( ap, char *** )) = cookie.result;
 			break;
 		}
 
 		case 'V':	/* sequence of strings + lengths */
 		{
-			bgbvr cookie = { BvVec };
-			cookie.ber = ber;
-			cookie.res.bv = va_arg( ap, struct berval *** );
-			cookie.alloc = 1;
-			rc = ber_get_stringbvl( &cookie, NULL );
+			bgbvr cookie = {
+				BvVec, LBER_BV_ALLOC, sizeof( struct berval * )
+			};
+			rc = ber_get_stringbvl( ber, &cookie );
+			*(va_arg( ap, struct berval *** )) = cookie.result;
 			break;
 		}
 
 		case 'W':	/* bvarray */
 		{
-			bgbvr cookie = { BvArray };
-			cookie.ber = ber;
-			cookie.res.ba = va_arg( ap, struct berval ** );
-			cookie.alloc = 1;
-			rc = ber_get_stringbvl( &cookie, NULL );
+			bgbvr cookie = {
+				BvArray, LBER_BV_ALLOC, sizeof( struct berval )
+			};
+			rc = ber_get_stringbvl( ber, &cookie );
+			*(va_arg( ap, struct berval ** )) = cookie.result;
 			break;
 		}
 
 		case 'x':	/* skip the next element - whatever it is */
-			if ( (rc = ber_skip_tag( ber, &len )) == LBER_DEFAULT )
-				break;
-			ber->ber_ptr += len;
-			ber->ber_tag = *(unsigned char *)ber->ber_ptr;
+			rc = ber_skip_element( ber, &data );
 			break;
 
 		case '{':	/* begin sequence */
 		case '[':	/* begin set */
-			if ( *(fmt + 1) != 'v' && *(fmt + 1) != 'V'
-				&& *(fmt + 1) != 'W' && *(fmt + 1) != 'M' )
+			switch ( fmt[1] ) {
+			case 'v': case 'V': case 'W': case 'M':
+				break;
+			default:
 				rc = ber_skip_tag( ber, &len );
+				break;
+			}
 			break;
 
 		case '}':	/* end sequence */
@@ -894,6 +908,7 @@ ber_scanf ( BerElement *ber,
 	}
 
 	va_end( ap );
+
 	if ( rc == LBER_DEFAULT ) {
 		/*
 		 * Error.  Reclaim malloced memory that was given to the caller.
@@ -916,42 +931,48 @@ ber_scanf ( BerElement *ber,
 		case 'a':	/* octet string - allocate storage as needed */
 		case 'A':
 			ss = va_arg( ap, char ** );
-			if ( *ss ) {
-				LBER_FREE( *ss );
-				*ss = NULL;
-			}
+			ber_memfree_x( *ss, ber->ber_memctx );
+			*ss = NULL;
 			break;
 
 		case 'b':	/* boolean */
 		case 'e':	/* enumerated */
-		case 'i':	/* int */
-			(void) va_arg( ap, int * );
+		case 'i':	/* integer */
+			(void) va_arg( ap, ber_int_t * );
 			break;
 
 		case 'l':	/* length of next item */
-			(void) va_arg( ap, ber_len_t * );
+			*(va_arg( ap, ber_len_t * )) = 0;
+			break;
+
+		case 'm':	/* berval in-place */
+			bval = va_arg( ap, struct berval * );
+			BER_BVZERO( bval );
+			break;
+
+		case 'M':	/* BVoff array in-place */
+			bvp = va_arg( ap, struct berval ** );
+			ber_memfree_x( *bvp, ber->ber_memctx );
+			*bvp = NULL;
+			*(va_arg( ap, ber_len_t * )) = 0;
+			(void) va_arg( ap, ber_len_t );
 			break;
 
 		case 'o':	/* octet string in a supplied berval */
 			bval = va_arg( ap, struct berval * );
-			if ( bval->bv_val != NULL ) {
-				LBER_FREE( bval->bv_val );
-				bval->bv_val = NULL;
-			}
-			bval->bv_len = 0;
+			ber_memfree_x( bval->bv_val, ber->ber_memctx );
+			BER_BVZERO( bval );
 			break;
 
 		case 'O':	/* octet string - allocate & include length */
 			bvp = va_arg( ap, struct berval ** );
-			if ( *bvp ) {
-				ber_bvfree( *bvp );
-				*bvp = NULL;
-			}
+			ber_bvfree_x( *bvp, ber->ber_memctx );
+			*bvp = NULL;
 			break;
 
 		case 's':	/* octet string - in a buffer */
 			(void) va_arg( ap, char * );
-			(void) va_arg( ap, ber_len_t * );
+			*(va_arg( ap, ber_len_t * )) = 0;
 			break;
 
 		case 't':	/* tag of next item */
@@ -961,19 +982,30 @@ ber_scanf ( BerElement *ber,
 
 		case 'B':	/* bit string - allocate storage as needed */
 			ss = va_arg( ap, char ** );
-			if ( *ss ) {
-				LBER_FREE( *ss );
-				*ss = NULL;
-			}
+			ber_memfree_x( *ss, ber->ber_memctx );
+			*ss = NULL;
 			*(va_arg( ap, ber_len_t * )) = 0; /* for length, in bits */
 			break;
 
-		case 'm':	/* berval in-place */
-		case 'M':	/* BVoff array in-place */
-		case 'n':	/* null */
 		case 'v':	/* sequence of strings */
+			sss = va_arg( ap, char *** );
+			ber_memvfree_x( (void **) *sss, ber->ber_memctx );
+			*sss = NULL;
+			break;
+
 		case 'V':	/* sequence of strings + lengths */
+			bvpp = va_arg( ap, struct berval *** );
+			ber_bvecfree_x( *bvpp, ber->ber_memctx );
+			*bvpp = NULL;
+			break;
+
 		case 'W':	/* BerVarray */
+			bvp = va_arg( ap, struct berval ** );
+			ber_bvarray_free_x( *bvp, ber->ber_memctx );
+			*bvp = NULL;
+			break;
+
+		case 'n':	/* null */
 		case 'x':	/* skip the next element - whatever it is */
 		case '{':	/* begin sequence */
 		case '[':	/* begin set */
