@@ -1,8 +1,8 @@
 /* distproc.c - implement distributed procedures */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-ldap/distproc.c,v 1.3.2.4 2007/02/13 20:22:24 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2005-2007 The OpenLDAP Foundation.
+ * Copyright 2005-2020 The OpenLDAP Foundation.
  * Portions Copyright 2003 Howard Chu.
  * All rights reserved.
  *
@@ -107,6 +107,7 @@ typedef struct ContinuationReference_t {
 } ContinuationReference_t;
 #define	CR_INIT		{ NULL, BER_BVNULL, LDAP_DP_RT_UNKNOWN, BER_BVNULL, LDAP_DP_SS_UNKNOWN, NULL, BER_BVNULL }
 
+#ifdef unused
 static struct berval	bv2rt[] = {
 	BER_BVC( "superior" ),
 	BER_BVC( "subordinate" ),
@@ -196,6 +197,7 @@ ldap_distproc_str2ss( const char *s )
 	ber_str2bv( s, 0, 0, &bv );
 	return ldap_distproc_bv2ss( &bv );
 }
+#endif /* unused */
 
 /*
  * NOTE: this overlay assumes that the chainingBehavior control
@@ -237,13 +239,7 @@ static int ldap_distproc_db_init_common( BackendDB	*be );
 static int ldap_distproc_db_init_one( BackendDB *be );
 #define	ldap_distproc_db_open_one(be)		(lback)->bi_db_open( (be) )
 #define	ldap_distproc_db_close_one(be)		(0)
-#define	ldap_distproc_db_destroy_one(be)	(lback)->bi_db_destroy( (be) )
-
-static int
-ldap_distproc_parse_ctrl(
-	Operation	*op,
-	SlapReply	*rs,
-	LDAPControl	*ctrl );
+#define	ldap_distproc_db_destroy_one(be, ca)	(lback)->bi_db_destroy( (be), (ca) )
 
 static int
 ldap_distproc_uri_cmp( const void *c1, const void *c2 )
@@ -298,7 +294,7 @@ ldap_distproc_operational( Operation *op, SlapReply *rs )
 	 * database.  This fix is likely to intercept also entries
 	 * generated by back-perl and so. */
 	if ( rs->sr_entry->e_private == NULL ) {
-		return 0;
+		return LDAP_SUCCESS;
 	}
 
 	return SLAP_CB_CONTINUE;
@@ -401,7 +397,7 @@ distproc_ldadd( CfEntryInfo *p, Entry *e, ConfigArgs *ca )
 		Debug( LDAP_DEBUG_ANY, "slapd-distproc: "
 			"first underlying database \"%s\" "
 			"cannot contain attribute \"%s\".\n",
-			e->e_name.bv_val, ad->ad_cname.bv_val, 0 );
+			e->e_name.bv_val, ad->ad_cname.bv_val );
 		rc = LDAP_CONSTRAINT_VIOLATION;
 		goto done;
 
@@ -411,7 +407,7 @@ distproc_ldadd( CfEntryInfo *p, Entry *e, ConfigArgs *ca )
 		Debug( LDAP_DEBUG_ANY, "slapd-distproc: "
 			"subsequent underlying database \"%s\" "
 			"must contain attribute \"%s\".\n",
-			e->e_name.bv_val, ad->ad_cname.bv_val, 0 );
+			e->e_name.bv_val, ad->ad_cname.bv_val );
 		rc = LDAP_CONSTRAINT_VIOLATION;
 		goto done;
 	}
@@ -426,8 +422,9 @@ distproc_ldadd( CfEntryInfo *p, Entry *e, ConfigArgs *ca )
 	if ( rc != 0 ) {
 		Debug( LDAP_DEBUG_ANY, "slapd-distproc: "
 			"unable to init %sunderlying database \"%s\".\n",
-			lc->lc_common_li == NULL ? "common " : "", e->e_name.bv_val, 0 );
-		return LDAP_CONSTRAINT_VIOLATION;
+			lc->lc_common_li == NULL ? "common " : "", e->e_name.bv_val );
+		rc = LDAP_CONSTRAINT_VIOLATION;
+		goto done;
 	}
 
 	li = ca->be->be_private;
@@ -440,14 +437,14 @@ distproc_ldadd( CfEntryInfo *p, Entry *e, ConfigArgs *ca )
 	{
 		Debug( LDAP_DEBUG_ANY, "slapd-distproc: "
 			"database \"%s\" insert failed.\n",
-			e->e_name.bv_val, 0, 0 );
+			e->e_name.bv_val );
 		rc = LDAP_CONSTRAINT_VIOLATION;
 		goto done;
 	}
 
 done:;
 	if ( rc != LDAP_SUCCESS ) {
-		(void)ldap_distproc_db_destroy_one( ca->be );
+		(void)ldap_distproc_db_destroy_one( ca->be, NULL );
 		ch_free( ca->be );
 		ca->be = NULL;
 	}
@@ -472,9 +469,9 @@ ldap_distproc_cfadd_apply( void *datum, void *arg )
 	struct berval			bv;
 
 	/* FIXME: should not hardcode "olcDatabase" here */
-	bv.bv_len = snprintf( lca->ca->msg, sizeof( lca->ca->msg ),
+	bv.bv_len = snprintf( lca->ca->cr_msg, sizeof( lca->ca->cr_msg ),
 		"olcDatabase={%d}%s", lca->count, lback->bi_type );
-	bv.bv_val = lca->ca->msg;
+	bv.bv_val = lca->ca->cr_msg;
 
 	lca->ca->be->be_private = (void *)li;
 	config_build_entry( lca->op, lca->rs, lca->p->e_private, lca->ca,
@@ -567,7 +564,8 @@ distproc_cfgen( ConfigArgs *c )
 
 static int
 ldap_distproc_db_init(
-	BackendDB *be )
+	BackendDB *be,
+	ConfigReply *cr )
 {
 	slap_overinst	*on = (slap_overinst *)be->bd_info;
 	ldap_distproc_t	*lc = NULL;
@@ -627,7 +625,7 @@ ldap_distproc_db_config(
 			if ( rc != 0 ) {
 				Debug( LDAP_DEBUG_ANY, "%s: line %d: "
 					"underlying slapd-ldap initialization failed.\n.",
-					fname, lineno, 0 );
+					fname, lineno );
 				return 1;
 			}
 			lc->lc_cfg_li = be->be_private;
@@ -659,7 +657,7 @@ private_destroy:;
 
 				db.bd_info = lback;
 				db.be_private = (void *)lc->lc_cfg_li;
-				ldap_distproc_db_destroy_one( &db );
+				ldap_distproc_db_destroy_one( &db, NULL );
 				lc->lc_cfg_li = NULL;
 
 			} else {
@@ -669,7 +667,7 @@ private_destroy:;
 				{
 					Debug( LDAP_DEBUG_ANY, "%s: line %d: "
 						"no URI list allowed in slapo-distproc.\n",
-						fname, lineno, 0 );
+						fname, lineno );
 					rc = 1;
 					goto private_destroy;
 				}
@@ -680,7 +678,7 @@ private_destroy:;
 				{
 					Debug( LDAP_DEBUG_ANY, "%s: line %d: "
 						"duplicate URI in slapo-distproc.\n",
-						fname, lineno, 0 );
+						fname, lineno );
 					rc = 1;
 					goto private_destroy;
 				}
@@ -712,7 +710,7 @@ ldap_distproc_db_apply( void *datum, void *arg )
 
 	lca->be->be_private = (void *)li;
 
-	return lca->func( lca->be );
+	return lca->func( lca->be, NULL );
 }
 
 static int
@@ -735,7 +733,7 @@ ldap_distproc_db_func(
 			db.bd_info = lback;
 			db.be_private = lc->lc_common_li;
 
-			rc = func( &db );
+			rc = func( &db, NULL );
 
 			if ( rc != 0 ) {
 				return rc;
@@ -759,21 +757,24 @@ ldap_distproc_db_func(
 
 static int
 ldap_distproc_db_open(
-	BackendDB	*be )
+	BackendDB	*be,
+	ConfigReply	*cr )
 {
 	return ldap_distproc_db_func( be, db_open );
 }
 
 static int
 ldap_distproc_db_close(
-	BackendDB	*be )
+	BackendDB	*be,
+	ConfigReply	*cr )
 {
 	return ldap_distproc_db_func( be, db_close );
 }
 
 static int
 ldap_distproc_db_destroy(
-	BackendDB	*be )
+	BackendDB	*be,
+	ConfigReply	*cr )
 {
 	slap_overinst	*on = (slap_overinst *) be->bd_info;
 	ldap_distproc_t	*lc = (ldap_distproc_t *)on->on_bi.bi_private;
@@ -804,7 +805,7 @@ ldap_distproc_db_init_common(
 
 	be->bd_info = lback;
 	be->be_private = NULL;
-	t = lback->bi_db_init( be );
+	t = lback->bi_db_init( be, NULL );
 	if ( t != 0 ) {
 		return t;
 	}
@@ -835,7 +836,7 @@ ldap_distproc_db_init_one(
 
 	be->bd_info = lback;
 	be->be_private = NULL;
-	t = lback->bi_db_init( be );
+	t = lback->bi_db_init( be, NULL );
 	if ( t != 0 ) {
 		return t;
 	}
@@ -924,8 +925,8 @@ ldap_exop_chained_request(
 		Operation	*op,
 		SlapReply	*rs )
 {
-	Statslog( LDAP_DEBUG_STATS, "%s CHAINED REQUEST\n",
-	    op->o_log_prefix, 0, 0, 0, 0 );
+	Debug( LDAP_DEBUG_STATS, "%s CHAINED REQUEST\n",
+	    op->o_log_prefix );
 
 	rs->sr_err = backend_check_restrictions( op, rs, 
 			(struct berval *)&slap_EXOP_CHAINEDREQUEST );
@@ -954,19 +955,17 @@ distproc_initialize( void )
 	if ( rc != LDAP_SUCCESS ) {
 		Debug( LDAP_DEBUG_ANY, "slapd-distproc: "
 			"unable to register chainedRequest exop: %d.\n",
-			rc, 0, 0 );
+			rc );
 		return rc;
 	}
 
-#ifdef LDAP_DEVEL
 	rc = supported_feature_load( &slap_FEATURE_CANCHAINOPS );
 	if ( rc != LDAP_SUCCESS ) {
 		Debug( LDAP_DEBUG_ANY, "slapd-distproc: "
 			"unable to register canChainOperations supported feature: %d.\n",
-			rc, 0, 0 );
+			rc );
 		return rc;
 	}
-#endif
 
 	rc = register_supported_control( LDAP_CONTROL_X_RETURNCONTREF,
 			SLAP_CTRL_GLOBAL|SLAP_CTRL_ACCESS|SLAP_CTRL_HIDE, NULL,
@@ -974,7 +973,7 @@ distproc_initialize( void )
 	if ( rc != LDAP_SUCCESS ) {
 		Debug( LDAP_DEBUG_ANY, "slapd-distproc: "
 			"unable to register returnContinuationReference control: %d.\n",
-			rc, 0, 0 );
+			rc );
 		return rc;
 	}
 
