@@ -1,8 +1,8 @@
 /* back-monitor.h - ldap monitor back-end header file */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-monitor/back-monitor.h,v 1.52.2.2 2007/02/13 20:22:26 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2001-2007 The OpenLDAP Foundation.
+ * Copyright 2001-2020 The OpenLDAP Foundation.
  * Portions Copyright 2001-2003 Pierangelo Masarati.
  * All rights reserved.
  *
@@ -72,8 +72,11 @@ typedef struct monitor_entry_t {
 /* NOTE: flags with 0xF0000000U mask are reserved for subsystem internals */
 
 	struct monitor_callback_t	*mp_cb;		/* callback sequence */
+	void		*mp_private;
 } monitor_entry_t;
 
+struct entry_limbo_t;			/* in init.c */
+
 typedef struct monitor_info_t {
 
 	/*
@@ -128,6 +131,7 @@ typedef struct monitor_info_t {
 	AttributeDescription	*mi_ad_monitorIsShadow;
 	AttributeDescription	*mi_ad_monitorUpdateRef;
 	AttributeDescription	*mi_ad_monitorRuntimeConfig;
+	AttributeDescription	*mi_ad_monitorSuperiorDN;
 
 	/*
 	 * Generic description attribute
@@ -135,7 +139,7 @@ typedef struct monitor_info_t {
 	AttributeDescription	*mi_ad_readOnly;
 	AttributeDescription	*mi_ad_restrictedOperation;
 
-	void			*mi_entry_limbo;
+	struct entry_limbo_t	*mi_entry_limbo;
 } monitor_info_t;
 
 /*
@@ -265,6 +269,8 @@ typedef struct monitor_subsys_t {
 				struct berval *ndn, Entry *, Entry ** );
 	/* modify entry and subentries */
 	int		( *mss_modify )( Operation *, SlapReply *, Entry * );
+
+	void		*mss_private;
 } monitor_subsys_t;
 
 extern BackendDB *be_monitor;
@@ -285,6 +291,10 @@ typedef struct monitor_extra_t {
 	monitor_subsys_t * (*get_subsys_by_dn)( struct berval *ndn, int sub );
 
 	int (*register_subsys)( monitor_subsys_t *ms );
+	int (*register_backend)( BackendInfo *bi );
+	int (*register_database)( BackendDB *be, struct berval *ndn_out );
+	int (*register_overlay_info)( slap_overinst *on );
+	int (*register_overlay)( BackendDB *be, slap_overinst *on, struct berval *ndn_out );
 	int (*register_entry)( Entry *e, monitor_callback_t *cb,
 		monitor_subsys_t *ms, unsigned long flags );
 	int (*register_entry_parent)( Entry *e, monitor_callback_t *cb,
@@ -297,6 +307,15 @@ typedef struct monitor_extra_t {
 	monitor_cbfunc *unregister_entry_parent;
 	monitor_cbafunc *unregister_entry_attrs;
 	monitor_cbfunc *unregister_entry_callback;
+	Entry * (*entry_stub)( struct berval *pdn,
+		struct berval *pndn,
+		struct berval *rdn,
+		ObjectClass *oc,
+		struct berval *create,
+		struct berval *modify );
+	monitor_entry_t * (*entrypriv_create)( void );
+	int (*register_subsys_late)( monitor_subsys_t *ms );
+	Entry * (*entry_get_unlocked)( struct berval *ndn );
 } monitor_extra_t;
 
 LDAP_END_DECL
