@@ -1,8 +1,8 @@
 /* database.c - deals with database subsystem */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-monitor/database.c,v 1.80.2.4 2007/02/13 20:22:26 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2001-2007 The OpenLDAP Foundation.
+ * Copyright 2001-2020 The OpenLDAP Foundation.
  * Portions Copyright 2001-2003 Pierangelo Masarati.
  * All rights reserved.
  *
@@ -33,14 +33,6 @@
 static int monitor_back_add_plugin( monitor_info_t *mi, Backend *be, Entry *e );
 #endif /* defined(LDAP_SLAPI) */
 
-/* for PATH_MAX on some systems (e.g. Solaris) */
-#ifdef HAVE_LIMITS_H
-#include <limits.h>
-#endif /* HAVE_LIMITS_H */
-#ifndef PATH_MAX
-#define PATH_MAX	4095
-#endif /* ! PATH_MAX */
-
 static int
 monitor_subsys_database_modify(
 	Operation	*op,
@@ -74,7 +66,7 @@ init_readOnly( monitor_info_t *mi, Entry *e, slap_mask_t restrictops )
 	struct berval	*tf = ( ( restrictops & SLAP_RESTRICT_OP_MASK ) == SLAP_RESTRICT_OP_WRITES ) ?
 		(struct berval *)&slap_true_bv : (struct berval *)&slap_false_bv;
 
-	return attr_merge_one( e, mi->mi_ad_readOnly, tf, tf );
+	return attr_merge_one( e, mi->mi_ad_readOnly, tf, NULL );
 }
 
 static int
@@ -107,6 +99,101 @@ init_restrictedOperation( monitor_info_t *mi, Entry *e, slap_mask_t restrictops
 	return LDAP_SUCCESS;
 }
 
+static int
+monitor_subsys_overlay_init_one(
+	monitor_info_t		*mi,
+	BackendDB		*be,
+	monitor_subsys_t	*ms,
+	monitor_subsys_t	*ms_overlay,
+	slap_overinst		*on,
+	Entry			*e_database,
+	Entry			**ep_overlay )
+{
+	char			buf[ BACKMONITOR_BUFSIZE ];
+	int			j, o;
+	Entry			*e_overlay;
+	slap_overinst		*on2;
+	slap_overinfo		*oi = NULL;
+	BackendInfo		*bi;
+	monitor_entry_t		*mp_overlay;
+	struct berval		bv;
+
+	assert( overlay_is_over( be ) );
+
+	oi = (slap_overinfo *)be->bd_info->bi_private;
+	bi = oi->oi_orig;
+
+	/* find the overlay number, o */
+	for ( o = 0, on2 = oi->oi_list; on2 && on2 != on; on2 = on2->on_next, o++ )
+		;
+
+	if ( on2 == NULL ) {
+		return -1;
+	}
+
+	/* find the overlay type number, j */
+	for ( on2 = overlay_next( NULL ), j = 0; on2; on2 = overlay_next( on2 ), j++ ) {
+		if ( on2->on_bi.bi_type == on->on_bi.bi_type ) {
+			break;
+		}
+	}
+	assert( on2 != NULL );
+
+	bv.bv_len = snprintf( buf, sizeof( buf ), "cn=Overlay %d", o );
+	bv.bv_val = buf;
+
+	e_overlay = monitor_entry_stub( &e_database->e_name, &e_database->e_nname, &bv,
+		mi->mi_oc_monitoredObject, NULL, NULL );
+
+	if ( e_overlay == NULL ) {
+		Debug( LDAP_DEBUG_ANY,
+			"monitor_subsys_overlay_init_one: "
+			"unable to create entry "
+			"\"cn=Overlay %d,%s\"\n",
+			o, e_database->e_name.bv_val );
+		return( -1 );
+	}
+	ber_str2bv( on->on_bi.bi_type, 0, 0, &bv );
+	attr_merge_normalize_one( e_overlay, mi->mi_ad_monitoredInfo, &bv, NULL );
+
+	bv.bv_len = snprintf( buf, sizeof( buf ), "cn=Overlay %d,%s",
+		j, ms_overlay->mss_dn.bv_val );
+	bv.bv_val = buf;
+	attr_merge_normalize_one( e_overlay, slap_schema.si_ad_seeAlso,
+		&bv, NULL );
+
+	if ( SLAP_MONITOR( be ) ) {
+		attr_merge( e_overlay, slap_schema.si_ad_monitorContext,
+				be->be_suffix, be->be_nsuffix );
+
+	} else {
+		attr_merge( e_overlay, slap_schema.si_ad_namingContexts,
+				be->be_suffix, NULL );
+	}
+
+	mp_overlay = monitor_entrypriv_create();
+	if ( mp_overlay == NULL ) {
+		return -1;
+	}
+	e_overlay->e_private = ( void * )mp_overlay;
+	mp_overlay->mp_info = ms;
+	mp_overlay->mp_flags = ms->mss_flags | MONITOR_F_SUB;
+	
+	if ( monitor_cache_add( mi, e_overlay ) ) {
+		Debug( LDAP_DEBUG_ANY,
+			"monitor_subsys_overlay_init_one: "
+			"unable to add entry "
+			"\"cn=Overlay %d,%s\"\n",
+			o, e_database->e_name.bv_val );
+		return -1;
+	}
+
+	*ep_overlay = e_overlay;
+	ep_overlay = &mp_overlay->mp_next;
+
+	return 0;
+}
+
 static int
 monitor_subsys_database_init_one(
 	monitor_info_t		*mi,
@@ -129,29 +216,32 @@ monitor_subsys_database_init_one(
 
 	bi = be->bd_info;
 
+	if ( be->be_suffix == NULL ) {
+		Debug( LDAP_DEBUG_ANY,
+			"monitor_subsys_database_init_one: "
+			"missing suffix for %s\n",
+			rdnval );
+		return( -1 );
+	}
+
 	if ( overlay_is_over( be ) ) {
 		oi = (slap_overinfo *)be->bd_info->bi_private;
 		bi = oi->oi_orig;
 	}
 
-	/* Subordinates are not exposed as their own naming context */
-	if ( SLAP_GLUE_SUBORDINATE( be ) ) {
-		return 0;
-	}
-
 	e = monitor_entry_stub( &ms->mss_dn, &ms->mss_ndn, rdn,
-		mi->mi_oc_monitoredObject, mi, NULL, NULL );
+		mi->mi_oc_monitoredObject, NULL, NULL );
 
 	if ( e == NULL ) {
 		Debug( LDAP_DEBUG_ANY,
-			"monitor_subsys_database_init: "
+			"monitor_subsys_database_init_one: "
 			"unable to create entry \"%s,%s\"\n",
-			rdn->bv_val, ms->mss_dn.bv_val, 0 );
+			rdn->bv_val, ms->mss_dn.bv_val );
 		return( -1 );
 	}
 
 	ber_str2bv( bi->bi_type, 0, 0, &bv );
-	attr_merge_one( e, mi->mi_ad_monitoredInfo, &bv, NULL );
+	attr_merge_normalize_one( e, mi->mi_ad_monitoredInfo, &bv, NULL );
 	attr_merge_one( e, mi->mi_ad_monitorIsShadow,
 		SLAP_SHADOW( be ) ? (struct berval *)&slap_true_bv :
 			(struct berval *)&slap_false_bv, NULL );
@@ -163,17 +253,24 @@ monitor_subsys_database_init_one(
 				be->be_suffix, be->be_nsuffix );
 
 	} else {
-		if ( be->be_suffix == NULL ) {
-			Debug( LDAP_DEBUG_ANY,
-				"monitor_subsys_database_init: "
-				"missing suffix for %s\n",
-				rdnval, 0, 0 );
-			return -1;
-		}
 		attr_merge( e, slap_schema.si_ad_namingContexts,
-				be->be_suffix, be->be_nsuffix );
+			be->be_suffix, NULL );
 		attr_merge( e_database, slap_schema.si_ad_namingContexts,
-				be->be_suffix, be->be_nsuffix );
+			be->be_suffix, NULL );
+
+		if ( SLAP_GLUE_SUBORDINATE( be ) ) {
+			BackendDB *sup_be = select_backend( &be->be_nsuffix[ 0 ], 1 );
+			if ( sup_be == NULL ) {
+				Debug( LDAP_DEBUG_ANY,
+					"monitor_subsys_database_init: "
+					"unable to get superior for %s\n",
+					be->be_suffix[ 0 ].bv_val );
+
+			} else {
+				attr_merge( e, mi->mi_ad_monitorSuperiorDN,
+					sup_be->be_suffix, sup_be->be_nsuffix );
+			}
+		}
 	}
 
 	(void)init_readOnly( mi, e, be->be_restrictops );
@@ -249,12 +346,13 @@ monitor_subsys_database_init_one(
 	mp->mp_info = ms;
 	mp->mp_flags = ms->mss_flags
 		| MONITOR_F_SUB;
+	mp->mp_private = be;
 
 	if ( monitor_cache_add( mi, e ) ) {
 		Debug( LDAP_DEBUG_ANY,
-			"monitor_subsys_database_init: "
+			"monitor_subsys_database_init_one: "
 			"unable to add entry \"%s,%s\"\n",
-			rdn->bv_val, ms->mss_dn.bv_val, 0 );
+			rdn->bv_val, ms->mss_dn.bv_val );
 		return( -1 );
 	}
 
@@ -264,74 +362,11 @@ monitor_subsys_database_init_one(
 
 	if ( oi != NULL ) {
 		Entry		**ep_overlay = &mp->mp_children;
-		monitor_entry_t	*mp_overlay;
 		slap_overinst	*on = oi->oi_list;
-		int		o;
 
-		for ( o = 0; on; o++, on = on->on_next ) {
-			Entry			*e_overlay;
-			slap_overinst		*on2;
-
-			/* find the overlay number, j */
-			for ( on2 = overlay_next( NULL ), j = 0; on2; on2 = overlay_next( on2 ), j++ ) {
-				if ( on2->on_bi.bi_type == on->on_bi.bi_type ) {
-					break;
-				}
-			}
-			assert( on2 != NULL );
-
-			bv.bv_len = snprintf( buf, sizeof( buf ), "cn=Overlay %d", o );
-			bv.bv_val = buf;
-
-			e_overlay = monitor_entry_stub( &e->e_name, &e->e_nname, &bv,
-				mi->mi_oc_monitoredObject, mi, NULL, NULL );
-
-			if ( e_overlay == NULL ) {
-				Debug( LDAP_DEBUG_ANY,
-					"monitor_subsys_database_init: "
-					"unable to create entry "
-					"\"cn=Overlay %d,%s,%s\"\n",
-					o, rdn->bv_val, ms->mss_dn.bv_val );
-				return( -1 );
-			}
-			ber_str2bv( on->on_bi.bi_type, 0, 0, &bv );
-			attr_merge_one( e_overlay, mi->mi_ad_monitoredInfo, &bv, NULL );
-
-			bv.bv_len = snprintf( buf, sizeof( buf ), "cn=Overlay %d,%s",
-				j, ms_overlay->mss_dn.bv_val );
-			bv.bv_val = buf;
-			attr_merge_normalize_one( e_overlay, slap_schema.si_ad_seeAlso,
-				&bv, NULL );
-
-			if ( SLAP_MONITOR( be ) ) {
-				attr_merge( e_overlay, slap_schema.si_ad_monitorContext,
-						be->be_suffix, be->be_nsuffix );
-
-			} else {
-				attr_merge( e_overlay, slap_schema.si_ad_namingContexts,
-						be->be_suffix, be->be_nsuffix );
-			}
-
-			mp_overlay = monitor_entrypriv_create();
-			if ( mp_overlay == NULL ) {
-				return -1;
-			}
-			e_overlay->e_private = ( void * )mp_overlay;
-			mp_overlay->mp_info = ms;
-			mp_overlay->mp_flags = ms->mss_flags
-				| MONITOR_F_SUB;
-	
-			if ( monitor_cache_add( mi, e_overlay ) ) {
-				Debug( LDAP_DEBUG_ANY,
-					"monitor_subsys_database_init: "
-					"unable to add entry "
-					"\"cn=Overlay %d,%s,%s\"\n",
-					o, rdn->bv_val, ms->mss_dn.bv_val );
-				return( -1 );
-			}
-
-			*ep_overlay = e_overlay;
-			ep_overlay = &mp_overlay->mp_next;
+		for ( ; on; on = on->on_next ) {
+			monitor_subsys_overlay_init_one( mi, be,
+				ms, ms_overlay, on, e, ep_overlay );
 		}
 	}
 
@@ -341,6 +376,149 @@ monitor_subsys_database_init_one(
 	return 0;
 }
 
+static int
+monitor_back_register_database_and_overlay(
+	BackendDB		*be,
+	struct slap_overinst	*on,
+	struct berval		*ndn_out )
+{
+	monitor_info_t		*mi;
+	Entry			*e_database, **ep;
+	int			i, rc;
+	monitor_entry_t		*mp;
+	monitor_subsys_t	*ms_backend,
+				*ms_database,
+				*ms_overlay;
+	struct berval		bv;
+	char			buf[ BACKMONITOR_BUFSIZE ];
+
+	assert( be_monitor != NULL );
+
+	if ( !monitor_subsys_is_opened() ) {
+		if ( on ) {
+			return monitor_back_register_overlay_limbo( be, on, ndn_out );
+
+		} else {
+			return monitor_back_register_database_limbo( be, ndn_out );
+		}
+	}
+
+	mi = ( monitor_info_t * )be_monitor->be_private;
+
+	ms_backend = monitor_back_get_subsys( SLAPD_MONITOR_BACKEND_NAME );
+	if ( ms_backend == NULL ) {
+		Debug( LDAP_DEBUG_ANY,
+			"monitor_back_register_database: "
+			"unable to get "
+			"\"" SLAPD_MONITOR_BACKEND_NAME "\" "
+			"subsystem\n" );
+		return -1;
+	}
+
+	ms_database = monitor_back_get_subsys( SLAPD_MONITOR_DATABASE_NAME );
+	if ( ms_database == NULL ) {
+		Debug( LDAP_DEBUG_ANY,
+			"monitor_back_register_database: "
+			"unable to get "
+			"\"" SLAPD_MONITOR_DATABASE_NAME "\" "
+			"subsystem\n" );
+		return -1;
+	}
+
+	ms_overlay = monitor_back_get_subsys( SLAPD_MONITOR_OVERLAY_NAME );
+	if ( ms_overlay == NULL ) {
+		Debug( LDAP_DEBUG_ANY,
+			"monitor_back_register_database: "
+			"unable to get "
+			"\"" SLAPD_MONITOR_OVERLAY_NAME "\" "
+			"subsystem\n" );
+		return -1;
+	}
+
+	if ( monitor_cache_get( mi, &ms_database->mss_ndn, &e_database ) ) {
+		Debug( LDAP_DEBUG_ANY,
+			"monitor_subsys_database_init: "
+			"unable to get entry \"%s\"\n",
+			ms_database->mss_ndn.bv_val );
+		return( -1 );
+	}
+
+	mp = ( monitor_entry_t * )e_database->e_private;
+	for ( i = -1, ep = &mp->mp_children; *ep; i++ ) {
+		mp = ( monitor_entry_t * )(*ep)->e_private;
+
+		assert( mp != NULL );
+		if ( mp->mp_private == be->bd_self ) {
+			rc = 0;
+			goto done;
+		}
+		ep = &mp->mp_next;
+	}
+
+	bv.bv_val = buf;
+	bv.bv_len = snprintf( buf, sizeof( buf ), "cn=Database %d", i );
+	if ( bv.bv_len >= sizeof( buf ) ) {
+		rc = -1;
+		goto done;
+	}
+	
+	rc = monitor_subsys_database_init_one( mi, be,
+		ms_database, ms_backend, ms_overlay, &bv, e_database, &ep );
+	if ( rc != 0 ) {
+		goto done;
+	}
+	/* database_init_one advanced ep past where we want.
+	 * But it stored the entry we want in mp->mp_next.
+	 */
+	ep = &mp->mp_next;
+
+done:;
+	monitor_cache_release( mi, e_database );
+	if ( rc == 0 && ndn_out && ep && *ep ) {
+		if ( on ) {
+			Entry *e_ov;
+			struct berval ov_type;
+
+			ber_str2bv( on->on_bi.bi_type, 0, 0, &ov_type );
+
+			mp = ( monitor_entry_t * ) (*ep)->e_private;
+			for ( e_ov = mp->mp_children; e_ov; ) {
+				Attribute *a = attr_find( e_ov->e_attrs, mi->mi_ad_monitoredInfo );
+
+				if ( a != NULL && bvmatch( &a->a_nvals[ 0 ], &ov_type ) ) {
+					*ndn_out = e_ov->e_nname;
+					break;
+				}
+
+				mp = ( monitor_entry_t * ) e_ov->e_private;
+				e_ov = mp->mp_next;
+			}
+			
+		} else {
+			*ndn_out = (*ep)->e_nname;
+		}
+	}
+
+	return rc;
+}
+
+int
+monitor_back_register_database(
+	BackendDB		*be,
+	struct berval		*ndn_out )
+{
+	return monitor_back_register_database_and_overlay( be, NULL, ndn_out );
+}
+
+int
+monitor_back_register_overlay(
+	BackendDB		*be,
+	struct slap_overinst	*on,
+	struct berval		*ndn_out )
+{
+	return monitor_back_register_database_and_overlay( be, on, ndn_out );
+}
+
 int
 monitor_subsys_database_init(
 	BackendDB		*be,
@@ -366,8 +544,7 @@ monitor_subsys_database_init(
 			"monitor_subsys_database_init: "
 			"unable to get "
 			"\"" SLAPD_MONITOR_BACKEND_NAME "\" "
-			"subsystem\n",
-			0, 0, 0 );
+			"subsystem\n" );
 		return -1;
 	}
 
@@ -377,8 +554,7 @@ monitor_subsys_database_init(
 			"monitor_subsys_database_init: "
 			"unable to get "
 			"\"" SLAPD_MONITOR_OVERLAY_NAME "\" "
-			"subsystem\n",
-			0, 0, 0 );
+			"subsystem\n" );
 		return -1;
 	}
 
@@ -386,7 +562,7 @@ monitor_subsys_database_init(
 		Debug( LDAP_DEBUG_ANY,
 			"monitor_subsys_database_init: "
 			"unable to get entry \"%s\"\n",
-			ms->mss_ndn.bv_val, 0, 0 );
+			ms->mss_ndn.bv_val );
 		return( -1 );
 	}
 
@@ -721,6 +897,7 @@ monitor_subsys_database_modify(
 	
 							BER_BVZERO( &a->a_vals[ k - 1 ] );
 							BER_BVZERO( &a->a_nvals[ k - 1 ] );
+							a->a_numvals--;
 						}
 					}
 				}
@@ -746,9 +923,16 @@ monitor_subsys_database_modify(
 	
 							BER_BVZERO( &a->a_vals[ k - 1 ] );
 							BER_BVZERO( &a->a_nvals[ k - 1 ] );
+							a->a_numvals--;
 						}
 					}
 				}
+
+				if ( a->a_vals == NULL ) {
+					assert( a->a_numvals == 0 );
+
+					attr_delete( &e->e_attrs, mi->mi_ad_restrictedOperation );
+				}
 			}
 		}
 
@@ -812,17 +996,21 @@ monitor_back_add_plugin( monitor_info_t *mi, Backend *be, Entry *e_database )
 		if ( rc != LDAP_SUCCESS ) {
 			goto done;
 		}
-
-		snprintf( buf, sizeof(buf),
-				"plugin %d name: %s; "
-				"vendor: %s; "
-				"version: %s; "
-				"description: %s", 
-				i,
-				srchdesc->spd_id,
-				srchdesc->spd_vendor,
-				srchdesc->spd_version,
-				srchdesc->spd_description );
+		if ( srchdesc ) {
+			snprintf( buf, sizeof(buf),
+					"plugin %d name: %s; "
+					"vendor: %s; "
+					"version: %s; "
+					"description: %s", 
+					i,
+					srchdesc->spd_id,
+					srchdesc->spd_vendor,
+					srchdesc->spd_version,
+					srchdesc->spd_description );
+		} else {
+			snprintf( buf, sizeof(buf),
+					"plugin %d name: <no description available>", i );
+		}
 
 		ber_str2bv( buf, 0, 0, &bv );
 		attr_merge_normalize_one( e_database,
