@@ -1,8 +1,8 @@
 /* init.c - initialize monitor backend */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-monitor/init.c,v 1.125.2.2 2007/02/13 20:22:26 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2001-2007 The OpenLDAP Foundation.
+ * Copyright 2001-2020 The OpenLDAP Foundation.
  * Portions Copyright 2001-2003 Pierangelo Masarati.
  * All rights reserved.
  *
@@ -52,6 +52,10 @@ static const monitor_extra_t monitor_extra = {
 	monitor_back_get_subsys_by_dn,
 
 	monitor_back_register_subsys,
+	monitor_back_register_backend,
+	monitor_back_register_database,
+	monitor_back_register_overlay_info,
+	monitor_back_register_overlay,
 	monitor_back_register_entry,
 	monitor_back_register_entry_parent,
 	monitor_back_register_entry_attrs,
@@ -60,7 +64,12 @@ static const monitor_extra_t monitor_extra = {
 	monitor_back_unregister_entry,
 	monitor_back_unregister_entry_parent,
 	monitor_back_unregister_entry_attrs,
-	monitor_back_unregister_entry_callback
+	monitor_back_unregister_entry_callback,
+
+	monitor_back_entry_stub,
+	monitor_back_entrypriv_create,
+	monitor_back_register_subsys_late,
+	monitor_back_entry_get_unlocked
 };
 	
 
@@ -221,6 +230,12 @@ static struct monitor_subsys_t known_monitor_subsys[] = {
        	}, { NULL }
 };
 
+int
+monitor_subsys_is_opened( void )
+{
+	return monitor_subsys_opened;
+}
+
 int
 monitor_back_register_subsys(
 	monitor_subsys_t	*ms )
@@ -245,7 +260,7 @@ monitor_back_register_subsys(
 	/* if a subsystem is registered __AFTER__ subsystem 
 	 * initialization (depending on the sequence the databases
 	 * are listed in slapd.conf), init it */
-	if ( monitor_subsys_opened ) {
+	if ( monitor_subsys_is_opened() ) {
 
 		/* FIXME: this should only be possible
 		 * if be_monitor is already initialized */
@@ -265,14 +280,24 @@ enum {
 	LIMBO_ENTRY,
 	LIMBO_ENTRY_PARENT,
 	LIMBO_ATTRS,
-	LIMBO_CB
+	LIMBO_CB,
+	LIMBO_BACKEND,
+	LIMBO_DATABASE,
+	LIMBO_OVERLAY_INFO,
+	LIMBO_OVERLAY,
+	LIMBO_SUBSYS,
+
+	LIMBO_LAST
 };
 
 typedef struct entry_limbo_t {
 	int			el_type;
+	BackendInfo		*el_bi;
+	BackendDB		*el_be;
+	slap_overinst		*el_on;
 	Entry			*el_e;
 	Attribute		*el_a;
-	struct berval		el_ndn;
+	struct berval		*el_ndn;
 	struct berval		el_nbase;
 	int			el_scope;
 	struct berval		el_filter;
@@ -288,6 +313,147 @@ monitor_back_is_configured( void )
 	return be_monitor != NULL;
 }
 
+int
+monitor_back_register_subsys_late(
+	monitor_subsys_t	*ms )
+{
+	entry_limbo_t	**elpp, el = { 0 };
+	monitor_info_t 	*mi;
+
+	if ( be_monitor == NULL ) {
+		Debug( LDAP_DEBUG_ANY,
+			"monitor_back_register_subsys_late: "
+			"monitor database not configured.\n" );
+		return -1;
+	}
+
+	/* everything is ready, can register already */
+	if ( monitor_subsys_is_opened() ) {
+		return monitor_back_register_subsys( ms );
+	}
+
+	mi = ( monitor_info_t * )be_monitor->be_private;
+
+
+	el.el_type = LIMBO_SUBSYS;
+
+	el.el_mss = ms;
+
+	for ( elpp = &mi->mi_entry_limbo;
+			*elpp;
+			elpp = &(*elpp)->el_next )
+		/* go to last */;
+
+	*elpp = (entry_limbo_t *)ch_malloc( sizeof( entry_limbo_t ) );
+
+	el.el_next = NULL;
+	**elpp = el;
+
+	return 0;
+}
+
+int
+monitor_back_register_backend(
+	BackendInfo		*bi )
+{
+	return -1;
+}
+
+int
+monitor_back_register_overlay_info(
+	slap_overinst		*on )
+{
+	return -1;
+}
+
+int
+monitor_back_register_backend_limbo(
+	BackendInfo		*bi )
+{
+	return -1;
+}
+
+int
+monitor_back_register_database_limbo(
+	BackendDB		*be,
+	struct berval		*ndn_out )
+{
+	entry_limbo_t	**elpp, el = { 0 };
+	monitor_info_t 	*mi;
+
+	if ( be_monitor == NULL ) {
+		Debug( LDAP_DEBUG_ANY,
+			"monitor_back_register_database_limbo: "
+			"monitor database not configured.\n" );
+		return -1;
+	}
+
+	mi = ( monitor_info_t * )be_monitor->be_private;
+
+
+	el.el_type = LIMBO_DATABASE;
+
+	el.el_be = be->bd_self;
+	el.el_ndn = ndn_out;
+	
+	for ( elpp = &mi->mi_entry_limbo;
+			*elpp;
+			elpp = &(*elpp)->el_next )
+		/* go to last */;
+
+	*elpp = (entry_limbo_t *)ch_malloc( sizeof( entry_limbo_t ) );
+
+	el.el_next = NULL;
+	**elpp = el;
+
+	return 0;
+}
+
+int
+monitor_back_register_overlay_info_limbo(
+	slap_overinst		*on )
+{
+	return -1;
+}
+
+int
+monitor_back_register_overlay_limbo(
+	BackendDB		*be,
+	struct slap_overinst	*on,
+	struct berval		*ndn_out )
+{
+	entry_limbo_t	**elpp, el = { 0 };
+	monitor_info_t 	*mi;
+
+	if ( be_monitor == NULL ) {
+		Debug( LDAP_DEBUG_ANY,
+			"monitor_back_register_overlay_limbo: "
+			"monitor database not configured.\n" );
+		return -1;
+	}
+
+	mi = ( monitor_info_t * )be_monitor->be_private;
+
+
+	el.el_type = LIMBO_OVERLAY;
+
+	el.el_be = be->bd_self;
+	el.el_on = on;
+	el.el_ndn = ndn_out;
+	
+	for ( elpp = &mi->mi_entry_limbo;
+			*elpp;
+			elpp = &(*elpp)->el_next )
+		/* go to last */;
+
+	*elpp = (entry_limbo_t *)ch_malloc( sizeof( entry_limbo_t ) );
+
+	el.el_next = NULL;
+	**elpp = el;
+
+	return 0;
+}
+
 int
 monitor_back_register_entry(
 	Entry			*e,
@@ -301,7 +467,7 @@ monitor_back_register_entry(
 		Debug( LDAP_DEBUG_ANY,
 			"monitor_back_register_entry(\"%s\"): "
 			"monitor database not configured.\n",
-			e->e_name.bv_val, 0, 0 );
+			e->e_name.bv_val );
 		return -1;
 	}
 
@@ -311,7 +477,7 @@ monitor_back_register_entry(
 	assert( e != NULL );
 	assert( e->e_private == NULL );
 	
-	if ( monitor_subsys_opened ) {
+	if ( monitor_subsys_is_opened() ) {
 		Entry		*e_parent = NULL,
 				*e_new = NULL,
 				**ep = NULL;
@@ -325,7 +491,7 @@ monitor_back_register_entry(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_register_entry(\"%s\"): "
 				"entry exists\n",
-				e->e_name.bv_val, 0, 0 );
+				e->e_name.bv_val );
 			monitor_cache_release( mi, e_parent );
 			return -1;
 		}
@@ -336,7 +502,7 @@ monitor_back_register_entry(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_register_entry(\"%s\"): "
 				"parent \"%s\" not found\n",
-				e->e_name.bv_val, pdn.bv_val, 0 );
+				e->e_name.bv_val, pdn.bv_val );
 			return -1;
 		}
 
@@ -348,7 +514,7 @@ monitor_back_register_entry(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_register_entry(\"%s\"): "
 				"parent \"%s\" is volatile\n",
-				e->e_name.bv_val, e_parent->e_name.bv_val, 0 );
+				e->e_name.bv_val, e_parent->e_name.bv_val );
 			rc = -1;
 			goto done;
 		}
@@ -358,7 +524,7 @@ monitor_back_register_entry(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_register_entry(\"%s\"): "
 				"monitor_entrypriv_create() failed\n",
-				e->e_name.bv_val, 0, 0 );
+				e->e_name.bv_val );
 			rc = -1;
 			goto done;
 		}
@@ -368,7 +534,7 @@ monitor_back_register_entry(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_register_entry(\"%s\"): "
 				"entry_dup() failed\n",
-				e->e_name.bv_val, 0, 0 );
+				e->e_name.bv_val );
 			rc = -1;
 			goto done;
 		}
@@ -395,7 +561,7 @@ monitor_back_register_entry(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_register_entry(\"%s\"): "
 				"unable to add entry\n",
-				e->e_name.bv_val, 0, 0 );
+				e->e_name.bv_val );
 			rc = -1;
 			goto done;
 		}
@@ -425,7 +591,7 @@ done:;
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_register_entry(\"%s\"): "
 				"entry_dup() failed\n",
-				e->e_name.bv_val, 0, 0 );
+				e->e_name.bv_val );
 			return -1;
 		}
 		
@@ -433,7 +599,7 @@ done:;
 		el.el_mss = mss;
 		el.el_flags = flags;
 
-		for ( elpp = (entry_limbo_t **)&mi->mi_entry_limbo;
+		for ( elpp = &mi->mi_entry_limbo;
 				*elpp;
 				elpp = &(*elpp)->el_next )
 			/* go to last */;
@@ -485,12 +651,11 @@ monitor_back_register_entry_parent(
 		/* need a filter */
 		Debug( LDAP_DEBUG_ANY,
 			"monitor_back_register_entry_parent(\"\"): "
-			"need a valid filter\n",
-			0, 0, 0 );
+			"need a valid filter\n" );
 		return -1;
 	}
 
-	if ( monitor_subsys_opened ) {
+	if ( monitor_subsys_is_opened() ) {
 		Entry		*e_parent = NULL,
 				*e_new = NULL,
 				**ep = NULL;
@@ -517,7 +682,7 @@ monitor_back_register_entry_parent(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_register_entry_parent(\"%s\"): "
 				"parent entry does not exist\n",
-				ndn.bv_val, 0, 0 );
+				ndn.bv_val );
 			rc = -1;
 			goto done;
 		}
@@ -530,7 +695,7 @@ monitor_back_register_entry_parent(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_register_entry_parent(\"%s\"): "
 				"entry is volatile\n",
-				e_parent->e_name.bv_val, 0, 0 );
+				e_parent->e_name.bv_val );
 			rc = -1;
 			goto done;
 		}
@@ -543,7 +708,7 @@ monitor_back_register_entry_parent(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_register_entry_parent(\"%s\"): "
 				"entry already exists\n",
-				e_name.bv_val, 0, 0 );
+				e_name.bv_val );
 			monitor_cache_release( mi, e_new );
 			e_new = NULL;
 			rc = -1;
@@ -555,7 +720,7 @@ monitor_back_register_entry_parent(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_register_entry_parent(\"%s\"): "
 				"monitor_entrypriv_create() failed\n",
-				e->e_name.bv_val, 0, 0 );
+				e->e_name.bv_val );
 			rc = -1;
 			goto done;
 		}
@@ -565,7 +730,7 @@ monitor_back_register_entry_parent(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_register_entry(\"%s\"): "
 				"entry_dup() failed\n",
-				e->e_name.bv_val, 0, 0 );
+				e->e_name.bv_val );
 			rc = -1;
 			goto done;
 		}
@@ -596,7 +761,7 @@ monitor_back_register_entry_parent(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_register_entry(\"%s\"): "
 				"unable to add entry\n",
-				e->e_name.bv_val, 0, 0 );
+				e->e_name.bv_val );
 			rc = -1;
 			goto done;
 		}
@@ -630,7 +795,7 @@ done:;
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_register_entry(\"%s\"): "
 				"entry_dup() failed\n",
-				e->e_name.bv_val, 0, 0 );
+				e->e_name.bv_val );
 			goto done_limbo;
 		}
 		
@@ -647,7 +812,7 @@ done:;
 		el.el_mss = mss;
 		el.el_flags = flags;
 
-		for ( elpp = (entry_limbo_t **)&mi->mi_entry_limbo;
+		for ( elpp = &mi->mi_entry_limbo;
 				*elpp;
 				elpp = &(*elpp)->el_next )
 			/* go to last */;
@@ -707,7 +872,7 @@ monitor_search2ndn(
 	OperationBuffer	opbuf;
 	Operation	*op;
 	void	*thrctx;
-	SlapReply	rs = { 0 };
+	SlapReply	rs = { REP_RESULT };
 	slap_callback	cb = { NULL, monitor_search2ndn_cb, NULL, NULL };
 	int		rc;
 
@@ -717,9 +882,9 @@ monitor_search2ndn(
 		return -1;
 	}
 
-	op = (Operation *) &opbuf;
 	thrctx = ldap_pvt_thread_pool_context();
-	connection_fake_init( &conn, op, thrctx );
+	connection_fake_init2( &conn, &opbuf, thrctx, 0 );
+	op = &opbuf.ob_op;
 
 	op->o_tag = LDAP_REQ_SEARCH;
 
@@ -770,7 +935,7 @@ monitor_search2ndn(
 
 cleanup:;
 	if ( op->ors_filter != NULL ) {
-		filter_free_x( op, op->ors_filter );
+		filter_free_x( op, op->ors_filter, 1 );
 	}
 	if ( !BER_BVISNULL( &op->ors_filterstr ) ) {
 		op->o_tmpfree( op->ors_filterstr.bv_val, op->o_tmpmemctx );
@@ -818,18 +983,17 @@ monitor_back_register_entry_attrs(
 	monitor_info_t 	*mi;
 	struct berval	ndn = BER_BVNULL;
 	char		*fname = ( a == NULL ? "callback" : "attrs" );
+	struct berval	empty_bv = BER_BVC("");
+
+	if ( nbase == NULL ) nbase = &empty_bv;
+	if ( filter == NULL ) filter = &empty_bv;
 
 	if ( be_monitor == NULL ) {
-		char		buf[ SLAP_TEXT_BUFLEN ];
-
-		snprintf( buf, sizeof( buf ),
-			"monitor_back_register_entry_%s(base=\"%s\" scope=%s filter=\"%s\"): "
-			"monitor database not configured.\n",
-			fname,
-			BER_BVISNULL( nbase ) ? "" : nbase->bv_val,
-			ldap_pvt_scope2str( scope ),
-			BER_BVISNULL( filter ) ? "" : filter->bv_val );
-		Debug( LDAP_DEBUG_ANY, "%s\n", buf, 0, 0 );
+		Debug(LDAP_DEBUG_ANY,
+		      "monitor_back_register_entry_%s(base=\"%s\" scope=%s filter=\"%s\"): " "monitor database not configured.\n\n",
+		      fname, BER_BVISNULL(nbase) ? "" : nbase->bv_val,
+		      ldap_pvt_scope2str(scope),
+		      BER_BVISNULL(filter) ? "" : filter->bv_val );
 
 		return -1;
 	}
@@ -854,11 +1018,11 @@ monitor_back_register_entry_attrs(
 		Debug( LDAP_DEBUG_ANY,
 			"monitor_back_register_entry_%s(\"\"): "
 			"need a valid filter\n",
-			fname, 0, 0 );
+			fname );
 		return -1;
 	}
 
-	if ( monitor_subsys_opened ) {
+	if ( monitor_subsys_is_opened() ) {
 		Entry			*e = NULL;
 		Attribute		**atp = NULL;
 		monitor_entry_t 	*mp = NULL;
@@ -868,19 +1032,12 @@ monitor_back_register_entry_attrs(
 
 		if ( BER_BVISNULL( &ndn ) ) {
 			if ( monitor_search2ndn( nbase, scope, filter, &ndn ) ) {
-				char		buf[ SLAP_TEXT_BUFLEN ];
-
-				snprintf( buf, sizeof( buf ),
-					"monitor_back_register_entry_%s(\"\"): "
-					"base=\"%s\" scope=%s filter=\"%s\": "
-					"unable to find entry\n",
-					fname,
-					nbase->bv_val ? nbase->bv_val : "\"\"",
-					ldap_pvt_scope2str( scope ),
-					filter->bv_val );
-
-				/* entry does not exist */
-				Debug( LDAP_DEBUG_ANY, "%s\n", buf, 0, 0 );
+				Debug(LDAP_DEBUG_ANY,
+				      "monitor_back_register_entry_%s(\"\"): " "base=\"%s\" scope=%s filter=\"%s\": " "unable to find entry\n\n",
+				      fname,
+				      nbase->bv_val ? nbase->bv_val : "\"\"",
+				      ldap_pvt_scope2str(scope),
+				      filter->bv_val );
 				return -1;
 			}
 
@@ -892,7 +1049,7 @@ monitor_back_register_entry_attrs(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_register_entry_%s(\"%s\"): "
 				"entry does not exist\n",
-				fname, ndn.bv_val, 0 );
+				fname, ndn.bv_val );
 			rc = -1;
 			goto done;
 		}
@@ -905,7 +1062,7 @@ monitor_back_register_entry_attrs(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_register_entry_%s(\"%s\"): "
 				"entry is volatile\n",
-				fname, e->e_name.bv_val, 0 );
+				fname, e->e_name.bv_val );
 			rc = -1;
 			goto done;
 		}
@@ -928,7 +1085,7 @@ monitor_back_register_entry_attrs(
 						Debug( LDAP_DEBUG_ANY,
 							"monitor_back_register_entry_%s(\"%s\"): "
 							"attr_dup() failed\n",
-							fname, e->e_name.bv_val, 0 );
+							fname, e->e_name.bv_val );
 						rc = -1;
 						goto done;
 					}
@@ -966,9 +1123,7 @@ done:;
 		entry_limbo_t	**elpp, el = { 0 };
 
 		el.el_type = LIMBO_ATTRS;
-		if ( !BER_BVISNULL( &ndn ) ) {
-			ber_dupbv( &el.el_ndn, &ndn );
-		}
+		el.el_ndn = ndn_in;
 		if ( !BER_BVISNULL( nbase ) ) {
 			ber_dupbv( &el.el_nbase, nbase);
 		}
@@ -980,23 +1135,13 @@ done:;
 		el.el_a = attrs_dup( a );
 		el.el_cb = cb;
 
-		for ( elpp = (entry_limbo_t **)&mi->mi_entry_limbo;
+		for ( elpp = &mi->mi_entry_limbo;
 				*elpp;
 				elpp = &(*elpp)->el_next )
 			/* go to last */;
 
 		*elpp = (entry_limbo_t *)ch_malloc( sizeof( entry_limbo_t ) );
 		if ( *elpp == NULL ) {
-			el.el_e->e_private = NULL;
-			entry_free( el.el_e );
-			return -1;
-		}
-
-		if ( *elpp != NULL ) {
-			el.el_next = NULL;
-			**elpp = el;
-
-		} else {
 			if ( !BER_BVISNULL( &el.el_filter ) ) {
 				ch_free( el.el_filter.bv_val );
 			}
@@ -1006,11 +1151,11 @@ done:;
 			if ( !BER_BVISNULL( &el.el_nbase ) ) {
 				ch_free( &el.el_nbase.bv_val );
 			}
-			if ( !BER_BVISNULL( &el.el_ndn ) ) {
-				ch_free( el.el_ndn.bv_val );
-			}
 			return -1;
 		}
+
+		el.el_next = NULL;
+		**elpp = el;
 	}
 
 	return 0;
@@ -1043,7 +1188,7 @@ monitor_back_unregister_entry(
 		Debug( LDAP_DEBUG_ANY,
 			"monitor_back_unregister_entry(\"%s\"): "
 			"monitor database not configured.\n",
-			ndn->bv_val, 0, 0 );
+			ndn->bv_val );
 
 		return -1;
 	}
@@ -1058,7 +1203,7 @@ monitor_back_unregister_entry(
 
 	assert( mi != NULL );
 
-	if ( monitor_subsys_opened ) {
+	if ( monitor_subsys_is_opened() ) {
 		Entry			*e = NULL;
 		monitor_entry_t 	*mp = NULL;
 		monitor_callback_t	*cb = NULL;
@@ -1068,7 +1213,7 @@ monitor_back_unregister_entry(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_unregister_entry(\"%s\"): "
 				"entry removal failed.\n",
-				ndn->bv_val, 0, 0 );
+				ndn->bv_val );
 			return -1;
 		}
 
@@ -1093,7 +1238,7 @@ monitor_back_unregister_entry(
 	} else {
 		entry_limbo_t	**elpp;
 
-		for ( elpp = (entry_limbo_t **)&mi->mi_entry_limbo;
+		for ( elpp = &mi->mi_entry_limbo;
 			*elpp;
 			elpp = &(*elpp)->el_next )
 		{
@@ -1107,6 +1252,9 @@ monitor_back_unregister_entry(
 				for ( cb = elp->el_cb; cb; cb = next ) {
 					/* FIXME: call callbacks? */
 					next = cb->mc_next;
+					if ( cb->mc_dispose ) {
+						cb->mc_dispose( &cb->mc_private );
+					}
 					ch_free( cb );
 				}
 				assert( elp->el_e != NULL );
@@ -1166,12 +1314,11 @@ monitor_back_unregister_entry_parent(
 		/* need a filter */
 		Debug( LDAP_DEBUG_ANY,
 			"monitor_back_unregister_entry_parent(\"\"): "
-			"need a valid filter\n",
-			0, 0, 0 );
+			"need a valid filter\n" );
 		return -1;
 	}
 
-	if ( monitor_subsys_opened ) {
+	if ( monitor_subsys_is_opened() ) {
 		Entry			*e = NULL;
 		monitor_entry_t 	*mp = NULL;
 
@@ -1192,7 +1339,7 @@ monitor_back_unregister_entry_parent(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_unregister_entry(\"%s\"): "
 				"entry removal failed.\n",
-				ndn.bv_val, 0, 0 );
+				ndn.bv_val );
 			ber_memfree( ndn.bv_val );
 			return -1;
 		}
@@ -1224,7 +1371,7 @@ monitor_back_unregister_entry_parent(
 	} else {
 		entry_limbo_t	**elpp;
 
-		for ( elpp = (entry_limbo_t **)&mi->mi_entry_limbo;
+		for ( elpp = &mi->mi_entry_limbo;
 			*elpp;
 			elpp = &(*elpp)->el_next )
 		{
@@ -1241,6 +1388,9 @@ monitor_back_unregister_entry_parent(
 				for ( cb = elp->el_cb; cb; cb = next ) {
 					/* FIXME: call callbacks? */
 					next = cb->mc_next;
+					if ( cb->mc_dispose ) {
+						cb->mc_dispose( &cb->mc_private );
+					}
 					ch_free( cb );
 				}
 				assert( elp->el_e != NULL );
@@ -1282,16 +1432,11 @@ monitor_back_unregister_entry_attrs(
 	char		*fname = ( target_a == NULL ? "callback" : "attrs" );
 
 	if ( be_monitor == NULL ) {
-		char		buf[ SLAP_TEXT_BUFLEN ];
-
-		snprintf( buf, sizeof( buf ),
-			"monitor_back_unregister_entry_%s(base=\"%s\" scope=%s filter=\"%s\"): "
-			"monitor database not configured.\n",
-			fname,
-			BER_BVISNULL( nbase ) ? "" : nbase->bv_val,
-			ldap_pvt_scope2str( scope ),
-			BER_BVISNULL( filter ) ? "" : filter->bv_val );
-		Debug( LDAP_DEBUG_ANY, "%s\n", buf, 0, 0 );
+		Debug(LDAP_DEBUG_ANY,
+		      "monitor_back_unregister_entry_%s(base=\"%s\" scope=%s filter=\"%s\"): " "monitor database not configured.\n\n",
+		      fname, BER_BVISNULL(nbase) ? "" : nbase->bv_val,
+		      ldap_pvt_scope2str(scope),
+		      BER_BVISNULL(filter) ? "" : filter->bv_val );
 
 		return -1;
 	}
@@ -1322,29 +1467,22 @@ monitor_back_unregister_entry_attrs(
 		Debug( LDAP_DEBUG_ANY,
 			"monitor_back_unregister_entry_%s(\"\"): "
 			"need a valid filter\n",
-			fname, 0, 0 );
+			fname );
 		return -1;
 	}
 
-	if ( monitor_subsys_opened ) {
+	if ( monitor_subsys_is_opened() ) {
 		Entry			*e = NULL;
 		monitor_entry_t 	*mp = NULL;
 		int			freeit = 0;
 
 		if ( BER_BVISNULL( &ndn ) ) {
 			if ( monitor_search2ndn( nbase, scope, filter, &ndn ) ) {
-				char		buf[ SLAP_TEXT_BUFLEN ];
-
-				snprintf( buf, sizeof( buf ),
-					"monitor_back_unregister_entry_%s(\"\"): "
-					"base=\"%s\" scope=%d filter=\"%s\": "
-					"unable to find entry\n",
-					fname,
-					nbase->bv_val ? nbase->bv_val : "\"\"",
-					scope, filter->bv_val );
-
-				/* entry does not exist */
-				Debug( LDAP_DEBUG_ANY, "%s\n", buf, 0, 0 );
+				Debug(LDAP_DEBUG_ANY,
+				      "monitor_back_unregister_entry_%s(\"\"): " "base=\"%s\" scope=%d filter=\"%s\": " "unable to find entry\n\n",
+				      fname,
+				      nbase->bv_val ? nbase->bv_val : "\"\"",
+				      scope, filter->bv_val );
 				return -1;
 			}
 
@@ -1356,7 +1494,7 @@ monitor_back_unregister_entry_attrs(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_unregister_entry(\"%s\"): "
 				"entry removal failed.\n",
-				ndn.bv_val, 0, 0 );
+				ndn.bv_val );
 			return -1;
 		}
 
@@ -1400,14 +1538,12 @@ monitor_back_unregister_entry_attrs(
 			ber_memfree( ndn.bv_val );
 		}
 
-		if ( e ) {
-			monitor_cache_release( mi, e );
-		}
+		monitor_cache_release( mi, e );
 
 	} else {
 		entry_limbo_t	**elpp;
 
-		for ( elpp = (entry_limbo_t **)&mi->mi_entry_limbo;
+		for ( elpp = &mi->mi_entry_limbo;
 			*elpp;
 			elpp = &(*elpp)->el_next )
 		{
@@ -1423,6 +1559,9 @@ monitor_back_unregister_entry_attrs(
 				for ( cb = elp->el_cb; cb; cb = next ) {
 					/* FIXME: call callbacks? */
 					next = cb->mc_next;
+					if ( cb->mc_dispose ) {
+						cb->mc_dispose( &cb->mc_private );
+					}
 					ch_free( cb );
 				}
 				assert( elp->el_e == NULL );
@@ -1803,6 +1942,15 @@ monitor_back_initialize(
 			"SINGLE-VALUE "
 			"USAGE dSAOperation )", SLAP_AT_HIDE,
 			offsetof(monitor_info_t, mi_ad_monitorRuntimeConfig) },
+		{ "( 1.3.6.1.4.1.4203.666.1.55.30 "
+			"NAME 'monitorSuperiorDN' "
+			"DESC 'monitor superior DN' "
+			/* "SUP distinguishedName " */
+			"EQUALITY distinguishedNameMatch "
+			"SYNTAX 1.3.6.1.4.1.1466.115.121.1.12 "
+			"NO-USER-MODIFICATION "
+			"USAGE dSAOperation )", SLAP_AT_FINAL|SLAP_AT_HIDE,
+			offsetof(monitor_info_t, mi_ad_monitorSuperiorDN) },
 		{ NULL, 0, -1 }
 	};
 
@@ -1814,9 +1962,10 @@ monitor_back_initialize(
 		{ "olmSubSystemAttributes",		"olmAttributes:0" },
 		{ "olmGenericAttributes",		"olmSubSystemAttributes:0" },
 		{ "olmDatabaseAttributes",		"olmSubSystemAttributes:1" },
+		{ "olmOverlayAttributes",		"olmSubSystemAttributes:2" },
 
-		/* for example, back-bdb specific attrs
-		 * are in "olmDatabaseAttributes:1"
+		/* for example, back-mdb specific attrs
+		 * are in "olmDatabaseAttributes:12"
 		 *
 		 * NOTE: developers, please record here OID assignments
 		 * for other modules */
@@ -1825,9 +1974,10 @@ monitor_back_initialize(
 		{ "olmSubSystemObjectClasses",		"olmObjectClasses:0" },
 		{ "olmGenericObjectClasses",		"olmSubSystemObjectClasses:0" },
 		{ "olmDatabaseObjectClasses",		"olmSubSystemObjectClasses:1" },
+		{ "olmOverlayObjectClasses",		"olmSubSystemObjectClasses:2" },
 
-		/* for example, back-bdb specific objectClasses
-		 * are in "olmDatabaseObjectClasses:1"
+		/* for example, back-mdb specific objectClasses
+		 * are in "olmDatabaseObjectClasses:12"
 		 *
 		 * NOTE: developers, please record here OID assignments
 		 * for other modules */
@@ -1853,7 +2003,7 @@ monitor_back_initialize(
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_back_initialize: unable to add "
 				"objectIdentifier \"%s=%s\"\n",
-				s_oid[ i ].name, s_oid[ i ].oid, 0 );
+				s_oid[ i ].name, s_oid[ i ].oid );
 			return 1;
 		}
 	}
@@ -1869,7 +2019,7 @@ monitor_back_initialize(
 
 		if ( code ) {
 			Debug( LDAP_DEBUG_ANY,
-				"monitor_back_db_init: register_at failed\n", 0, 0, 0 );
+				"monitor_back_db_init: register_at failed\n" );
 			return -1;
 		}
 		(*ad)->ad_type->sat_flags |= mat[ i ].flags;
@@ -1883,7 +2033,7 @@ monitor_back_initialize(
 		code = register_oc( moc[ i ].schema, Oc, 0 );
 		if ( code ) {
 			Debug( LDAP_DEBUG_ANY,
-				"monitor_back_db_init: register_oc failed\n", 0, 0, 0 );
+				"monitor_back_db_init: register_oc failed\n" );
 			return -1;
 		}
 		(*Oc)->soc_flags |= moc[ i ].flags;
@@ -1917,7 +2067,7 @@ monitor_back_initialize(
 
 	bi->bi_extended = 0;
 
-	bi->bi_entry_release_rw = 0;
+	bi->bi_entry_release_rw = monitor_back_release;
 	bi->bi_chk_referrals = 0;
 	bi->bi_operational = monitor_back_operational;
 
@@ -1927,13 +2077,13 @@ monitor_back_initialize(
 	bi->bi_tool_entry_open = 0;
 	bi->bi_tool_entry_close = 0;
 	bi->bi_tool_entry_first = 0;
+	bi->bi_tool_entry_first_x = 0;
 	bi->bi_tool_entry_next = 0;
 	bi->bi_tool_entry_get = 0;
 	bi->bi_tool_entry_put = 0;
 	bi->bi_tool_entry_reindex = 0;
 	bi->bi_tool_sync = 0;
 	bi->bi_tool_dn2id_get = 0;
-	bi->bi_tool_id2entry_get = 0;
 	bi->bi_tool_entry_modify = 0;
 
 	bi->bi_connection_init = 0;
@@ -1956,7 +2106,8 @@ monitor_back_initialize(
 
 int
 monitor_back_db_init(
-	BackendDB	*be )
+	BackendDB	*be,
+	ConfigReply	*c)
 {
 	int			rc;
 	struct berval		dn = BER_BVC( SLAPD_MONITOR_DN ),
@@ -1966,6 +2117,17 @@ monitor_back_db_init(
 
 	monitor_subsys_t	*ms;
 
+	/*
+	 * database monitor can be defined once only
+	 */
+	if ( be_monitor != NULL ) {
+		if (c) {
+			snprintf(c->msg, sizeof(c->msg),"only one monitor database allowed");
+		}
+		return( -1 );
+	}
+	be_monitor = be;
+
 	/*
 	 * register subsys
 	 */
@@ -1975,16 +2137,6 @@ monitor_back_db_init(
 		}
 	}
 
-	/*
-	 * database monitor can be defined once only
-	 */
-	if ( be_monitor != NULL ) {
-		Debug( LDAP_DEBUG_ANY,
-			"only one monitor database is allowed\n", 0, 0, 0 );
-		return( -1 );
-	}
-	be_monitor = be;
-
 	/* indicate system schema supported */
 	SLAP_BFLAGS(be) |= SLAP_BFLAG_MONITOR;
 
@@ -1992,7 +2144,7 @@ monitor_back_db_init(
 	if( rc != LDAP_SUCCESS ) {
 		Debug( LDAP_DEBUG_ANY,
 			"unable to normalize/pretty monitor DN \"%s\" (%d)\n",
-			dn.bv_val, rc, 0 );
+			dn.bv_val, rc );
 		return -1;
 	}
 
@@ -2005,7 +2157,7 @@ monitor_back_db_init(
 
 	be->be_private = &monitor_info;
 
-	be2 = select_backend( &ndn, 0, 0 );
+	be2 = select_backend( &ndn, 0 );
 	if ( be2 != be ) {
 		char	*type = be2->bd_info->bi_type;
 
@@ -2014,19 +2166,57 @@ monitor_back_db_init(
 			type = oi->oi_orig->bi_type;
 		}
 
-		Debug( LDAP_DEBUG_ANY,
-			"\"monitor\" database serving namingContext \"%s\" "
-			"is hidden by \"%s\" database serving namingContext \"%s\".\n",
-			pdn.bv_val, type, be2->be_nsuffix[ 0 ].bv_val );
+		if (c) {
+			snprintf(c->msg, sizeof(c->msg),
+					"\"monitor\" database serving namingContext \"%s\" "
+					"is hidden by \"%s\" database serving namingContext \"%s\".\n",
+					pdn.bv_val, type, be2->be_nsuffix[ 0 ].bv_val );
+		}
 		return -1;
 	}
 
 	return 0;
 }
 
+static void
+monitor_back_destroy_limbo_entry(
+	entry_limbo_t	*el,
+	int		dispose )
+{
+	if ( el->el_e ) {
+		entry_free( el->el_e );
+	}
+	if ( el->el_a ) {
+		attrs_free( el->el_a );
+	}
+	if ( !BER_BVISNULL( &el->el_nbase ) ) {
+		ber_memfree( el->el_nbase.bv_val );
+	}
+	if ( !BER_BVISNULL( &el->el_filter ) ) {
+		ber_memfree( el->el_filter.bv_val );
+	}
+
+	/* NOTE: callbacks are not copied; so only free them
+	 * if disposing of */
+	if ( el->el_cb && dispose != 0 ) {
+		monitor_callback_t *next;
+
+		for ( ; el->el_cb; el->el_cb = next ) {
+			next = el->el_cb->mc_next;
+			if ( el->el_cb->mc_dispose ) {
+				el->el_cb->mc_dispose( &el->el_cb->mc_private );
+			}
+			ch_free( el->el_cb );
+		}
+	}
+
+	ch_free( el );
+}
+
 int
 monitor_back_db_open(
-	BackendDB	*be )
+	BackendDB	*be,
+	ConfigReply	*cr)
 {
 	monitor_info_t 		*mi = (monitor_info_t *)be->be_private;
 	struct monitor_subsys_t	**ms;
@@ -2034,10 +2224,7 @@ monitor_back_db_open(
 	monitor_entry_t		*mp;
 	int			i;
 	struct berval		bv, rdn = BER_BVC(SLAPD_MONITOR_DN);
-	struct tm		*tms;
-#ifdef HAVE_GMTIME_R
-	struct tm		tm_buf;
-#endif
+	struct tm		tms;
 	static char		tmbuf[ LDAP_LUTIL_GENTIME_BUFSIZE ];
 	struct berval	desc[] = {
 		BER_BVC("This subtree contains monitoring/managing objects."),
@@ -2046,6 +2233,8 @@ monitor_back_db_open(
 		" attributes, which must be explicitly requested."),
 		BER_BVNULL };
 
+	int			retcode = 0;
+
 	assert( be_monitor != NULL );
 	if ( be != be_monitor ) {
 		be_monitor = be;
@@ -2054,27 +2243,8 @@ monitor_back_db_open(
 	/*
 	 * Start
 	 */
-#ifndef HAVE_GMTIME_R
-	ldap_pvt_thread_mutex_lock( &gmtime_mutex );
-#endif
-#ifdef HACK_LOCAL_TIME
-# ifdef HAVE_LOCALTIME_R
-	tms = localtime_r( &starttime, &tm_buf );
-# else
-	tms = localtime( &starttime );
-# endif /* HAVE_LOCALTIME_R */
-	lutil_localtime( tmbuf, sizeof(tmbuf), tms, -timezone );
-#else /* !HACK_LOCAL_TIME */
-# ifdef HAVE_GMTIME_R
-	tms = gmtime_r( &starttime, &tm_buf );
-# else
-	tms = gmtime( &starttime );
-# endif /* HAVE_GMTIME_R */
-	lutil_gentime( tmbuf, sizeof(tmbuf), tms );
-#endif /* !HACK_LOCAL_TIME */
-#ifndef HAVE_GMTIME_R
-	ldap_pvt_thread_mutex_unlock( &gmtime_mutex );
-#endif
+	ldap_pvt_gmtime( &starttime, &tms );
+	lutil_gentime( tmbuf, sizeof(tmbuf), &tms );
 
 	mi->mi_startTime.bv_val = tmbuf;
 	mi->mi_startTime.bv_len = strlen( tmbuf );
@@ -2090,13 +2260,13 @@ monitor_back_db_open(
 	/*
 	 * creates the "cn=Monitor" entry 
 	 */
-	e = monitor_entry_stub( NULL, NULL, &rdn, mi->mi_oc_monitorServer, mi,
+	e = monitor_entry_stub( NULL, NULL, &rdn, mi->mi_oc_monitorServer,
 		NULL, NULL );
 
 	if ( e == NULL) {
 		Debug( LDAP_DEBUG_ANY,
 			"unable to create \"%s\" entry\n",
-			SLAPD_MONITOR_DN, 0, 0 );
+			SLAPD_MONITOR_DN );
 		return( -1 );
 	}
 
@@ -2129,7 +2299,7 @@ monitor_back_db_open(
 					&bv, NULL ) ) {
 			Debug( LDAP_DEBUG_ANY,
 				"unable to add monitoredInfo to \"%s\" entry\n",
-				SLAPD_MONITOR_DN, 0, 0 );
+				SLAPD_MONITOR_DN );
 			return( -1 );
 		}
 	}
@@ -2144,7 +2314,7 @@ monitor_back_db_open(
 	if ( monitor_cache_add( mi, e ) ) {
 		Debug( LDAP_DEBUG_ANY,
 			"unable to add entry \"%s\" to cache\n",
-			SLAPD_MONITOR_DN, 0, 0 );
+			SLAPD_MONITOR_DN );
 		return -1;
 	}
 	root = e;
@@ -2166,18 +2336,18 @@ monitor_back_db_open(
 		if ( rc != LDAP_SUCCESS ) {
 			Debug( LDAP_DEBUG_ANY,
 				"monitor RDN \"%s\" is invalid\n", 
-				dn.bv_val, 0, 0 );
+				dn.bv_val );
 			return( -1 );
 		}
 
 		e = monitor_entry_stub( &root->e_name, &root->e_nname,
-			&monitor_subsys[ i ]->mss_rdn, mi->mi_oc_monitorContainer, mi,
+			&monitor_subsys[ i ]->mss_rdn, mi->mi_oc_monitorContainer,
 			NULL, NULL );
 
 		if ( e == NULL) {
 			Debug( LDAP_DEBUG_ANY,
 				"unable to create \"%s\" entry\n", 
-				monitor_subsys[ i ]->mss_dn.bv_val, 0, 0 );
+				monitor_subsys[ i ]->mss_dn.bv_val );
 			return( -1 );
 		}
 		monitor_subsys[i]->mss_dn = e->e_name;
@@ -2199,7 +2369,7 @@ monitor_back_db_open(
 		if ( monitor_cache_add( mi, e ) ) {
 			Debug( LDAP_DEBUG_ANY,
 				"unable to add entry \"%s\" to cache\n",
-				monitor_subsys[ i ]->mss_dn.bv_val, 0, 0 );
+				monitor_subsys[ i ]->mss_dn.bv_val );
 			return -1;
 		}
 
@@ -2215,8 +2385,7 @@ monitor_back_db_open(
 	 * opens the monitor backend subsystems
 	 */
 	for ( ms = monitor_subsys; ms[ 0 ] != NULL; ms++ ) {
-		if ( ms[ 0 ]->mss_open && ( *ms[ 0 ]->mss_open )( be, ms[ 0 ] ) )
-		{
+		if ( ms[ 0 ]->mss_open && ms[ 0 ]->mss_open( be, ms[ 0 ] ) ) {
 			return( -1 );
 		}
 		ms[ 0 ]->mss_flags |= MONITOR_F_OPENED;
@@ -2225,7 +2394,7 @@ monitor_back_db_open(
 	monitor_subsys_opened = 1;
 
 	if ( mi->mi_entry_limbo ) {
-		entry_limbo_t	*el = (entry_limbo_t *)mi->mi_entry_limbo;
+		entry_limbo_t	*el = mi->mi_entry_limbo;
 
 		for ( ; el; ) {
 			entry_limbo_t	*tmp;
@@ -2254,7 +2423,7 @@ monitor_back_db_open(
 
 			case LIMBO_ATTRS:
 				rc = monitor_back_register_entry_attrs(
-						&el->el_ndn,
+						el->el_ndn,
 						el->el_a,
 						el->el_cb,
 						&el->el_nbase,
@@ -2264,48 +2433,51 @@ monitor_back_db_open(
 
 			case LIMBO_CB:
 				rc = monitor_back_register_entry_callback(
-						&el->el_ndn,
+						el->el_ndn,
 						el->el_cb,
 						&el->el_nbase,
 						el->el_scope,
 						&el->el_filter );
 				break;
 
+			case LIMBO_BACKEND:
+				rc = monitor_back_register_backend( el->el_bi );
+				break;
+
+			case LIMBO_DATABASE:
+				rc = monitor_back_register_database( el->el_be, el->el_ndn );
+				break;
+
+			case LIMBO_OVERLAY_INFO:
+				rc = monitor_back_register_overlay_info( el->el_on );
+				break;
+
+			case LIMBO_OVERLAY:
+				rc = monitor_back_register_overlay( el->el_be, el->el_on, el->el_ndn );
+				break;
+
+			case LIMBO_SUBSYS:
+				rc = monitor_back_register_subsys( el->el_mss );
+				break;
+
 			default:
 				assert( 0 );
 			}
 
-			if ( el->el_e ) {
-				entry_free( el->el_e );
-			}
-			if ( el->el_a ) {
-				attrs_free( el->el_a );
-			}
-			if ( !BER_BVISNULL( &el->el_ndn ) ) {
-				ber_memfree( el->el_ndn.bv_val );
-			}
-			if ( !BER_BVISNULL( &el->el_nbase ) ) {
-				ber_memfree( el->el_nbase.bv_val );
-			}
-			if ( !BER_BVISNULL( &el->el_filter ) ) {
-				ber_memfree( el->el_filter.bv_val );
-			}
-			if ( el->el_cb && rc != 0 ) {
-				if ( el->el_cb->mc_dispose ) {
-					el->el_cb->mc_dispose( &el->el_cb->mc_private );
-				}
-				ch_free( el->el_cb );
-			}
-
 			tmp = el;
 			el = el->el_next;
-			ch_free( tmp );
+			monitor_back_destroy_limbo_entry( tmp, rc );
+
+			if ( rc != 0 ) {
+				/* try all, but report error at end */
+				retcode = 1;
+			}
 		}
 
 		mi->mi_entry_limbo = NULL;
 	}
 
-	return( 0 );
+	return retcode;
 }
 
 int
@@ -2342,7 +2514,8 @@ monitor_back_db_config(
 
 int
 monitor_back_db_destroy(
-	BackendDB	*be )
+	BackendDB	*be,
+	ConfigReply	*cr)
 {
 	monitor_info_t	*mi = ( monitor_info_t * )be->be_private;
 
@@ -2372,6 +2545,16 @@ monitor_back_db_destroy(
 
 		ch_free( monitor_subsys );
 	}
+
+	if ( mi->mi_entry_limbo ) {
+		entry_limbo_t	*el = mi->mi_entry_limbo;
+
+		for ( ; el; ) {
+			entry_limbo_t *tmp = el;
+			el = el->el_next;
+			monitor_back_destroy_limbo_entry( tmp, 1 );
+		}
+	}
 	
 	ldap_pvt_thread_mutex_destroy( &monitor_info.mi_cache_mutex );
 
@@ -2379,11 +2562,3 @@ monitor_back_db_destroy(
 
 	return 0;
 }
-
-#if SLAPD_MONITOR == SLAPD_MOD_DYNAMIC
-
-/* conditionally define the init_module() function */
-SLAP_BACKEND_INIT_MODULE( monitor )
-
-#endif /* SLAPD_MONITOR == SLAPD_MOD_DYNAMIC */
-
