@@ -1,8 +1,8 @@
 /* thread.c - deal with thread subsystem */
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-monitor/thread.c,v 1.38.2.4 2007/02/13 20:22:26 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2001-2007 The OpenLDAP Foundation.
+ * Copyright 2001-2020 The OpenLDAP Foundation.
  * Portions Copyright 2001-2003 Pierangelo Masarati.
  * All rights reserved.
  *
@@ -29,7 +29,6 @@
 
 #include <ldap_rq.h>
 
-#ifndef NO_THREADS
 typedef enum {
 	MT_UNKNOWN,
 	MT_RUNQUEUE,
@@ -96,7 +95,6 @@ monitor_subsys_thread_update(
 	Operation		*op,
 	SlapReply		*rs,
 	Entry 			*e );
-#endif /* ! NO_THREADS */
 
 /*
  * initializes log subentry
@@ -106,7 +104,6 @@ monitor_subsys_thread_init(
 	BackendDB       	*be,
 	monitor_subsys_t	*ms )
 {
-#ifndef NO_THREADS
 	monitor_info_t	*mi;
 	monitor_entry_t	*mp;
 	Entry		*e, **ep, *e_thread;
@@ -119,8 +116,7 @@ monitor_subsys_thread_init(
 	if ( monitor_cache_get( mi, &ms->mss_ndn, &e_thread ) ) {
 		Debug( LDAP_DEBUG_ANY,
 			"monitor_subsys_thread_init: unable to get entry \"%s\"\n",
-			ms->mss_dn.bv_val, 
-			0, 0 );
+			ms->mss_dn.bv_val );
 		return( -1 );
 	}
 
@@ -139,13 +135,13 @@ monitor_subsys_thread_init(
 		 */
 		e = monitor_entry_stub( &ms->mss_dn, &ms->mss_ndn,
 			&mt[ i ].rdn,
-			mi->mi_oc_monitoredObject, mi, NULL, NULL );
+			mi->mi_oc_monitoredObject, NULL, NULL );
 		if ( e == NULL ) {
 			Debug( LDAP_DEBUG_ANY,
 				"monitor_subsys_thread_init: "
 				"unable to create entry \"%s,%s\"\n",
 				mt[ i ].rdn.bv_val,
-				ms->mss_ndn.bv_val, 0 );
+				ms->mss_ndn.bv_val );
 			return( -1 );
 		}
 
@@ -201,7 +197,7 @@ monitor_subsys_thread_init(
 				"monitor_subsys_thread_init: "
 				"unable to add entry \"%s,%s\"\n",
 				mt[ i ].rdn.bv_val,
-				ms->mss_dn.bv_val, 0 );
+				ms->mss_dn.bv_val );
 			return( -1 );
 		}
 	
@@ -211,11 +207,9 @@ monitor_subsys_thread_init(
 
 	monitor_cache_release( mi, e_thread );
 
-#endif /* ! NO_THREADS */
 	return( 0 );
 }
 
-#ifndef NO_THREADS
 static int 
 monitor_subsys_thread_update( 
 	Operation		*op,
@@ -260,6 +254,7 @@ monitor_subsys_thread_update(
 				ber_bvarray_free( a->a_vals );
 				a->a_vals = NULL;
 				a->a_nvals = NULL;
+				a->a_numvals = 0;
 			}
 
 			i = 0;
@@ -292,6 +287,7 @@ monitor_subsys_thread_update(
 				ber_bvarray_free( a->a_vals );
 				a->a_vals = NULL;
 				a->a_nvals = NULL;
+				a->a_numvals = 0;
 			}
 	
 			i = 0;
@@ -353,4 +349,3 @@ monitor_subsys_thread_update(
 
 	return SLAP_CB_CONTINUE;
 }
-#endif /* ! NO_THREADS */
