@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-sql/add.c,v 1.50.2.3 2007/02/13 20:22:28 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2007 The OpenLDAP Foundation.
+ * Copyright 1999-2020 The OpenLDAP Foundation.
  * Portions Copyright 1999 Dmitry Kovalev.
  * Portions Copyright 2002 Pierangelo Masarati.
  * Portions Copyright 2004 Mark Adamson.
@@ -72,7 +72,7 @@ backsql_modify_delete_all_values(
 			"   backsql_modify_delete_all_values(): "
 			"missing attribute value delete procedure "
 			"for attr \"%s\"\n",
-			at->bam_ad->ad_cname.bv_val, 0, 0 );
+			at->bam_ad->ad_cname.bv_val );
 		if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
 			rs->sr_text = "SQL-backend error";
 			return rs->sr_err = LDAP_OTHER;
@@ -87,7 +87,7 @@ backsql_modify_delete_all_values(
 			"   backsql_modify_delete_all_values(): "
 			"error preparing attribute value select query "
 			"\"%s\"\n",
-			at->bam_query, 0, 0 );
+			at->bam_query );
 		backsql_PrintErrors( bi->sql_db_env, dbh, 
 				asth, rc );
 
@@ -100,8 +100,7 @@ backsql_modify_delete_all_values(
 		Debug( LDAP_DEBUG_TRACE,
 			"   backsql_modify_delete_all_values(): "
 			"error binding key value parameter "
-			"to attribute value select query\n",
-			0, 0, 0 );
+			"to attribute value select query\n" );
 		backsql_PrintErrors( bi->sql_db_env, dbh, 
 				asth, rc );
 		SQLFreeStmt( asth, SQL_DROP );
@@ -114,8 +113,7 @@ backsql_modify_delete_all_values(
 	if ( !BACKSQL_SUCCESS( rc ) ) {
 		Debug( LDAP_DEBUG_TRACE,
 			"   backsql_modify_delete_all_values(): "
-			"error executing attribute value select query\n",
-			0, 0, 0 );
+			"error executing attribute value select query\n" );
 		backsql_PrintErrors( bi->sql_db_env, dbh, 
 				asth, rc );
 		SQLFreeStmt( asth, SQL_DROP );
@@ -147,7 +145,7 @@ backsql_modify_delete_all_values(
 					"error preparing attribute value "
 					"delete procedure "
 					"\"%s\"\n",
-					at->bam_delete_proc, 0, 0 );
+					at->bam_delete_proc );
 				backsql_PrintErrors( bi->sql_db_env, dbh, 
 						sth, rc );
 
@@ -164,7 +162,7 @@ backsql_modify_delete_all_values(
 					Debug( LDAP_DEBUG_TRACE,
 						"   backsql_modify_delete_all_values(): "
 						"error binding output parameter for %s[%d]\n",
-						at->bam_ad->ad_cname.bv_val, i, 0 );
+						at->bam_ad->ad_cname.bv_val, i );
 					backsql_PrintErrors( bi->sql_db_env, dbh, 
 						sth, rc );
 					SQLFreeStmt( sth, SQL_DROP );
@@ -181,7 +179,7 @@ backsql_modify_delete_all_values(
 				Debug( LDAP_DEBUG_TRACE,
 					"   backsql_modify_delete_all_values(): "
 					"error binding keyval parameter for %s[%d]\n",
-					at->bam_ad->ad_cname.bv_val, i, 0 );
+					at->bam_ad->ad_cname.bv_val, i );
 				backsql_PrintErrors( bi->sql_db_env, dbh, 
 					sth, rc );
 				SQLFreeStmt( sth, SQL_DROP );
@@ -190,17 +188,12 @@ backsql_modify_delete_all_values(
 				rs->sr_err = LDAP_OTHER;
 				goto done;
 			}
-#ifdef BACKSQL_ARBITRARY_KEY
+
 			Debug( LDAP_DEBUG_TRACE,
 				"   backsql_modify_delete_all_values() "
-				"arg(%d)=%s\n",
-				pno + 1 + po, e_id->eid_keyval.bv_val, 0 );
-#else /* ! BACKSQL_ARBITRARY_KEY */
-			Debug( LDAP_DEBUG_TRACE,
-				"   backsql_modify_delete_all_values() "
-				"arg(%d)=%lu\n",
-				pno + 1 + po, e_id->eid_keyval, 0 );
-#endif /* ! BACKSQL_ARBITRARY_KEY */
+				"arg(%d)=" BACKSQL_IDFMT "\n",
+				pno + 1 + po,
+				BACKSQL_IDARG(e_id->eid_keyval) );
 
 			/*
 			 * check for syntax needed here 
@@ -213,7 +206,7 @@ backsql_modify_delete_all_values(
 				Debug( LDAP_DEBUG_TRACE,
 					"   backsql_modify_delete_all_values(): "
 					"error binding value parameter for %s[%d]\n",
-					at->bam_ad->ad_cname.bv_val, i, 0 );
+					at->bam_ad->ad_cname.bv_val, i );
 				backsql_PrintErrors( bi->sql_db_env, dbh, 
 					sth, rc );
 				SQLFreeStmt( sth, SQL_DROP );
@@ -237,7 +230,7 @@ backsql_modify_delete_all_values(
 					"   backsql_modify_delete_all_values(): "
 					"delete_proc "
 					"execution failed (rc=%d, prc=%d)\n",
-					rc, prc, 0 );
+					rc, prc );
 				if ( prc != LDAP_SUCCESS ) {
 					/* SQL procedure executed fine 
 					 * but returned an error */
@@ -279,7 +272,7 @@ backsql_modify_internal(
 	Modifications	*ml;
 
 	Debug( LDAP_DEBUG_TRACE, "==>backsql_modify_internal(): "
-		"traversing modifications list\n", 0, 0, 0 );
+		"traversing modifications list\n" );
 
 	for ( ml = modlist; ml != NULL; ml = ml->sml_next ) {
 		AttributeDescription	*ad;
@@ -317,7 +310,7 @@ backsql_modify_internal(
 			Debug( LDAP_DEBUG_TRACE, "   backsql_modify_internal(): "
 				"attribute \"%s\" is not registered "
 				"in objectClass \"%s\"\n",
-				ad->ad_cname.bv_val, BACKSQL_OC_NAME( oc ), 0 );
+				ad->ad_cname.bv_val, BACKSQL_OC_NAME( oc ) );
 
 			if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
 				rs->sr_err = LDAP_UNWILLING_TO_PERFORM;
@@ -333,7 +326,7 @@ backsql_modify_internal(
 		case LDAP_MOD_REPLACE: {
 			Debug( LDAP_DEBUG_TRACE, "   backsql_modify_internal(): "
 				"replacing values for attribute \"%s\"\n",
-				at->bam_ad->ad_cname.bv_val, 0, 0 );
+				at->bam_ad->ad_cname.bv_val );
 
 			if ( at->bam_add_proc == NULL ) {
 				Debug( LDAP_DEBUG_TRACE,
@@ -341,7 +334,7 @@ backsql_modify_internal(
 					"add procedure is not defined "
 					"for attribute \"%s\" "
 					"- unable to perform replacements\n",
-					at->bam_ad->ad_cname.bv_val, 0, 0 );
+					at->bam_ad->ad_cname.bv_val );
 
 				if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
 					rs->sr_err = LDAP_UNWILLING_TO_PERFORM;
@@ -359,7 +352,7 @@ backsql_modify_internal(
 						"   backsql_modify_internal(): "
 						"delete procedure is not defined "
 						"for attribute \"%s\"\n",
-						at->bam_ad->ad_cname.bv_val, 0, 0 );
+						at->bam_ad->ad_cname.bv_val );
 
 					rs->sr_err = LDAP_UNWILLING_TO_PERFORM;
 					rs->sr_text = "operation not permitted "
@@ -372,7 +365,7 @@ backsql_modify_internal(
 					"delete procedure is not defined "
 					"for attribute \"%s\" "
 					"- adding only\n",
-					at->bam_ad->ad_cname.bv_val, 0, 0 );
+					at->bam_ad->ad_cname.bv_val );
 
 				goto add_only;
 			}
@@ -394,13 +387,14 @@ del_all:
 		 */
 		case LDAP_MOD_ADD:
 		/* case SLAP_MOD_SOFTADD: */
+		/* case SLAP_MOD_ADD_IF_NOT_PRESENT: */
 add_only:;
 			if ( at->bam_add_proc == NULL ) {
 				Debug( LDAP_DEBUG_TRACE,
 					"   backsql_modify_internal(): "
 					"add procedure is not defined "
 					"for attribute \"%s\"\n",
-					at->bam_ad->ad_cname.bv_val, 0, 0 );
+					at->bam_ad->ad_cname.bv_val );
 
 				if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
 					rs->sr_err = LDAP_UNWILLING_TO_PERFORM;
@@ -414,7 +408,7 @@ add_only:;
 			
 			Debug( LDAP_DEBUG_TRACE, "   backsql_modify_internal(): "
 				"adding new values for attribute \"%s\"\n",
-				at->bam_ad->ad_cname.bv_val, 0, 0 );
+				at->bam_ad->ad_cname.bv_val );
 
 			/* can't add a NULL val array */
 			assert( sm_values != NULL );
@@ -434,8 +428,7 @@ add_only:;
 				if ( rc != SQL_SUCCESS ) {
 					Debug( LDAP_DEBUG_TRACE,
 						"   backsql_modify_internal(): "
-						"error preparing add query\n", 
-						0, 0, 0 );
+						"error preparing add query\n" );
 					backsql_PrintErrors( bi->sql_db_env, dbh, sth, rc );
 
 					rs->sr_err = LDAP_OTHER;
@@ -451,7 +444,7 @@ add_only:;
 						Debug( LDAP_DEBUG_TRACE,
 							"   backsql_modify_internal(): "
 							"error binding output parameter for %s[%d]\n",
-							at->bam_ad->ad_cname.bv_val, i, 0 );
+							at->bam_ad->ad_cname.bv_val, i );
 						backsql_PrintErrors( bi->sql_db_env, dbh, 
 							sth, rc );
 						SQLFreeStmt( sth, SQL_DROP );
@@ -468,7 +461,7 @@ add_only:;
 					Debug( LDAP_DEBUG_TRACE,
 						"   backsql_modify_internal(): "
 						"error binding keyval parameter for %s[%d]\n",
-						at->bam_ad->ad_cname.bv_val, i, 0 );
+						at->bam_ad->ad_cname.bv_val, i );
 					backsql_PrintErrors( bi->sql_db_env, dbh, 
 						sth, rc );
 					SQLFreeStmt( sth, SQL_DROP );
@@ -477,17 +470,12 @@ add_only:;
 					rs->sr_err = LDAP_OTHER;
 					goto done;
 				}
-#ifdef BACKSQL_ARBITRARY_KEY
+
 				Debug( LDAP_DEBUG_TRACE,
 					"   backsql_modify_internal(): "
-					"arg(%d)=\"%s\"\n", 
-					pno + 1 + po, e_id->eid_keyval.bv_val, 0 );
-#else /* ! BACKSQL_ARBITRARY_KEY */
-				Debug( LDAP_DEBUG_TRACE,
-					"   backsql_modify_internal(): "
-					"arg(%d)=\"%lu\"\n", 
-					pno + 1 + po, e_id->eid_keyval, 0 );
-#endif /* ! BACKSQL_ARBITRARY_KEY */
+					"arg(%d)=" BACKSQL_IDFMT "\n", 
+					pno + 1 + po,
+					BACKSQL_IDARG(e_id->eid_keyval) );
 
 				/*
 				 * check for syntax needed here
@@ -499,7 +487,7 @@ add_only:;
 					Debug( LDAP_DEBUG_TRACE,
 						"   backsql_modify_internal(): "
 						"error binding value parameter for %s[%d]\n",
-						at->bam_ad->ad_cname.bv_val, i, 0 );
+						at->bam_ad->ad_cname.bv_val, i );
 					backsql_PrintErrors( bi->sql_db_env, dbh, 
 						sth, rc );
 					SQLFreeStmt( sth, SQL_DROP );
@@ -523,7 +511,7 @@ add_only:;
 						"   backsql_modify_internal(): "
 						"add_proc execution failed "
 						"(rc=%d, prc=%d)\n",
-						rc, prc, 0 );
+						rc, prc );
 					if ( prc != LDAP_SUCCESS ) {
 						/* SQL procedure executed fine 
 						 * but returned an error */
@@ -551,12 +539,13 @@ add_only:;
 			break;
 			
 	      	case LDAP_MOD_DELETE:
+		/* case SLAP_MOD_SOFTDEL: */
 			if ( at->bam_delete_proc == NULL ) {
 				Debug( LDAP_DEBUG_TRACE,
 					"   backsql_modify_internal(): "
 					"delete procedure is not defined "
 					"for attribute \"%s\"\n",
-					at->bam_ad->ad_cname.bv_val, 0, 0 );
+					at->bam_ad->ad_cname.bv_val );
 
 				if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
 					rs->sr_err = LDAP_UNWILLING_TO_PERFORM;
@@ -574,13 +563,13 @@ add_only:;
 					"no values given to delete "
 					"for attribute \"%s\" "
 					"-- deleting all values\n",
-					at->bam_ad->ad_cname.bv_val, 0, 0 );
+					at->bam_ad->ad_cname.bv_val );
 				goto del_all;
 			}
 
 			Debug( LDAP_DEBUG_TRACE, "   backsql_modify_internal(): "
 				"deleting values for attribute \"%s\"\n",
-				at->bam_ad->ad_cname.bv_val, 0, 0 );
+				at->bam_ad->ad_cname.bv_val );
 
 			for ( i = 0, at_val = sm_values;
 					!BER_BVISNULL( at_val );
@@ -597,8 +586,7 @@ add_only:;
 				if ( rc != SQL_SUCCESS ) {
 					Debug( LDAP_DEBUG_TRACE,
 						"   backsql_modify_internal(): "
-						"error preparing delete query\n", 
-						0, 0, 0 );
+						"error preparing delete query\n" );
 					backsql_PrintErrors( bi->sql_db_env, dbh, sth, rc );
 
 					rs->sr_err = LDAP_OTHER;
@@ -614,7 +602,7 @@ add_only:;
 						Debug( LDAP_DEBUG_TRACE,
 							"   backsql_modify_internal(): "
 							"error binding output parameter for %s[%d]\n",
-							at->bam_ad->ad_cname.bv_val, i, 0 );
+							at->bam_ad->ad_cname.bv_val, i );
 						backsql_PrintErrors( bi->sql_db_env, dbh, 
 							sth, rc );
 						SQLFreeStmt( sth, SQL_DROP );
@@ -631,7 +619,7 @@ add_only:;
 					Debug( LDAP_DEBUG_TRACE,
 						"   backsql_modify_internal(): "
 						"error binding keyval parameter for %s[%d]\n",
-						at->bam_ad->ad_cname.bv_val, i, 0 );
+						at->bam_ad->ad_cname.bv_val, i );
 					backsql_PrintErrors( bi->sql_db_env, dbh, 
 						sth, rc );
 					SQLFreeStmt( sth, SQL_DROP );
@@ -640,17 +628,12 @@ add_only:;
 					rs->sr_err = LDAP_OTHER;
 					goto done;
 				}
-#ifdef BACKSQL_ARBITRARY_KEY
+
 				Debug( LDAP_DEBUG_TRACE,
 					"   backsql_modify_internal(): "
-					"arg(%d)=\"%s\"\n", 
-					pno + 1 + po, e_id->eid_keyval.bv_val, 0 );
-#else /* ! BACKSQL_ARBITRARY_KEY */
-				Debug( LDAP_DEBUG_TRACE,
-					"   backsql_modify_internal(): "
-					"arg(%d)=\"%lu\"\n", 
-					pno + 1 + po, e_id->eid_keyval, 0 );
-#endif /* ! BACKSQL_ARBITRARY_KEY */
+					"arg(%d)=" BACKSQL_IDFMT "\n", 
+					pno + 1 + po,
+					BACKSQL_IDARG(e_id->eid_keyval) );
 
 				/*
 				 * check for syntax needed here 
@@ -662,7 +645,7 @@ add_only:;
 					Debug( LDAP_DEBUG_TRACE,
 						"   backsql_modify_internal(): "
 						"error binding value parameter for %s[%d]\n",
-						at->bam_ad->ad_cname.bv_val, i, 0 );
+						at->bam_ad->ad_cname.bv_val, i );
 					backsql_PrintErrors( bi->sql_db_env, dbh, 
 						sth, rc );
 					SQLFreeStmt( sth, SQL_DROP );
@@ -675,7 +658,7 @@ add_only:;
 				Debug( LDAP_DEBUG_TRACE,
 					"   backsql_modify_internal(): "
 					"executing \"%s\"\n", 
-					at->bam_delete_proc, 0, 0 );
+					at->bam_delete_proc );
 				rc = SQLExecute( sth );
 				if ( rc == SQL_SUCCESS && prc == LDAP_SUCCESS )
 				{
@@ -686,7 +669,7 @@ add_only:;
 						"   backsql_modify_internal(): "
 						"delete_proc execution "
 						"failed (rc=%d, prc=%d)\n",
-						rc, prc, 0 );
+						rc, prc );
 
 					if ( prc != LDAP_SUCCESS ) {
 						/* SQL procedure executed fine
@@ -710,7 +693,7 @@ add_only:;
 
 	      	case LDAP_MOD_INCREMENT:
 			Debug( LDAP_DEBUG_TRACE, "   backsql_modify_internal(): "
-				"increment not supported yet\n", 0, 0, 0 );
+				"increment not supported yet\n" );
 			if ( BACKSQL_FAIL_IF_NO_MAPPING( bi ) ) {
 				rs->sr_err = LDAP_OTHER;
 				rs->sr_text = "SQL-backend error";
@@ -739,7 +722,7 @@ backsql_add_attr(
 	SQLHDBC 		dbh,
 	backsql_oc_map_rec 	*oc,
 	Attribute		*at,
-	unsigned long		new_keyval )
+	backsql_key_t		new_keyval )
 {
 	backsql_info		*bi = (backsql_info*)op->o_bd->be_private;
 	backsql_at_map_rec	*at_rec = NULL;
@@ -794,7 +777,6 @@ backsql_add_attr(
 		int		prc = LDAP_SUCCESS;
 		/* first parameter #, parameter order */
 		SQLUSMALLINT	pno, po;
-		char		logbuf[] = "val[18446744073709551615UL], id=18446744073709551615UL";
 		
 		/*
 		 * Do not deal with the objectClass that is used
@@ -820,7 +802,7 @@ backsql_add_attr(
 				Debug( LDAP_DEBUG_TRACE,
 					"   backsql_add_attr(): "
 					"error binding output parameter for %s[%lu]\n",
-					at_rec->bam_ad->ad_cname.bv_val, i, 0 );
+					at_rec->bam_ad->ad_cname.bv_val, i );
 				backsql_PrintErrors( bi->sql_db_env, dbh, 
 					sth, rc );
 				SQLFreeStmt( sth, SQL_DROP );
@@ -835,13 +817,13 @@ backsql_add_attr(
 
 		po = ( BACKSQL_IS_ADD( at_rec->bam_param_order ) ) > 0;
 		currpos = pno + 1 + po;
-		rc = backsql_BindParamInt( sth, currpos,
+		rc = backsql_BindParamNumID( sth, currpos,
 				SQL_PARAM_INPUT, &new_keyval );
 		if ( rc != SQL_SUCCESS ) {
 			Debug( LDAP_DEBUG_TRACE,
 				"   backsql_add_attr(): "
 				"error binding keyval parameter for %s[%lu]\n",
-				at_rec->bam_ad->ad_cname.bv_val, i, 0 );
+				at_rec->bam_ad->ad_cname.bv_val, i );
 			backsql_PrintErrors( bi->sql_db_env, dbh, 
 				sth, rc );
 			SQLFreeStmt( sth, SQL_DROP );
@@ -862,7 +844,7 @@ backsql_add_attr(
 			Debug( LDAP_DEBUG_TRACE,
 				"   backsql_add_attr(): "
 				"error binding value parameter for %s[%lu]\n",
-				at_rec->bam_ad->ad_cname.bv_val, i, 0 );
+				at_rec->bam_ad->ad_cname.bv_val, i );
 			backsql_PrintErrors( bi->sql_db_env, dbh, 
 				sth, rc );
 			SQLFreeStmt( sth, SQL_DROP );
@@ -872,12 +854,10 @@ backsql_add_attr(
 		}
 
 #ifdef LDAP_DEBUG
-		snprintf( logbuf, sizeof( logbuf ), "val[%lu], id=%lu",
-				i, new_keyval );
-		Debug( LDAP_DEBUG_TRACE, "   backsql_add_attr(\"%s\"): "
-			"executing \"%s\" %s\n", 
-			op->ora_e->e_name.bv_val,
-			at_rec->bam_add_proc, logbuf );
+		Debug(LDAP_DEBUG_TRACE,
+		      "   backsql_add_attr(\"%s\"): " "executing \"%s\" val[%lu], id=" BACKSQL_IDNUMFMT "\n",
+		      op->ora_e->e_name.bv_val, at_rec->bam_add_proc,
+		      i, new_keyval );
 #endif
 		rc = SQLExecute( sth );
 		if ( rc == SQL_SUCCESS && prc == LDAP_SUCCESS ) {
@@ -917,7 +897,7 @@ backsql_add( Operation *op, SlapReply *rs )
 	backsql_info		*bi = (backsql_info*)op->o_bd->be_private;
 	SQLHDBC 		dbh = SQL_NULL_HDBC;
 	SQLHSTMT 		sth = SQL_NULL_HSTMT;
-	unsigned long		new_keyval = 0;
+	backsql_key_t		new_keyval = 0;
 	RETCODE			rc;
 	backsql_oc_map_rec 	*oc = NULL;
 	backsql_srch_info	bsi = { 0 };
@@ -939,7 +919,7 @@ backsql_add( Operation *op, SlapReply *rs )
 	 * NOTE: fake successful result to force contextCSN to be bumped up
 	 */
 	if ( op->o_sync ) {
-		char		buf[ LDAP_LUTIL_CSNSTR_BUFSIZE ];
+		char		buf[ LDAP_PVT_CSNSTR_BUFSIZE ];
 		struct berval	csn;
 
 		csn.bv_val = buf;
@@ -956,18 +936,18 @@ backsql_add( Operation *op, SlapReply *rs )
 #endif /* BACKSQL_SYNCPROV */
 
 	Debug( LDAP_DEBUG_TRACE, "==>backsql_add(\"%s\")\n",
-			op->ora_e->e_name.bv_val, 0, 0 );
+			op->ora_e->e_name.bv_val );
 
 	/* check schema */
 	if ( BACKSQL_CHECK_SCHEMA( bi ) ) {
 		char		textbuf[ SLAP_TEXT_BUFLEN ] = { '\0' };
 
-		rs->sr_err = entry_schema_check( op, op->ora_e, NULL, 0, 1,
+		rs->sr_err = entry_schema_check( op, op->ora_e, NULL, 0, 1, NULL,
 			&rs->sr_text, textbuf, sizeof( textbuf ) );
 		if ( rs->sr_err != LDAP_SUCCESS ) {
 			Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 				"entry failed schema check -- aborting\n",
-				op->ora_e->e_name.bv_val, 0, 0 );
+				op->ora_e->e_name.bv_val );
 			e = NULL;
 			goto done;
 		}
@@ -975,6 +955,17 @@ backsql_add( Operation *op, SlapReply *rs )
 
 	slap_add_opattrs( op, &rs->sr_text, textbuf, textlen, 1 );
 
+	if ( get_assert( op ) &&
+		( test_filter( op, op->ora_e, get_assertion( op )) != LDAP_COMPARE_TRUE ))
+	{
+		Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
+			"assertion control failed -- aborting\n",
+			op->ora_e->e_name.bv_val );
+		e = NULL;
+		rs->sr_err = LDAP_ASSERTION_FAILED;
+		goto done;
+	}
+
 	/* search structuralObjectClass */
 	for ( at = op->ora_e->e_attrs; at != NULL; at = at->a_next ) {
 		if ( at->a_desc == slap_schema.si_ad_structuralObjectClass ) {
@@ -997,7 +988,7 @@ backsql_add( Operation *op, SlapReply *rs )
 		if ( at == NULL ) {
 			Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 				"no objectClass\n",
-				op->ora_e->e_name.bv_val, 0, 0 );
+				op->ora_e->e_name.bv_val );
 			rs->sr_err = LDAP_OBJECT_CLASS_VIOLATION;
 			e = NULL;
 			goto done;
@@ -1025,7 +1016,7 @@ backsql_add( Operation *op, SlapReply *rs )
 		Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 			"cannot map structuralObjectClass \"%s\" -- aborting\n",
 			op->ora_e->e_name.bv_val,
-			scname.bv_val, 0 );
+			scname.bv_val );
 		rs->sr_err = LDAP_UNWILLING_TO_PERFORM;
 		rs->sr_text = "operation not permitted within namingContext";
 		e = NULL;
@@ -1037,7 +1028,7 @@ backsql_add( Operation *op, SlapReply *rs )
 			"create procedure is not defined "
 			"for structuralObjectClass \"%s\" - aborting\n",
 			op->ora_e->e_name.bv_val,
-			scname.bv_val, 0 );
+			scname.bv_val );
 		rs->sr_err = LDAP_UNWILLING_TO_PERFORM;
 		rs->sr_text = "operation not permitted within namingContext";
 		e = NULL;
@@ -1050,7 +1041,7 @@ backsql_add( Operation *op, SlapReply *rs )
 			"but none is defined for structuralObjectClass \"%s\" "
 			"- aborting\n",
 			op->ora_e->e_name.bv_val,
-			scname.bv_val, 0 );
+			scname.bv_val );
 		rs->sr_err = LDAP_UNWILLING_TO_PERFORM;
 		rs->sr_text = "operation not permitted within namingContext";
 		e = NULL;
@@ -1071,7 +1062,7 @@ backsql_add( Operation *op, SlapReply *rs )
 	if ( rs->sr_err != LDAP_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 			"could not get connection handle - exiting\n", 
-			op->ora_e->e_name.bv_val, 0, 0 );
+			op->ora_e->e_name.bv_val );
 		rs->sr_text = ( rs->sr_err == LDAP_OTHER )
 			?  "SQL-backend error" : NULL;
 		e = NULL;
@@ -1088,7 +1079,7 @@ backsql_add( Operation *op, SlapReply *rs )
 	if ( backsql_api_dn2odbc( op, rs, &realdn ) ) {
 		Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 			"backsql_api_dn2odbc(\"%s\") failed\n", 
-			op->ora_e->e_name.bv_val, realdn.bv_val, 0 );
+			op->ora_e->e_name.bv_val, realdn.bv_val );
 		rs->sr_err = LDAP_OTHER;
 		rs->sr_text = "SQL-backend error";
 		e = NULL;
@@ -1099,7 +1090,7 @@ backsql_add( Operation *op, SlapReply *rs )
 	if ( rs->sr_err == LDAP_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 			"entry exists\n",
-			op->ora_e->e_name.bv_val, 0, 0 );
+			op->ora_e->e_name.bv_val );
 		rs->sr_err = LDAP_ALREADY_EXISTS;
 		e = op->ora_e;
 		goto done;
@@ -1159,13 +1150,13 @@ backsql_add( Operation *op, SlapReply *rs )
 
 	colnum = 1;
 	if ( BACKSQL_IS_ADD( oc->bom_expect_return ) ) {
-		rc = backsql_BindParamInt( sth, 1, SQL_PARAM_OUTPUT, &new_keyval );
+		rc = backsql_BindParamNumID( sth, 1, SQL_PARAM_OUTPUT, &new_keyval );
 		if ( rc != SQL_SUCCESS ) {
 			Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 				"error binding keyval parameter "
 				"for objectClass %s\n",
 				op->ora_e->e_name.bv_val,
-				oc->bom_oc->soc_cname.bv_val, 0 );
+				oc->bom_oc->soc_cname.bv_val );
 			backsql_PrintErrors( bi->sql_db_env, dbh, 
 				sth, rc );
 			SQLFreeStmt( sth, SQL_DROP );
@@ -1186,26 +1177,25 @@ backsql_add( Operation *op, SlapReply *rs )
 					at->a_vals[0].bv_len );
 			Debug( LDAP_DEBUG_TRACE, "backsql_add(): "
 					"create_proc hint: param = '%s'\n",
-					at->a_vals[0].bv_val, 0, 0 );
+					at->a_vals[0].bv_val );
 
 		} else {
 			backsql_BindParamStr( sth, colnum, SQL_PARAM_INPUT,
 					"", 0 );
 			Debug( LDAP_DEBUG_TRACE, "backsql_add(): "
-					"create_proc hint (%s) not avalable\n",
-					oc->bom_create_hint->ad_cname.bv_val,
-					0, 0 );
+					"create_proc hint (%s) not available\n",
+					oc->bom_create_hint->ad_cname.bv_val );
 		}
 		colnum++;
 	}
 
 	Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): executing \"%s\"\n",
-		op->ora_e->e_name.bv_val, oc->bom_create_proc, 0 );
+		op->ora_e->e_name.bv_val, oc->bom_create_proc );
 	rc = SQLExecute( sth );
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 			"create_proc execution failed\n",
-			op->ora_e->e_name.bv_val, 0, 0 );
+			op->ora_e->e_name.bv_val );
 		backsql_PrintErrors( bi->sql_db_env, dbh, sth, rc);
 		SQLFreeStmt( sth, SQL_DROP );
 		rs->sr_err = LDAP_OTHER;
@@ -1219,7 +1209,7 @@ backsql_add( Operation *op, SlapReply *rs )
 
 	if ( !BACKSQL_IS_ADD( oc->bom_expect_return ) ) {
 		SWORD		ncols;
-		SQLINTEGER	value_len;
+		SQLLEN		value_len;
 
 		if ( BACKSQL_CREATE_NEEDS_SELECT( bi ) ) {
 			SQLFreeStmt( sth, SQL_DROP );
@@ -1249,7 +1239,7 @@ backsql_add( Operation *op, SlapReply *rs )
 		if ( rc != SQL_SUCCESS ) {
 			Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 				"create_proc result evaluation failed\n",
-				op->ora_e->e_name.bv_val, 0, 0 );
+				op->ora_e->e_name.bv_val );
 			backsql_PrintErrors( bi->sql_db_env, dbh, sth, rc);
 			SQLFreeStmt( sth, SQL_DROP );
 			rs->sr_err = LDAP_OTHER;
@@ -1260,7 +1250,7 @@ backsql_add( Operation *op, SlapReply *rs )
 		} else if ( ncols != 1 ) {
 			Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 				"create_proc result is bogus (ncols=%d)\n",
-				op->ora_e->e_name.bv_val, ncols, 0 );
+				op->ora_e->e_name.bv_val, ncols );
 			backsql_PrintErrors( bi->sql_db_env, dbh, sth, rc);
 			SQLFreeStmt( sth, SQL_DROP );
 			rs->sr_err = LDAP_OTHER;
@@ -1297,7 +1287,7 @@ backsql_add( Operation *op, SlapReply *rs )
 		if ( value_len <= 0 ) {
 			Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 				"create_proc result is empty?\n",
-				op->ora_e->e_name.bv_val, 0, 0 );
+				op->ora_e->e_name.bv_val );
 			backsql_PrintErrors( bi->sql_db_env, dbh, sth, rc);
 			SQLFreeStmt( sth, SQL_DROP );
 			rs->sr_err = LDAP_OTHER;
@@ -1310,8 +1300,8 @@ backsql_add( Operation *op, SlapReply *rs )
 	SQLFreeStmt( sth, SQL_DROP );
 
 	Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
-		"create_proc returned keyval=%ld\n",
-		op->ora_e->e_name.bv_val, new_keyval, 0 );
+		"create_proc returned keyval=" BACKSQL_IDNUMFMT "\n",
+		op->ora_e->e_name.bv_val, new_keyval );
 
 	rc = backsql_Prepare( dbh, &sth, bi->sql_insentry_stmt, 0 );
 	if ( rc != SQL_SUCCESS ) {
@@ -1326,7 +1316,7 @@ backsql_add( Operation *op, SlapReply *rs )
 		Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 			"error binding DN parameter for objectClass %s\n",
 			op->ora_e->e_name.bv_val,
-			oc->bom_oc->soc_cname.bv_val, 0 );
+			oc->bom_oc->soc_cname.bv_val );
 		backsql_PrintErrors( bi->sql_db_env, dbh, 
 			sth, rc );
 		SQLFreeStmt( sth, SQL_DROP );
@@ -1337,13 +1327,13 @@ backsql_add( Operation *op, SlapReply *rs )
 		goto done;
 	}
 
-	rc = backsql_BindParamInt( sth, 2, SQL_PARAM_INPUT, &oc->bom_id );
+	rc = backsql_BindParamNumID( sth, 2, SQL_PARAM_INPUT, &oc->bom_id );
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 			"error binding objectClass ID parameter "
 			"for objectClass %s\n",
 			op->ora_e->e_name.bv_val,
-			oc->bom_oc->soc_cname.bv_val, 0 );
+			oc->bom_oc->soc_cname.bv_val );
 		backsql_PrintErrors( bi->sql_db_env, dbh, 
 			sth, rc );
 		SQLFreeStmt( sth, SQL_DROP );
@@ -1360,7 +1350,7 @@ backsql_add( Operation *op, SlapReply *rs )
 			"error binding parent ID parameter "
 			"for objectClass %s\n",
 			op->ora_e->e_name.bv_val,
-			oc->bom_oc->soc_cname.bv_val, 0 );
+			oc->bom_oc->soc_cname.bv_val );
 		backsql_PrintErrors( bi->sql_db_env, dbh, 
 			sth, rc );
 		SQLFreeStmt( sth, SQL_DROP );
@@ -1371,13 +1361,13 @@ backsql_add( Operation *op, SlapReply *rs )
 		goto done;
 	}
 
-	rc = backsql_BindParamInt( sth, 4, SQL_PARAM_INPUT, &new_keyval );
+	rc = backsql_BindParamNumID( sth, 4, SQL_PARAM_INPUT, &new_keyval );
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 			"error binding entry ID parameter "
 			"for objectClass %s\n",
 			op->ora_e->e_name.bv_val,
-			oc->bom_oc->soc_cname.bv_val, 0 );
+			oc->bom_oc->soc_cname.bv_val );
 		backsql_PrintErrors( bi->sql_db_env, dbh, 
 			sth, rc );
 		SQLFreeStmt( sth, SQL_DROP );
@@ -1388,22 +1378,17 @@ backsql_add( Operation *op, SlapReply *rs )
 		goto done;
 	}
 
-	Debug( LDAP_DEBUG_TRACE, "   backsql_add(): executing \"%s\" for dn \"%s\"\n",
-			bi->sql_insentry_stmt, op->ora_e->e_name.bv_val, 0 );
-#ifdef BACKSQL_ARBITRARY_KEY
-	Debug( LDAP_DEBUG_TRACE, "                  for oc_map_id=%ld, "
-			"p_id=%s, keyval=%ld\n",
-			oc->bom_id, bsi.bsi_base_id.eid_id.bv_val, new_keyval );
-#else /* ! BACKSQL_ARBITRARY_KEY */
-	Debug( LDAP_DEBUG_TRACE, "                  for oc_map_id=%ld, "
-			"p_id=%ld, keyval=%ld\n",
-			oc->bom_id, bsi.bsi_base_id.eid_id, new_keyval );
-#endif /* ! BACKSQL_ARBITRARY_KEY */
+	Debug(LDAP_DEBUG_TRACE,
+	      "   backsql_add(): executing \"%s\" for dn=\"%s\"  oc_map_id=" BACKSQL_IDNUMFMT " p_id=" BACKSQL_IDFMT " keyval=" BACKSQL_IDNUMFMT "\n",
+	      bi->sql_insentry_stmt, op->ora_e->e_name.bv_val,
+	      oc->bom_id, BACKSQL_IDARG(bsi.bsi_base_id.eid_id),
+	      new_keyval );
+
 	rc = SQLExecute( sth );
 	if ( rc != SQL_SUCCESS ) {
 		Debug( LDAP_DEBUG_TRACE, "   backsql_add(\"%s\"): "
 			"could not insert ldap_entries record\n",
-			op->ora_e->e_name.bv_val, 0, 0 );
+			op->ora_e->e_name.bv_val );
 		backsql_PrintErrors( bi->sql_db_env, dbh, sth, rc );
 		
 		/*
@@ -1421,7 +1406,7 @@ backsql_add( Operation *op, SlapReply *rs )
 	for ( at = op->ora_e->e_attrs; at != NULL; at = at->a_next ) {
 		Debug( LDAP_DEBUG_TRACE, "   backsql_add(): "
 			"adding attribute \"%s\"\n", 
-			at->a_desc->ad_cname.bv_val, 0, 0 );
+			at->a_desc->ad_cname.bv_val );
 
 		/*
 		 * Skip:
@@ -1520,7 +1505,7 @@ done:;
 	}
 
 	if ( !BER_BVISNULL( &bsi.bsi_base_id.eid_ndn ) ) {
-		(void)backsql_free_entryID( op, &bsi.bsi_base_id, 0 );
+		(void)backsql_free_entryID( &bsi.bsi_base_id, 0, op->o_tmpmemctx );
 	}
 
 	if ( !BER_BVISNULL( &p.e_nname ) ) {
