@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-sql/sql-wrap.c,v 1.43.2.2 2007/02/13 20:22:29 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2007 The OpenLDAP Foundation.
+ * Copyright 1999-2020 The OpenLDAP Foundation.
  * Portions Copyright 1999 Dmitry Kovalev.
  * Portions Copyright 2002 Pierangelo Masarati.
  * Portions Copyright 2004 Mark Adamson.
@@ -32,11 +32,6 @@
 
 #define MAX_ATTR_LEN 16384
 
-typedef struct backsql_db_conn {
-	unsigned long	ldap_cid;
-	SQLHDBC		dbh;
-} backsql_db_conn;
-
 void
 backsql_PrintErrors( SQLHENV henv, SQLHDBC hdbc, SQLHSTMT sth, int rc )
 {
@@ -45,7 +40,7 @@ backsql_PrintErrors( SQLHENV henv, SQLHDBC hdbc, SQLHSTMT sth, int rc )
 	SDWORD	iSqlCode;				/* return code    */
 	SWORD	len = SQL_MAX_MESSAGE_LENGTH - 1;	/* return length  */ 
 
-	Debug( LDAP_DEBUG_TRACE, "Return code: %d\n", rc, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "Return code: %d\n", rc );
 
 	for ( ; rc = SQLError( henv, hdbc, sth, state, &iSqlCode, msg,
 		SQL_MAX_MESSAGE_LENGTH - 1, &len ), BACKSQL_SUCCESS( rc ); )
@@ -57,7 +52,7 @@ backsql_PrintErrors( SQLHENV henv, SQLHDBC hdbc, SQLHSTMT sth, int rc )
 }
 
 RETCODE
-backsql_Prepare( SQLHDBC dbh, SQLHSTMT *sth, char *query, int timeout )
+backsql_Prepare( SQLHDBC dbh, SQLHSTMT *sth, const char *query, int timeout )
 {
 	RETCODE		rc;
 
@@ -67,7 +62,7 @@ backsql_Prepare( SQLHDBC dbh, SQLHSTMT *sth, char *query, int timeout )
 	}
 
 #ifdef BACKSQL_TRACE
-	Debug( LDAP_DEBUG_TRACE, "==>backsql_Prepare()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "==>backsql_Prepare()\n" );
 #endif /* BACKSQL_TRACE */
 
 #ifdef BACKSQL_MSSQL_WORKAROUND
@@ -79,7 +74,7 @@ backsql_Prepare( SQLHDBC dbh, SQLHSTMT *sth, char *query, int timeout )
 
 #ifdef BACKSQL_TRACE
 		Debug( LDAP_DEBUG_TRACE, "backsql_Prepare(): driver name=\"%s\"\n",
-				drv_name, 0, 0 );
+				drv_name );
 #endif /* BACKSQL_TRACE */
 
 		ldap_pvt_str2upper( drv_name );
@@ -92,14 +87,13 @@ backsql_Prepare( SQLHDBC dbh, SQLHSTMT *sth, char *query, int timeout )
 			 */
 			Debug( LDAP_DEBUG_TRACE, "_SQLprepare(): "
 				"enabling MS SQL Server default result "
-				"set workaround\n", 0, 0, 0 );
+				"set workaround\n" );
 			rc = SQLSetStmtOption( *sth, SQL_CONCURRENCY, 
 					SQL_CONCUR_ROWVER );
 			if ( rc != SQL_SUCCESS && rc != SQL_SUCCESS_WITH_INFO ) {
 				Debug( LDAP_DEBUG_TRACE, "backsql_Prepare(): "
 					"SQLSetStmtOption(SQL_CONCURRENCY,"
-					"SQL_CONCUR_ROWVER) failed:\n", 
-					0, 0, 0 );
+					"SQL_CONCUR_ROWVER) failed:\n" );
 				backsql_PrintErrors( SQL_NULL_HENV, dbh, *sth, rc );
 				SQLFreeStmt( *sth, SQL_DROP );
 				return rc;
@@ -111,7 +105,7 @@ backsql_Prepare( SQLHDBC dbh, SQLHSTMT *sth, char *query, int timeout )
 	if ( timeout > 0 ) {
 		Debug( LDAP_DEBUG_TRACE, "_SQLprepare(): "
 			"setting query timeout to %d sec.\n", 
-			timeout, 0, 0 );
+			timeout );
 		rc = SQLSetStmtOption( *sth, SQL_QUERY_TIMEOUT, timeout );
 		if ( rc != SQL_SUCCESS ) {
 			backsql_PrintErrors( SQL_NULL_HENV, dbh, *sth, rc );
@@ -121,8 +115,7 @@ backsql_Prepare( SQLHDBC dbh, SQLHSTMT *sth, char *query, int timeout )
 	}
 
 #ifdef BACKSQL_TRACE
-	Debug( LDAP_DEBUG_TRACE, "<==backsql_Prepare() calling SQLPrepare()\n",
-			0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<==backsql_Prepare() calling SQLPrepare()\n" );
 #endif /* BACKSQL_TRACE */
 
 	return SQLPrepare( *sth, (SQLCHAR *)query, SQL_NTS );
@@ -132,92 +125,150 @@ RETCODE
 backsql_BindRowAsStrings_x( SQLHSTMT sth, BACKSQL_ROW_NTS *row, void *ctx )
 {
 	RETCODE		rc;
-	SQLCHAR		colname[ 64 ];
-	SQLSMALLINT	name_len, col_type, col_scale, col_null;
-	UDWORD		col_prec;
-	int		i;
 
 	if ( row == NULL ) {
 		return SQL_ERROR;
 	}
 
 #ifdef BACKSQL_TRACE
-	Debug( LDAP_DEBUG_TRACE, "==> backsql_BindRowAsStrings()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "==> backsql_BindRowAsStrings()\n" );
 #endif /* BACKSQL_TRACE */
 	
 	rc = SQLNumResultCols( sth, &row->ncols );
 	if ( rc != SQL_SUCCESS ) {
 #ifdef BACKSQL_TRACE
 		Debug( LDAP_DEBUG_TRACE, "backsql_BindRowAsStrings(): "
-			"SQLNumResultCols() failed:\n", 0, 0, 0 );
+			"SQLNumResultCols() failed:\n" );
 #endif /* BACKSQL_TRACE */
 		
 		backsql_PrintErrors( SQL_NULL_HENV, SQL_NULL_HDBC, sth, rc );
 
 	} else {
+		SQLCHAR		colname[ 64 ];
+		SQLSMALLINT	name_len, col_type, col_scale, col_null;
+		SQLLEN		col_prec;
+		int		i;
+
 #ifdef BACKSQL_TRACE
 		Debug( LDAP_DEBUG_TRACE, "backsql_BindRowAsStrings: "
-			"ncols=%d\n", (int)row->ncols, 0, 0 );
+			"ncols=%d\n", (int)row->ncols );
 #endif /* BACKSQL_TRACE */
 
 		row->col_names = (BerVarray)ber_memcalloc_x( row->ncols + 1, 
 				sizeof( struct berval ), ctx );
-		if ( !row->col_names ) goto nomem3;
-		row->cols = (char **)ber_memcalloc_x( row->ncols + 1, 
-				sizeof( char * ), ctx );
-		if ( !row->cols ) goto nomem2;
+		if ( row->col_names == NULL ) {
+			goto nomem;
+		}
+
 		row->col_prec = (UDWORD *)ber_memcalloc_x( row->ncols,
 				sizeof( UDWORD ), ctx );
-		if ( !row->col_prec ) goto nomem1;
-		row->value_len = (SQLINTEGER *)ber_memcalloc_x( row->ncols,
-				sizeof( SQLINTEGER ), ctx );
-		if ( !row->value_len ) {
+		if ( row->col_prec == NULL ) {
+			goto nomem;
+		}
+
+		row->col_type = (SQLSMALLINT *)ber_memcalloc_x( row->ncols,
+				sizeof( SQLSMALLINT ), ctx );
+		if ( row->col_type == NULL ) {
+			goto nomem;
+		}
+
+		row->cols = (char **)ber_memcalloc_x( row->ncols + 1, 
+				sizeof( char * ), ctx );
+		if ( row->cols == NULL ) {
+			goto nomem;
+		}
+
+		row->value_len = (SQLLEN *)ber_memcalloc_x( row->ncols,
+				sizeof( SQLLEN ), ctx );
+		if ( row->value_len == NULL ) {
+			goto nomem;
+		}
+
+		if ( 0 ) {
+nomem:
+			ber_memfree_x( row->col_names, ctx );
+			row->col_names = NULL;
 			ber_memfree_x( row->col_prec, ctx );
 			row->col_prec = NULL;
-nomem1:		ber_memfree_x( row->cols, ctx );
+			ber_memfree_x( row->col_type, ctx );
+			row->col_type = NULL;
+			ber_memfree_x( row->cols, ctx );
 			row->cols = NULL;
-nomem2:		ber_memfree_x( row->col_names, ctx );
-			row->col_names = NULL;
-nomem3:		Debug( LDAP_DEBUG_ANY, "backsql_BindRowAsStrings: "
-				"out of memory\n", 0, 0, 0 );
+			ber_memfree_x( row->value_len, ctx );
+			row->value_len = NULL;
+
+			Debug( LDAP_DEBUG_ANY, "backsql_BindRowAsStrings: "
+				"out of memory\n" );
+
 			return LDAP_NO_MEMORY;
 		}
-		for ( i = 1; i <= row->ncols; i++ ) {
-			rc = SQLDescribeCol( sth, (SQLSMALLINT)i, &colname[ 0 ],
+
+		for ( i = 0; i < row->ncols; i++ ) {
+			SQLSMALLINT	TargetType;
+
+			rc = SQLDescribeCol( sth, (SQLSMALLINT)(i + 1), &colname[ 0 ],
 					(SQLUINTEGER)( sizeof( colname ) - 1 ),
 					&name_len, &col_type,
 					&col_prec, &col_scale, &col_null );
 			/* FIXME: test rc? */
 
 			ber_str2bv_x( (char *)colname, 0, 1,
-					&row->col_names[ i - 1 ], ctx );
+					&row->col_names[ i ], ctx );
 #ifdef BACKSQL_TRACE
 			Debug( LDAP_DEBUG_TRACE, "backsql_BindRowAsStrings: "
 				"col_name=%s, col_prec[%d]=%d\n",
-				colname, (int)i, (int)col_prec );
+				colname, (int)(i + 1), (int)col_prec );
 #endif /* BACKSQL_TRACE */
 			if ( col_type != SQL_CHAR && col_type != SQL_VARCHAR )
 			{
 				col_prec = MAX_ATTR_LEN;
 			}
 
-			row->cols[ i - 1 ] = (char *)ber_memcalloc_x( col_prec + 1,
+			row->cols[ i ] = (char *)ber_memcalloc_x( col_prec + 1,
 					sizeof( char ), ctx );
-			row->col_prec[ i - 1 ] = col_prec;
-			rc = SQLBindCol( sth, (SQLUSMALLINT)i,
-					 SQL_C_CHAR,
-					 (SQLPOINTER)row->cols[ i - 1 ],
-					 col_prec + 1,
-					 &row->value_len[ i - 1 ] );
+			row->col_prec[ i ] = col_prec;
+			row->col_type[ i ] = col_type;
+
+			/*
+			 * ITS#3386, ITS#3113 - 20070308
+			 * Note: there are many differences between various DPMS and ODBC
+			 * Systems; some support SQL_C_BLOB, SQL_C_BLOB_LOCATOR.  YMMV:
+			 * This has only been tested on Linux/MySQL/UnixODBC
+			 * For BINARY-type Fields (BLOB, etc), read the data as BINARY
+			 */
+			if ( BACKSQL_IS_BINARY( col_type ) ) {
+#ifdef BACKSQL_TRACE
+				Debug( LDAP_DEBUG_TRACE, "backsql_BindRowAsStrings: "
+					"col_name=%s, col_type[%d]=%d: reading binary data\n",
+					colname, (int)(i + 1), (int)col_type);
+#endif /* BACKSQL_TRACE */
+				TargetType = SQL_C_BINARY;
+
+			} else {
+				/* Otherwise read it as Character data */
+#ifdef BACKSQL_TRACE
+				Debug( LDAP_DEBUG_TRACE, "backsql_BindRowAsStrings: "
+					"col_name=%s, col_type[%d]=%d: reading character data\n",
+					colname, (int)(i + 1), (int)col_type);
+#endif /* BACKSQL_TRACE */
+				TargetType = SQL_C_CHAR;
+			}
+
+			rc = SQLBindCol( sth, (SQLUSMALLINT)(i + 1),
+				 TargetType,
+				 (SQLPOINTER)row->cols[ i ],
+				 col_prec + 1,
+				 &row->value_len[ i ] );
+
 			/* FIXME: test rc? */
 		}
 
-		BER_BVZERO( &row->col_names[ i - 1 ] );
-		row->cols[ i - 1 ] = NULL;
+		BER_BVZERO( &row->col_names[ i ] );
+		row->cols[ i ] = NULL;
 	}
 
 #ifdef BACKSQL_TRACE
-	Debug( LDAP_DEBUG_TRACE, "<== backsql_BindRowAsStrings()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<== backsql_BindRowAsStrings()\n" );
 #endif /* BACKSQL_TRACE */
 
 	return rc;
@@ -237,8 +288,9 @@ backsql_FreeRow_x( BACKSQL_ROW_NTS *row, void *ctx )
 	}
 
 	ber_bvarray_free_x( row->col_names, ctx );
-	ber_memvfree_x( (void **)row->cols, ctx );
 	ber_memfree_x( row->col_prec, ctx );
+	ber_memfree_x( row->col_type, ctx );
+	ber_memvfree_x( (void **)row->cols, ctx );
 	ber_memfree_x( row->value_len, ctx );
 
 	return SQL_SUCCESS;
@@ -251,29 +303,15 @@ backsql_FreeRow( BACKSQL_ROW_NTS *row )
 	return backsql_FreeRow_x( row, NULL );
 }
 
-static int
-backsql_cmp_connid( const void *v_c1, const void *v_c2 )
-{
-	const backsql_db_conn *c1 = v_c1, *c2 = v_c2;
-	if ( c1->ldap_cid > c2->ldap_cid ) {
-		return 1;
-	}
-	
-	if ( c1->ldap_cid < c2->ldap_cid ) {
-		return -1;
-	}
-	
-	return 0;
-}
-
 static void
-backsql_close_db_conn( void *v_conn )
+backsql_close_db_handle( SQLHDBC dbh )
 {
-	backsql_db_conn	*conn = 	(backsql_db_conn *)v_conn;
-	unsigned long	cid = conn->ldap_cid;
+	if ( dbh == SQL_NULL_HDBC ) {
+		return;
+	}
 
-	Debug( LDAP_DEBUG_TRACE, "==>backsql_close_db_conn(%lu)\n",
-		cid, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "==>backsql_close_db_handle(%p)\n",
+		(void *)dbh );
 
 	/*
 	 * Default transact is SQL_ROLLBACK; commit is required only
@@ -282,21 +320,18 @@ backsql_close_db_conn( void *v_conn )
 	 */
 
 	/* TimesTen */
-	SQLTransact( SQL_NULL_HENV, conn->dbh, SQL_ROLLBACK );
-	SQLDisconnect( conn->dbh );
-	SQLFreeConnect( conn->dbh );
-	ch_free( conn );
+	SQLTransact( SQL_NULL_HENV, dbh, SQL_ROLLBACK );
+	SQLDisconnect( dbh );
+	SQLFreeConnect( dbh );
 
-	Debug( LDAP_DEBUG_TRACE, "<==backsql_close_db_conn(%lu)\n",
-		cid, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<==backsql_close_db_handle(%p)\n",
+		(void *)dbh );
 }
 
 int
 backsql_conn_destroy(
 	backsql_info	*bi )
 {
-	avl_free( bi->sql_db_conns, backsql_close_db_conn );
-
 	return 0;
 }
 
@@ -306,18 +341,17 @@ backsql_init_db_env( backsql_info *bi )
 	RETCODE		rc;
 	int		ret = SQL_SUCCESS;
 	
-	Debug( LDAP_DEBUG_TRACE, "==>backsql_init_db_env()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "==>backsql_init_db_env()\n" );
 
 	rc = SQLAllocEnv( &bi->sql_db_env );
 	if ( rc != SQL_SUCCESS ) {
-		Debug( LDAP_DEBUG_TRACE, "init_db_env: SQLAllocEnv failed:\n",
-				0, 0, 0 );
+		Debug( LDAP_DEBUG_TRACE, "init_db_env: SQLAllocEnv failed:\n" );
 		backsql_PrintErrors( SQL_NULL_HENV, SQL_NULL_HDBC,
 				SQL_NULL_HENV, rc );
 		ret = SQL_ERROR;
 	}
 
-	Debug( LDAP_DEBUG_TRACE, "<==backsql_init_db_env()=%d\n", ret, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<==backsql_init_db_env()=%d\n", ret );
 
 	return ret;
 }
@@ -325,7 +359,7 @@ backsql_init_db_env( backsql_info *bi )
 int
 backsql_free_db_env( backsql_info *bi )
 {
-	Debug( LDAP_DEBUG_TRACE, "==>backsql_free_db_env()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "==>backsql_free_db_env()\n" );
 
 	(void)SQLFreeEnv( bi->sql_db_env );
 	bi->sql_db_env = SQL_NULL_HENV;
@@ -335,48 +369,47 @@ backsql_free_db_env( backsql_info *bi )
 	 * before calling this -- then what are we going to delete?? 
 	 * everything is already deleted...
 	 */
-	Debug( LDAP_DEBUG_TRACE, "<==backsql_free_db_env()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<==backsql_free_db_env()\n" );
 
 	return SQL_SUCCESS;
 }
 
 static int
-backsql_open_db_conn( backsql_info *bi, unsigned long ldap_cid, backsql_db_conn **pdbc )
+backsql_open_db_handle(
+	backsql_info	*bi,
+	SQLHDBC		*dbhp )
 {
 	/* TimesTen */
 	char			DBMSName[ 32 ];
-	backsql_db_conn		*dbc;
 	int			rc;
 
-	assert( pdbc != NULL );
-	*pdbc = NULL;
+	assert( dbhp != NULL );
+	*dbhp = SQL_NULL_HDBC;
  
-	Debug( LDAP_DEBUG_TRACE, "==>backsql_open_db_conn(%lu)\n",
-		ldap_cid, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "==>backsql_open_db_handle()\n" );
 
-	dbc = (backsql_db_conn *)ch_calloc( 1, sizeof( backsql_db_conn ) );
-	dbc->ldap_cid = ldap_cid;
-	rc = SQLAllocConnect( bi->sql_db_env, &dbc->dbh );
+	rc = SQLAllocConnect( bi->sql_db_env, dbhp );
 	if ( !BACKSQL_SUCCESS( rc ) ) {
-		Debug( LDAP_DEBUG_TRACE, "backsql_open_db_conn(%lu): "
-			"SQLAllocConnect() failed:\n", ldap_cid, 0, 0 );
+		Debug( LDAP_DEBUG_TRACE, "backsql_open_db_handle(): "
+			"SQLAllocConnect() failed:\n" );
 		backsql_PrintErrors( bi->sql_db_env, SQL_NULL_HDBC,
-				SQL_NULL_HENV, rc );
+			SQL_NULL_HENV, rc );
 		return LDAP_UNAVAILABLE;
 	}
 
-	rc = SQLConnect( dbc->dbh,
-			(SQLCHAR*)bi->sql_dbname, SQL_NTS,
-			(SQLCHAR*)bi->sql_dbuser, SQL_NTS,
-			(SQLCHAR*)bi->sql_dbpasswd, SQL_NTS );
+	rc = SQLConnect( *dbhp,
+		(SQLCHAR*)bi->sql_dbname, SQL_NTS,
+		(SQLCHAR*)bi->sql_dbuser, SQL_NTS,
+		(SQLCHAR*)bi->sql_dbpasswd, SQL_NTS );
 	if ( rc != SQL_SUCCESS ) {
-		Debug( LDAP_DEBUG_TRACE, "backsql_open_db_conn(%lu): "
+		Debug( LDAP_DEBUG_TRACE, "backsql_open_db_handle(): "
 			"SQLConnect() to database \"%s\" %s.\n",
-			ldap_cid, bi->sql_dbname,
+			bi->sql_dbname,
 			rc == SQL_SUCCESS_WITH_INFO ?
-			"succeeded with info" : "failed" );
-		backsql_PrintErrors( bi->sql_db_env, dbc->dbh, SQL_NULL_HENV, rc );
+				"succeeded with info" : "failed" );
+		backsql_PrintErrors( bi->sql_db_env, *dbhp, SQL_NULL_HENV, rc );
 		if ( rc != SQL_SUCCESS_WITH_INFO ) {
+			SQLFreeConnect( *dbhp );
 			return LDAP_UNAVAILABLE;
 		}
 	}
@@ -385,7 +418,8 @@ backsql_open_db_conn( backsql_info *bi, unsigned long ldap_cid, backsql_db_conn
 	 * TimesTen : Turn off autocommit.  We must explicitly
 	 * commit any transactions. 
 	 */
-	SQLSetConnectOption( dbc->dbh, SQL_AUTOCOMMIT, SQL_AUTOCOMMIT_OFF );
+	SQLSetConnectOption( *dbhp, SQL_AUTOCOMMIT,
+		BACKSQL_AUTOCOMMIT_ON( bi ) ?  SQL_AUTOCOMMIT_ON : SQL_AUTOCOMMIT_OFF );
 
 	/* 
 	 * See if this connection is to TimesTen.  If it is,
@@ -394,114 +428,101 @@ backsql_open_db_conn( backsql_info *bi, unsigned long ldap_cid, backsql_db_conn
 	/* Assume until proven otherwise */
 	bi->sql_flags &= ~BSQLF_USE_REVERSE_DN;
 	DBMSName[ 0 ] = '\0';
-	rc = SQLGetInfo( dbc->dbh, SQL_DBMS_NAME, (PTR)&DBMSName,
+	rc = SQLGetInfo( *dbhp, SQL_DBMS_NAME, (PTR)&DBMSName,
 			sizeof( DBMSName ), NULL );
 	if ( rc == SQL_SUCCESS ) {
 		if ( strcmp( DBMSName, "TimesTen" ) == 0 ||
-				strcmp( DBMSName, "Front-Tier" ) == 0 ) {
-			Debug( LDAP_DEBUG_TRACE, "backsql_open_db_conn(%lu): "
-				"TimesTen database!\n", ldap_cid, 0, 0 );
+			strcmp( DBMSName, "Front-Tier" ) == 0 )
+		{
+			Debug( LDAP_DEBUG_TRACE, "backsql_open_db_handle(): "
+				"TimesTen database!\n" );
 			bi->sql_flags |= BSQLF_USE_REVERSE_DN;
 		}
+
 	} else {
-		Debug( LDAP_DEBUG_TRACE, "backsql_open_db_conn(%lu): "
-			"SQLGetInfo() failed.\n", ldap_cid, 0, 0 );
-		backsql_PrintErrors( bi->sql_db_env, dbc->dbh, SQL_NULL_HENV, rc );
-		return rc;
+		Debug( LDAP_DEBUG_TRACE, "backsql_open_db_handle(): "
+			"SQLGetInfo() failed.\n" );
+		backsql_PrintErrors( bi->sql_db_env, *dbhp, SQL_NULL_HENV, rc );
+		SQLDisconnect( *dbhp );
+		SQLFreeConnect( *dbhp );
+		return LDAP_UNAVAILABLE;
 	}
 	/* end TimesTen */
 
-	Debug( LDAP_DEBUG_TRACE, "backsql_open_db_conn(%lu): "
-		"connected, adding to tree.\n", ldap_cid, 0, 0 );
-	ldap_pvt_thread_mutex_lock( &bi->sql_dbconn_mutex );
-	if ( avl_insert( &bi->sql_db_conns, dbc, backsql_cmp_connid, avl_dup_error ) ) {
-		Debug( LDAP_DEBUG_TRACE, "backsql_open_db_conn(%lu): "
-			"duplicate connection ID.\n", ldap_cid, 0, 0 );
-		return LDAP_OTHER;
-	}
-	ldap_pvt_thread_mutex_unlock( &bi->sql_dbconn_mutex );
-	Debug( LDAP_DEBUG_TRACE, "<==backsql_open_db_conn(%lu)\n", ldap_cid, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<==backsql_open_db_handle()\n" );
 
-	*pdbc = dbc;
+	return LDAP_SUCCESS;
+}
+
+static void	*backsql_db_conn_dummy;
+
+static void
+backsql_db_conn_keyfree(
+	void		*key,
+	void		*data )
+{
+	(void)backsql_close_db_handle( (SQLHDBC)data );
+}
+
+int
+backsql_free_db_conn( Operation *op, SQLHDBC dbh )
+{
+	Debug( LDAP_DEBUG_TRACE, "==>backsql_free_db_conn()\n" );
+
+	(void)backsql_close_db_handle( dbh );
+	ldap_pvt_thread_pool_setkey( op->o_threadctx,
+		&backsql_db_conn_dummy, (void *)SQL_NULL_HDBC,
+		backsql_db_conn_keyfree, NULL, NULL );
+
+	Debug( LDAP_DEBUG_TRACE, "<==backsql_free_db_conn()\n" );
 
 	return LDAP_SUCCESS;
 }
 
 int
-backsql_free_db_conn( Operation *op )
+backsql_get_db_conn( Operation *op, SQLHDBC *dbhp )
 {
-	backsql_info		*bi = (backsql_info *)op->o_bd->be_private;
-	backsql_db_conn		tmp = { 0 },
-				*conn;
+	backsql_info	*bi = (backsql_info *)op->o_bd->be_private;
+	int		rc = LDAP_SUCCESS;
+	SQLHDBC		dbh = SQL_NULL_HDBC;
 
-	Debug( LDAP_DEBUG_TRACE, "==>backsql_free_db_conn()\n", 0, 0, 0 );
-	tmp.ldap_cid = op->o_connid;
-	ldap_pvt_thread_mutex_lock( &bi->sql_dbconn_mutex );
-	conn = avl_delete( &bi->sql_db_conns, &tmp, backsql_cmp_connid );
-	ldap_pvt_thread_mutex_unlock( &bi->sql_dbconn_mutex );
+	Debug( LDAP_DEBUG_TRACE, "==>backsql_get_db_conn()\n" );
 
-	/*
-	 * we have one thread per connection, as I understand -- so we can
-	 * get this out of critical section
-	 */
-	if ( conn != NULL ) {
-		Debug( LDAP_DEBUG_TRACE, "backsql_free_db_conn(): "
-			"closing db connection %lu (%p)\n",
-			op->o_connid, (void *)conn, 0 );
-		backsql_close_db_conn( (void *)conn );
+	assert( dbhp != NULL );
+	*dbhp = SQL_NULL_HDBC;
+
+	if ( op->o_threadctx ) {
+		void		*data = NULL;
+
+		ldap_pvt_thread_pool_getkey( op->o_threadctx,
+				&backsql_db_conn_dummy, &data, NULL );
+		dbh = (SQLHDBC)data;
+
+	} else {
+		dbh = bi->sql_dbh;
 	}
 
-	Debug( LDAP_DEBUG_TRACE, "<==backsql_free_db_conn()\n", 0, 0, 0 );
-
-	return conn ? SQL_SUCCESS : SQL_ERROR;
-}
-
-int
-backsql_get_db_conn( Operation *op, SQLHDBC *dbh )
-{
-	backsql_info		*bi = (backsql_info *)op->o_bd->be_private;
-	backsql_db_conn		*dbc,
-				tmp = { 0 };
-	int			rc = LDAP_SUCCESS;
-
-	Debug( LDAP_DEBUG_TRACE, "==>backsql_get_db_conn()\n", 0, 0, 0 );
-
-	assert( dbh != NULL );
-	*dbh = SQL_NULL_HDBC;
-
-	tmp.ldap_cid = op->o_connid;
-
-	/*
-	 * we have one thread per connection, as I understand -- 
-	 * so we do not need locking here
-	 */
-	dbc = avl_find( bi->sql_db_conns, &tmp, backsql_cmp_connid );
-	if ( !dbc ) {
-		rc = backsql_open_db_conn( bi, op->o_connid, &dbc );
-		if ( rc != LDAP_SUCCESS) {
-			Debug( LDAP_DEBUG_TRACE, "backsql_get_db_conn(): "
-				"could not get connection handle "
-				"-- returning NULL\n", 0, 0, 0 );
-			return rc;
-		}
-	}
-
-	ldap_pvt_thread_mutex_lock( &bi->sql_schema_mutex );
-	if ( !BACKSQL_SCHEMA_LOADED( bi ) ) {
-		Debug( LDAP_DEBUG_TRACE, "backsql_get_db_conn(): "
-			"first call -- reading schema map\n", 0, 0, 0 );
-		rc = backsql_load_schema_map( bi, dbc->dbh );
+	if ( dbh == SQL_NULL_HDBC ) {
+		rc = backsql_open_db_handle( bi, &dbh );
 		if ( rc != LDAP_SUCCESS ) {
-			ldap_pvt_thread_mutex_unlock( &bi->sql_schema_mutex );
-			backsql_free_db_conn( op );
 			return rc;
 		}
+
+		if ( op->o_threadctx ) {
+			void		*data = (void *)dbh;
+
+			ldap_pvt_thread_pool_setkey( op->o_threadctx,
+					&backsql_db_conn_dummy, data,
+					backsql_db_conn_keyfree, NULL, NULL );
+
+		} else {
+			bi->sql_dbh = dbh;
+		}
 	}
-	ldap_pvt_thread_mutex_unlock( &bi->sql_schema_mutex );
 
-	*dbh = dbc->dbh;
+	*dbhp = dbh;
 
-	Debug( LDAP_DEBUG_TRACE, "<==backsql_get_db_conn()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<==backsql_get_db_conn()\n" );
 
 	return LDAP_SUCCESS;
 }
