@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/back-sql/util.c,v 1.45.2.1 2007/02/13 20:22:29 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2007 The OpenLDAP Foundation.
+ * Copyright 1999-2020 The OpenLDAP Foundation.
  * Portions Copyright 1999 Dmitry Kovalev.
  * Portions Copyright 2002 Pierangelo Masarati.
  * All rights reserved.
@@ -37,32 +37,32 @@
 
 #define BACKSQL_STR_GROW 256
 
-char backsql_def_oc_query[] = 
+const char backsql_def_oc_query[] = 
 	"SELECT id,name,keytbl,keycol,create_proc,delete_proc,expect_return "
 	"FROM ldap_oc_mappings";
-char backsql_def_needs_select_oc_query[] = 
+const char backsql_def_needs_select_oc_query[] = 
 	"SELECT id,name,keytbl,keycol,create_proc,create_keyval,delete_proc,"
 	"expect_return FROM ldap_oc_mappings";
-char backsql_def_at_query[] = 
+const char backsql_def_at_query[] = 
 	"SELECT name,sel_expr,from_tbls,join_where,add_proc,delete_proc,"
 	"param_order,expect_return,sel_expr_u FROM ldap_attr_mappings "
 	"WHERE oc_map_id=?";
-char backsql_def_delentry_stmt[] = "DELETE FROM ldap_entries WHERE id=?";
-char backsql_def_renentry_stmt[] =
+const char backsql_def_delentry_stmt[] = "DELETE FROM ldap_entries WHERE id=?";
+const char backsql_def_renentry_stmt[] =
 	"UPDATE ldap_entries SET dn=?,parent=?,keyval=? WHERE id=?";
-char backsql_def_insentry_stmt[] = 
+const char backsql_def_insentry_stmt[] = 
 	"INSERT INTO ldap_entries (dn,oc_map_id,parent,keyval) "
 	"VALUES (?,?,?,?)";
-char backsql_def_delobjclasses_stmt[] = "DELETE FROM ldap_entry_objclasses "
+const char backsql_def_delobjclasses_stmt[] = "DELETE FROM ldap_entry_objclasses "
 	"WHERE entry_id=?";
-char backsql_def_subtree_cond[] = "ldap_entries.dn LIKE CONCAT('%',?)";
-char backsql_def_upper_subtree_cond[] = "(ldap_entries.dn) LIKE CONCAT('%',?)";
-char backsql_id_query[] = "SELECT id,keyval,oc_map_id,dn FROM ldap_entries WHERE ";
+const char backsql_def_subtree_cond[] = "ldap_entries.dn LIKE CONCAT('%',?)";
+const char backsql_def_upper_subtree_cond[] = "(ldap_entries.dn) LIKE CONCAT('%',?)";
+const char backsql_id_query[] = "SELECT id,keyval,oc_map_id,dn FROM ldap_entries WHERE ";
 /* better ?||? or cast(?||? as varchar) */ 
-char backsql_def_concat_func[] = "CONCAT(?,?)";
+const char backsql_def_concat_func[] = "CONCAT(?,?)";
 
 /* TimesTen */
-char backsql_check_dn_ru_query[] = "SELECT dn_ru FROM ldap_entries";
+const char backsql_check_dn_ru_query[] = "SELECT dn_ru FROM ldap_entries";
 
 struct berbuf *
 backsql_strcat_x( struct berbuf *dest, void *memctx, ... )
@@ -76,7 +76,7 @@ backsql_strcat_x( struct berbuf *dest, void *memctx, ... )
 			|| dest->bb_val.bv_len == strlen( dest->bb_val.bv_val ) );
  
 #ifdef BACKSQL_TRACE
-	Debug( LDAP_DEBUG_TRACE, "==>backsql_strcat()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "==>backsql_strcat()\n" );
 #endif /* BACKSQL_TRACE */
 
 	va_start( strs, memctx );
@@ -103,8 +103,8 @@ backsql_strcat_x( struct berbuf *dest, void *memctx, ... )
 					dest->bb_len + grow * sizeof( char ), memctx );
 			if ( tmp_dest == NULL ) {
 				Debug( LDAP_DEBUG_ANY, "backsql_strcat(): "
-					"could not reallocate string buffer.\n",
-					0, 0, 0 );
+					"could not reallocate string buffer.\n" );
+				va_end( strs );
 				return NULL;
 			}
 			dest->bb_val.bv_val = tmp_dest;
@@ -113,7 +113,7 @@ backsql_strcat_x( struct berbuf *dest, void *memctx, ... )
 #ifdef BACKSQL_TRACE
 			Debug( LDAP_DEBUG_TRACE, "backsql_strcat(): "
 				"new buflen=%d, dest=%p\n",
-				dest->bb_len, dest, 0 );
+				dest->bb_len, dest );
 #endif /* BACKSQL_TRACE */
 		}
 		AC_MEMCPY( dest->bb_val.bv_val + cdlen, cstr, cslen + 1 );
@@ -123,7 +123,7 @@ backsql_strcat_x( struct berbuf *dest, void *memctx, ... )
 
 #ifdef BACKSQL_TRACE
 	Debug( LDAP_DEBUG_TRACE, "<==backsql_strcat() (dest=\"%s\")\n", 
-			dest->bb_val.bv_val, 0, 0 );
+			dest->bb_val.bv_val );
 #endif /* BACKSQL_TRACE */
 
 	dest->bb_val.bv_len = cdlen;
@@ -144,7 +144,7 @@ backsql_strfcat_x( struct berbuf *dest, void *memctx, const char *fmt, ... )
 			|| dest->bb_val.bv_len == strlen( dest->bb_val.bv_val ) );
  
 #ifdef BACKSQL_TRACE
-	Debug( LDAP_DEBUG_TRACE, "==>backsql_strfcat()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "==>backsql_strfcat()\n" );
 #endif /* BACKSQL_TRACE */
 
 	va_start( strs, fmt );
@@ -210,8 +210,8 @@ backsql_strfcat_x( struct berbuf *dest, void *memctx, const char *fmt, ... )
 					( dest->bb_len ) + grow * sizeof( char ), memctx );
 			if ( tmp_dest == NULL ) {
 				Debug( LDAP_DEBUG_ANY, "backsql_strfcat(): "
-					"could not reallocate string buffer.\n",
-					0, 0, 0 );
+					"could not reallocate string buffer.\n" );
+				va_end( strs );
 				return NULL;
 			}
 			dest->bb_val.bv_val = tmp_dest;
@@ -219,7 +219,7 @@ backsql_strfcat_x( struct berbuf *dest, void *memctx, const char *fmt, ... )
 
 #ifdef BACKSQL_TRACE
 			Debug( LDAP_DEBUG_TRACE, "backsql_strfcat(): "
-				"new buflen=%d, dest=%p\n", dest->bb_len, dest, 0 );
+				"new buflen=%d, dest=%p\n", dest->bb_len, dest );
 #endif /* BACKSQL_TRACE */
 		}
 
@@ -233,7 +233,7 @@ backsql_strfcat_x( struct berbuf *dest, void *memctx, const char *fmt, ... )
 
 #ifdef BACKSQL_TRACE
 	Debug( LDAP_DEBUG_TRACE, "<==backsql_strfcat() (dest=\"%s\")\n", 
-			dest->bb_val.bv_val, 0, 0 );
+			dest->bb_val.bv_val );
 #endif /* BACKSQL_TRACE */
 
 	dest->bb_val.bv_len = cdlen;
@@ -266,7 +266,7 @@ backsql_entry_addattr(
 
 #ifdef BACKSQL_TRACE
 	Debug( LDAP_DEBUG_TRACE, "<==backsql_entry_addattr(\"%s\")\n",
-		e->e_name.bv_val, 0, 0 );
+		e->e_name.bv_val );
 #endif /* BACKSQL_TRACE */
 
 	return LDAP_SUCCESS;
@@ -333,7 +333,7 @@ backsql_merge_from_clause(
 	Debug( LDAP_DEBUG_TRACE, "==>backsql_merge_from_clause(): "
 		"dest_from=\"%s\",src_from=\"%s\"\n",
  		dest_from ? dest_from->bb_val.bv_val : "<NULL>",
-		src_from->bv_val, 0 );
+		src_from->bv_val );
 #endif /* BACKSQL_TRACE */
 
 	srcc = ch_strdup( src_from->bv_val );
@@ -348,7 +348,7 @@ backsql_merge_from_clause(
 
 #ifdef BACKSQL_TRACE
 		Debug( LDAP_DEBUG_TRACE, "backsql_merge_from_clause(): "
-			"p=\"%s\" s=\"%s\"\n", p, s, 0 );
+			"p=\"%s\" s=\"%s\"\n", p, s );
 #endif /* BACKSQL_TRACE */
 
 		if ( BER_BVISNULL( &res.bb_val ) ) {
@@ -367,7 +367,7 @@ backsql_merge_from_clause(
 	}
 
 #ifdef BACKSQL_TRACE
-	Debug( LDAP_DEBUG_TRACE, "<==backsql_merge_from_clause()\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_TRACE, "<==backsql_merge_from_clause()\n" );
 #endif /* BACKSQL_TRACE */
 
 	free( srcc );
