@@ -1,8 +1,8 @@
 /* ch_malloc.c - malloc routines that test returns from malloc and friends */
-/* $OpenLDAP: pkg/ldap/servers/slapd/ch_malloc.c,v 1.28.2.1 2007/02/13 20:22:17 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2007 The OpenLDAP Foundation.
+ * Copyright 1998-2020 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -53,7 +53,7 @@ ch_malloc(
 
 	if ( (new = (void *) ber_memalloc_x( size, NULL )) == NULL ) {
 		Debug( LDAP_DEBUG_ANY, "ch_malloc of %lu bytes failed\n",
-			(long) size, 0, 0 );
+			(long) size );
 		assert( 0 );
 		exit( EXIT_FAILURE );
 	}
@@ -85,7 +85,7 @@ ch_realloc(
 
 	if ( (new = (void *) ber_memrealloc_x( block, size, NULL )) == NULL ) {
 		Debug( LDAP_DEBUG_ANY, "ch_realloc of %lu bytes failed\n",
-			(long) size, 0, 0 );
+			(long) size );
 		assert( 0 );
 		exit( EXIT_FAILURE );
 	}
@@ -103,7 +103,7 @@ ch_calloc(
 
 	if ( (new = (void *) ber_memcalloc_x( nelem, size, NULL )) == NULL ) {
 		Debug( LDAP_DEBUG_ANY, "ch_calloc of %lu elems of %lu bytes failed\n",
-		  (long) nelem, (long) size, 0 );
+		  (long) nelem, (long) size );
 		assert( 0 );
 		exit( EXIT_FAILURE );
 	}
@@ -119,7 +119,7 @@ ch_strdup(
 	char	*new;
 
 	if ( (new = ber_strdup_x( string, NULL )) == NULL ) {
-		Debug( LDAP_DEBUG_ANY, "ch_strdup(%s) failed\n", string, 0, 0 );
+		Debug( LDAP_DEBUG_ANY, "ch_strdup(%s) failed\n", string );
 		assert( 0 );
 		exit( EXIT_FAILURE );
 	}
