@@ -1,8 +1,8 @@
 /* cr.c - content rule routines */
-/* $OpenLDAP: pkg/ldap/servers/slapd/cr.c,v 1.22.2.1 2007/02/13 20:22:17 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2007 The OpenLDAP Foundation.
+ * Copyright 1998-2020 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -30,7 +30,7 @@ struct cindexrec {
 };
 
 static Avlnode	*cr_index = NULL;
-static LDAP_STAILQ_HEAD(CRList, slap_content_rule) cr_list
+static LDAP_STAILQ_HEAD(CRList, ContentRule) cr_list
 	= LDAP_STAILQ_HEAD_INITIALIZER(cr_list);
 
 static int
@@ -120,6 +120,8 @@ cr_insert(
 	struct cindexrec	*cir;
 	char			**names;
 
+	assert( scr != NULL );
+
 	if ( scr->scr_oid ) {
 		cir = (struct cindexrec *)
 			ch_calloc( 1, sizeof(struct cindexrec) );
@@ -127,9 +129,6 @@ cr_insert(
 		cir->cir_name.bv_len = strlen( scr->scr_oid );
 		cir->cir_cr = scr;
 
-		assert( cir->cir_name.bv_val != NULL );
-		assert( cir->cir_cr != NULL );
-
 		if ( avl_insert( &cr_index, (caddr_t) cir,
 		                 cr_index_cmp, avl_dup_error ) )
 		{
@@ -150,9 +149,6 @@ cr_insert(
 			cir->cir_name.bv_len = strlen( *names );
 			cir->cir_cr = scr;
 
-			assert( cir->cir_name.bv_val != NULL );
-			assert( cir->cir_cr != NULL );
-
 			if ( avl_insert( &cr_index, (caddr_t) cir,
 			                 cr_index_cmp, avl_dup_error ) )
 			{
@@ -489,7 +485,7 @@ cr_schema_info( Entry *e )
 #endif
 #if 0
 		Debug( LDAP_DEBUG_TRACE, "Merging cr [%ld] %s\n",
-	       (long) val.bv_len, val.bv_val, 0 );
+	       (long) val.bv_len, val.bv_val );
 #endif
 
 		nval.bv_val = cr->scr_oid;
