@@ -1,8 +1,8 @@
 /* filter.c - routines for parsing and dealing with filters */
-/* $OpenLDAP: pkg/ldap/servers/slapd/filter.c,v 1.134.2.2 2007/02/13 20:22:18 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2007 The OpenLDAP Foundation.
+ * Copyright 1998-2020 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -32,12 +32,21 @@
 #include <ac/string.h>
 
 #include "slap.h"
+#include "lutil.h"
+
+const Filter *slap_filter_objectClass_pres;
+const struct berval *slap_filterstr_objectClass_pres;
+
+#ifndef SLAPD_MAX_FILTER_DEPTH
+#define SLAPD_MAX_FILTER_DEPTH	5000
+#endif
 
 static int	get_filter_list(
 	Operation *op,
 	BerElement *ber,
 	Filter **f,
-	const char **text );
+	const char **text,
+	int depth );
 
 static int	get_ssa(
 	Operation *op,
@@ -57,18 +66,39 @@ static int	get_simple_vrFilter(
 	const char **text );
 
 int
-get_filter(
+filter_init( void )
+{
+	static Filter filter_objectClass_pres = { LDAP_FILTER_PRESENT };
+	static struct berval filterstr_objectClass_pres = BER_BVC("(objectClass=*)");
+
+	filter_objectClass_pres.f_desc = slap_schema.si_ad_objectClass;
+
+	slap_filter_objectClass_pres = &filter_objectClass_pres;
+	slap_filterstr_objectClass_pres = &filterstr_objectClass_pres;
+
+	return 0;
+}
+
+void
+filter_destroy( void )
+{
+	return;
+}
+
+static int
+get_filter0(
 	Operation *op,
 	BerElement *ber,
 	Filter **filt,
-	const char **text )
+	const char **text,
+	int depth )
 {
 	ber_tag_t	tag;
 	ber_len_t	len;
 	int		err;
 	Filter		f;
 
-	Debug( LDAP_DEBUG_FILTER, "begin get_filter\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "begin get_filter\n" );
 	/*
 	 * A filter looks like this coming in:
 	 *	Filter ::= CHOICE {
@@ -102,6 +132,11 @@ get_filter(
 	 *
 	 */
 
+	if( depth > SLAPD_MAX_FILTER_DEPTH ) {
+		*text = "filter nested too deeply";
+		return SLAPD_DISCONNECT;
+	}
+
 	tag = ber_peek_tag( ber, &len );
 
 	if( tag == LBER_ERROR ) {
@@ -116,7 +151,7 @@ get_filter(
 
 	switch ( f.f_choice ) {
 	case LDAP_FILTER_EQUALITY:
-		Debug( LDAP_DEBUG_FILTER, "EQUALITY\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "EQUALITY\n" );
 		err = get_ava( op, ber, &f, SLAP_MR_EQUALITY, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
@@ -126,7 +161,7 @@ get_filter(
 		break;
 
 	case LDAP_FILTER_SUBSTRINGS:
-		Debug( LDAP_DEBUG_FILTER, "SUBSTRINGS\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "SUBSTRINGS\n" );
 		err = get_ssa( op, ber, &f, text );
 		if( err != LDAP_SUCCESS ) {
 			break;
@@ -135,7 +170,7 @@ get_filter(
 		break;
 
 	case LDAP_FILTER_GE:
-		Debug( LDAP_DEBUG_FILTER, "GE\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "GE\n" );
 		err = get_ava( op, ber, &f, SLAP_MR_ORDERING, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
@@ -144,7 +179,7 @@ get_filter(
 		break;
 
 	case LDAP_FILTER_LE:
-		Debug( LDAP_DEBUG_FILTER, "LE\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "LE\n" );
 		err = get_ava( op, ber, &f, SLAP_MR_ORDERING, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
@@ -155,7 +190,7 @@ get_filter(
 	case LDAP_FILTER_PRESENT: {
 		struct berval type;
 
-		Debug( LDAP_DEBUG_FILTER, "PRESENT\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "PRESENT\n" );
 		if ( ber_scanf( ber, "m", &type ) == LBER_ERROR ) {
 			err = SLAPD_DISCONNECT;
 			*text = "error decoding filter";
@@ -187,7 +222,7 @@ get_filter(
 		} break;
 
 	case LDAP_FILTER_APPROX:
-		Debug( LDAP_DEBUG_FILTER, "APPROX\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "APPROX\n" );
 		err = get_ava( op, ber, &f, SLAP_MR_EQUALITY_APPROX, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
@@ -196,8 +231,8 @@ get_filter(
 		break;
 
 	case LDAP_FILTER_AND:
-		Debug( LDAP_DEBUG_FILTER, "AND\n", 0, 0, 0 );
-		err = get_filter_list( op, ber, &f.f_and, text );
+		Debug( LDAP_DEBUG_FILTER, "AND\n" );
+		err = get_filter_list( op, ber, &f.f_and, text, depth+1 );
 		if ( err != LDAP_SUCCESS ) {
 			break;
 		}
@@ -209,8 +244,8 @@ get_filter(
 		break;
 
 	case LDAP_FILTER_OR:
-		Debug( LDAP_DEBUG_FILTER, "OR\n", 0, 0, 0 );
-		err = get_filter_list( op, ber, &f.f_or, text );
+		Debug( LDAP_DEBUG_FILTER, "OR\n" );
+		err = get_filter_list( op, ber, &f.f_or, text, depth+1 );
 		if ( err != LDAP_SUCCESS ) {
 			break;
 		}
@@ -222,9 +257,9 @@ get_filter(
 		break;
 
 	case LDAP_FILTER_NOT:
-		Debug( LDAP_DEBUG_FILTER, "NOT\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "NOT\n" );
 		(void) ber_skip_tag( ber, &len );
-		err = get_filter( op, ber, &f.f_not, text );
+		err = get_filter0( op, ber, &f.f_not, text, depth+1 );
 		if ( err != LDAP_SUCCESS ) {
 			break;
 		}
@@ -250,7 +285,7 @@ get_filter(
 		break;
 
 	case LDAP_FILTER_EXT:
-		Debug( LDAP_DEBUG_FILTER, "EXTENSIBLE\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "EXTENSIBLE\n" );
 
 		err = get_mra( op, ber, &f, text );
 		if ( err != LDAP_SUCCESS ) {
@@ -263,7 +298,7 @@ get_filter(
 	default:
 		(void) ber_scanf( ber, "x" ); /* skip the element */
 		Debug( LDAP_DEBUG_ANY, "get_filter: unknown filter type=%lu\n",
-			f.f_choice, 0, 0 );
+			f.f_choice );
 		f.f_choice = SLAPD_FILTER_COMPUTED;
 		f.f_result = SLAPD_COMPARE_UNDEFINED;
 		break;
@@ -282,15 +317,27 @@ get_filter(
 		**filt = f;
 	}
 
-	Debug( LDAP_DEBUG_FILTER, "end get_filter %d\n", err, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "end get_filter %d\n", err );
 
 	return( err );
 }
 
+int
+get_filter(
+	Operation *op,
+	BerElement *ber,
+	Filter **filt,
+	const char **text )
+{
+	return get_filter0( op, ber, filt, text, 0 );
+}
+
+
 static int
 get_filter_list( Operation *op, BerElement *ber,
 	Filter **f,
-	const char **text )
+	const char **text,
+	int depth )
 {
 	Filter		**new;
 	int		err;
@@ -298,20 +345,20 @@ get_filter_list( Operation *op, BerElement *ber,
 	ber_len_t	len;
 	char		*last;
 
-	Debug( LDAP_DEBUG_FILTER, "begin get_filter_list\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "begin get_filter_list\n" );
 	new = f;
 	for ( tag = ber_first_element( ber, &len, &last );
 		tag != LBER_DEFAULT;
 		tag = ber_next_element( ber, &len, last ) )
 	{
-		err = get_filter( op, ber, new, text );
+		err = get_filter0( op, ber, new, text, depth );
 		if ( err != LDAP_SUCCESS )
 			return( err );
 		new = &(*new)->f_next;
 	}
 	*new = NULL;
 
-	Debug( LDAP_DEBUG_FILTER, "end get_filter_list\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "end get_filter_list\n" );
 	return( LDAP_SUCCESS );
 }
 
@@ -324,14 +371,14 @@ get_ssa(
 {
 	ber_tag_t	tag;
 	ber_len_t	len;
-	ber_tag_t	rc;
+	int	rc;
 	struct berval desc, value, nvalue;
 	char		*last;
 	SubstringsAssertion ssa;
 
 	*text = "error decoding filter";
 
-	Debug( LDAP_DEBUG_FILTER, "begin get_ssa\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "begin get_ssa\n" );
 	if ( ber_scanf( ber, "{m" /*}*/, &desc ) == LBER_ERROR ) {
 		return SLAPD_DISCONNECT;
 	}
@@ -361,14 +408,24 @@ get_ssa(
 
 	rc = LDAP_PROTOCOL_ERROR;
 
+	/* If there is no substring matching rule, there's nothing
+	 * we can do with this filter. But we continue to parse it
+	 * for logging purposes.
+	 */
+	if ( ssa.sa_desc->ad_type->sat_substr == NULL ) {
+		f->f_choice |= SLAPD_FILTER_UNDEFINED;
+		Debug( LDAP_DEBUG_FILTER,
+		"get_ssa: no substring matching rule for attributeType %s\n",
+			desc.bv_val );
+	}
+
 	for ( tag = ber_first_element( ber, &len, &last );
 		tag != LBER_DEFAULT;
 		tag = ber_next_element( ber, &len, last ) )
 	{
 		unsigned usage;
 
-		rc = ber_scanf( ber, "m", &value );
-		if ( rc == LBER_ERROR ) {
+		if ( ber_scanf( ber, "m", &value ) == LBER_ERROR ) {
 			rc = SLAPD_DISCONNECT;
 			goto return_error;
 		}
@@ -410,7 +467,7 @@ get_ssa(
 		default:
 			Debug( LDAP_DEBUG_FILTER,
 				"  unknown substring choice=%ld\n",
-				(long) tag, 0, 0 );
+				(long) tag );
 
 			rc = LDAP_PROTOCOL_ERROR;
 			goto return_error;
@@ -420,21 +477,27 @@ get_ssa(
 		rc = asserted_value_validate_normalize(
 			ssa.sa_desc, ssa.sa_desc->ad_type->sat_equality,
 			usage, &value, &nvalue, text, op->o_tmpmemctx );
-		if( rc != LDAP_SUCCESS ) goto return_error;
+		if( rc != LDAP_SUCCESS ) {
+			f->f_choice |= SLAPD_FILTER_UNDEFINED;
+			Debug( LDAP_DEBUG_FILTER,
+			"get_ssa: illegal value for attributeType %s (%d) %s\n",
+				desc.bv_val, rc, *text );
+			ber_dupbv_x( &nvalue, &value, op->o_tmpmemctx );
+		}
 
 		switch ( tag ) {
 		case LDAP_SUBSTRING_INITIAL:
-			Debug( LDAP_DEBUG_FILTER, "  INITIAL\n", 0, 0, 0 );
+			Debug( LDAP_DEBUG_FILTER, "  INITIAL\n" );
 			ssa.sa_initial = nvalue;
 			break;
 
 		case LDAP_SUBSTRING_ANY:
-			Debug( LDAP_DEBUG_FILTER, "  ANY\n", 0, 0, 0 );
+			Debug( LDAP_DEBUG_FILTER, "  ANY\n" );
 			ber_bvarray_add_x( &ssa.sa_any, &nvalue, op->o_tmpmemctx );
 			break;
 
 		case LDAP_SUBSTRING_FINAL:
-			Debug( LDAP_DEBUG_FILTER, "  FINAL\n", 0, 0, 0 );
+			Debug( LDAP_DEBUG_FILTER, "  FINAL\n" );
 			ssa.sa_final = nvalue;
 			break;
 
@@ -445,7 +508,7 @@ get_ssa(
 
 return_error:
 			Debug( LDAP_DEBUG_FILTER, "  error=%ld\n",
-				(long) rc, 0, 0 );
+				(long) rc );
 			slap_sl_free( ssa.sa_initial.bv_val, op->o_tmpmemctx );
 			ber_bvarray_free_x( ssa.sa_any, op->o_tmpmemctx );
 			if ( ssa.sa_desc->ad_flags & SLAP_DESC_TEMPORARY )
@@ -454,6 +517,7 @@ return_error:
 			return rc;
 		}
 
+		*text = NULL;
 		rc = LDAP_SUCCESS;
 	}
 
@@ -462,12 +526,12 @@ return_error:
 		*f->f_sub = ssa;
 	}
 
-	Debug( LDAP_DEBUG_FILTER, "end get_ssa\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "end get_ssa\n" );
 	return rc /* LDAP_SUCCESS */ ;
 }
 
 void
-filter_free_x( Operation *op, Filter *f )
+filter_free_x( Operation *op, Filter *f, int freeme )
 {
 	Filter	*p, *next;
 
@@ -479,6 +543,8 @@ filter_free_x( Operation *op, Filter *f )
 
 	switch ( f->f_choice ) {
 	case LDAP_FILTER_PRESENT:
+		if ( f->f_desc->ad_flags & SLAP_DESC_TEMPORARY )
+			op->o_tmpfree( f->f_desc, op->o_tmpmemctx );
 		break;
 
 	case LDAP_FILTER_EQUALITY:
@@ -506,7 +572,7 @@ filter_free_x( Operation *op, Filter *f )
 	case LDAP_FILTER_NOT:
 		for ( p = f->f_list; p != NULL; p = next ) {
 			next = p->f_next;
-			filter_free_x( op, p );
+			filter_free_x( op, p, 1 );
 		}
 		break;
 
@@ -519,11 +585,13 @@ filter_free_x( Operation *op, Filter *f )
 
 	default:
 		Debug( LDAP_DEBUG_ANY, "filter_free: unknown filter type=%lu\n",
-			f->f_choice, 0, 0 );
+			f->f_choice );
 		break;
 	}
 
-	op->o_tmpfree( f, op->o_tmpmemctx );
+	if ( freeme ) {
+		op->o_tmpfree( f, op->o_tmpmemctx );
+	}
 }
 
 void
@@ -535,25 +603,33 @@ filter_free( Filter *f )
 	op.o_hdr = &ohdr;
 	op.o_tmpmemctx = slap_sl_context( f );
 	op.o_tmpmfuncs = &slap_sl_mfuncs;
-	filter_free_x( &op, f );
+	filter_free_x( &op, f, 1 );
 }
 
 void
 filter2bv_x( Operation *op, Filter *f, struct berval *fstr )
+{
+	filter2bv_undef_x( op, f, 0, fstr );
+}
+
+void
+filter2bv_undef_x( Operation *op, Filter *f, int noundef, struct berval *fstr )
 {
 	int		i;
 	Filter		*p;
-	struct berval	tmp;
+	struct berval	tmp, value;
 	static struct berval
 			ber_bvfalse = BER_BVC( "(?=false)" ),
 			ber_bvtrue = BER_BVC( "(?=true)" ),
 			ber_bvundefined = BER_BVC( "(?=undefined)" ),
 			ber_bverror = BER_BVC( "(?=error)" ),
 			ber_bvunknown = BER_BVC( "(?=unknown)" ),
-			ber_bvnone = BER_BVC( "(?=none)" );
+			ber_bvnone = BER_BVC( "(?=none)" ),
+			ber_bvF = BER_BVC( "(|)" ),
+			ber_bvT = BER_BVC( "(&)" );
 	ber_len_t	len;
 	ber_tag_t	choice;
-	int undef;
+	int undef, undef2;
 	char *sign;
 
 	if ( f == NULL ) {
@@ -562,6 +638,7 @@ filter2bv_x( Operation *op, Filter *f, struct berval *fstr )
 	}
 
 	undef = f->f_choice & SLAPD_FILTER_UNDEFINED;
+	undef2 = (undef && !noundef);
 	choice = f->f_choice & SLAPD_FILTER_MASK;
 
 	switch ( choice ) {
@@ -582,34 +659,48 @@ filter2bv_x( Operation *op, Filter *f, struct berval *fstr )
 		sign = "~=";
 
 simple:
-		filter_escape_value_x( &f->f_av_value, &tmp, op->o_tmpmemctx );
+		value = f->f_av_value;
+		if ( f->f_av_desc->ad_type->sat_equality &&
+			!undef &&
+			( f->f_av_desc->ad_type->sat_equality->smr_usage & SLAP_MR_MUTATION_NORMALIZER ))
+		{
+			f->f_av_desc->ad_type->sat_equality->smr_normalize(
+				(SLAP_MR_DENORMALIZE|SLAP_MR_VALUE_OF_ASSERTION_SYNTAX),
+				NULL, NULL, &f->f_av_value, &value, op->o_tmpmemctx );
+		}
+
+		filter_escape_value_x( &value, &tmp, op->o_tmpmemctx );
 		/* NOTE: tmp can legitimately be NULL (meaning empty) 
 		 * since in a Filter values in AVAs are supposed
 		 * to have been normalized, meaning that an empty value
 		 * is legal for that attribute's syntax */
 
 		fstr->bv_len += f->f_av_desc->ad_cname.bv_len + tmp.bv_len;
-		if ( undef )
+		if ( undef2 )
 			fstr->bv_len++;
 		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 1, op->o_tmpmemctx );
 
 		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s%s%s%s)",
-			undef ? "?" : "",
+			undef2 ? "?" : "",
 			f->f_av_desc->ad_cname.bv_val, sign,
 			tmp.bv_len ? tmp.bv_val : "" );
 
+		if ( value.bv_val != f->f_av_value.bv_val ) {
+			ber_memfree_x( value.bv_val, op->o_tmpmemctx );
+		}
+
 		ber_memfree_x( tmp.bv_val, op->o_tmpmemctx );
 		break;
 
 	case LDAP_FILTER_SUBSTRINGS:
 		fstr->bv_len = f->f_sub_desc->ad_cname.bv_len +
 			STRLENOF("(=*)");
-		if ( undef )
+		if ( undef2 )
 			fstr->bv_len++;
 		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 128, op->o_tmpmemctx );
 
 		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s%s=*)",
-			undef ? "?" : "",
+			undef2 ? "?" : "",
 			f->f_sub_desc->ad_cname.bv_val );
 
 		if ( f->f_sub_initial.bv_val != NULL ) {
@@ -618,16 +709,16 @@ simple:
 			len = fstr->bv_len;
 
 			filter_escape_value_x( &f->f_sub_initial, &tmp, op->o_tmpmemctx );
-			tmplen = tmp.bv_len ? tmp.bv_len : STRLENOF( "(null)" );
+			tmplen = tmp.bv_len;
 
 			fstr->bv_len += tmplen;
 			fstr->bv_val = op->o_tmprealloc( fstr->bv_val,
 				fstr->bv_len + 1, op->o_tmpmemctx );
 
-			snprintf( &fstr->bv_val[len-2],
+			snprintf( &fstr->bv_val[len - 2],
 				tmplen + STRLENOF( /*(*/ "*)" ) + 1,
 				/* "(attr=" */ "%s*)",
-				tmp.bv_len ? tmp.bv_val : "(null)");
+				tmp.bv_len ? tmp.bv_val : "");
 
 			ber_memfree_x( tmp.bv_val, op->o_tmpmemctx );
 		}
@@ -639,16 +730,16 @@ simple:
 				len = fstr->bv_len;
 				filter_escape_value_x( &f->f_sub_any[i],
 					&tmp, op->o_tmpmemctx );
-				tmplen = tmp.bv_len ? tmp.bv_len : STRLENOF( "(null)" );
+				tmplen = tmp.bv_len;
 
 				fstr->bv_len += tmplen + STRLENOF( /*(*/ ")" );
 				fstr->bv_val = op->o_tmprealloc( fstr->bv_val,
 					fstr->bv_len + 1, op->o_tmpmemctx );
 
-				snprintf( &fstr->bv_val[len-1],
+				snprintf( &fstr->bv_val[len - 1],
 					tmplen + STRLENOF( /*(*/ "*)" ) + 1,
 					/* "(attr=[init]*[any*]" */ "%s*)",
-					tmp.bv_len ? tmp.bv_val : "(null)");
+					tmp.bv_len ? tmp.bv_val : "");
 				ber_memfree_x( tmp.bv_val, op->o_tmpmemctx );
 			}
 		}
@@ -659,16 +750,16 @@ simple:
 			len = fstr->bv_len;
 
 			filter_escape_value_x( &f->f_sub_final, &tmp, op->o_tmpmemctx );
-			tmplen = tmp.bv_len ? tmp.bv_len : STRLENOF( "(null)" );
+			tmplen = tmp.bv_len;
 
 			fstr->bv_len += tmplen;
 			fstr->bv_val = op->o_tmprealloc( fstr->bv_val,
 				fstr->bv_len + 1, op->o_tmpmemctx );
 
-			snprintf( &fstr->bv_val[len-1],
+			snprintf( &fstr->bv_val[len - 1],
 				tmplen + STRLENOF( /*(*/ ")" ) + 1,
 				/* "(attr=[init*][any*]" */ "%s)",
-				tmp.bv_len ? tmp.bv_val : "(null)");
+				tmp.bv_len ? tmp.bv_val : "");
 
 			ber_memfree_x( tmp.bv_val, op->o_tmpmemctx );
 		}
@@ -678,13 +769,13 @@ simple:
 	case LDAP_FILTER_PRESENT:
 		fstr->bv_len = f->f_desc->ad_cname.bv_len +
 			STRLENOF("(=*)");
-		if ( undef )
+		if ( undef2 )
 			fstr->bv_len++;
 
 		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 1, op->o_tmpmemctx );
 
 		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s%s=*)",
-			undef ? "?" : "",
+			undef2 ? "?" : "",
 			f->f_desc->ad_cname.bv_val );
 		break;
 
@@ -701,7 +792,7 @@ simple:
 		for ( p = f->f_list; p != NULL; p = p->f_next ) {
 			len = fstr->bv_len;
 
-			filter2bv_x( op, p, &tmp );
+			filter2bv_undef_x( op, p, noundef, &tmp );
 			
 			fstr->bv_len += tmp.bv_len;
 			fstr->bv_val = op->o_tmprealloc( fstr->bv_val, fstr->bv_len + 1,
@@ -733,13 +824,14 @@ simple:
 		}
 		
 		fstr->bv_len = ad.bv_len +
+			( undef2 ? 1 : 0 ) +
 			( f->f_mr_dnattrs ? STRLENOF(":dn") : 0 ) +
-			( f->f_mr_rule_text.bv_len ? f->f_mr_rule_text.bv_len+1 : 0 ) +
+			( f->f_mr_rule_text.bv_len ? f->f_mr_rule_text.bv_len + STRLENOF(":") : 0 ) +
 			tmp.bv_len + STRLENOF("(:=)");
 		fstr->bv_val = op->o_tmpalloc( fstr->bv_len + 1, op->o_tmpmemctx );
 
 		snprintf( fstr->bv_val, fstr->bv_len + 1, "(%s%s%s%s%s:=%s)",
-			undef ? "?" : "",
+			undef2 ? "?" : "",
 			ad.bv_val,
 			f->f_mr_dnattrs ? ":dn" : "",
 			f->f_mr_rule_text.bv_len ? ":" : "",
@@ -751,11 +843,11 @@ simple:
 	case SLAPD_FILTER_COMPUTED:
 		switch ( f->f_result ) {
 		case LDAP_COMPARE_FALSE:
-			tmp = ber_bvfalse;
+			tmp = ( noundef ? ber_bvF : ber_bvfalse );
 			break;
 
 		case LDAP_COMPARE_TRUE:
-			tmp = ber_bvtrue;
+			tmp = ( noundef ? ber_bvT : ber_bvtrue );
 			break;
 			
 		case SLAPD_COMPARE_UNDEFINED:
@@ -778,6 +870,12 @@ simple:
 
 void
 filter2bv( Filter *f, struct berval *fstr )
+{
+	filter2bv_undef( f, 0, fstr );
+}
+
+void
+filter2bv_undef( Filter *f, int noundef, struct berval *fstr )
 {
 	Operation op;
 	Opheader ohdr;
@@ -786,7 +884,7 @@ filter2bv( Filter *f, struct berval *fstr )
 	op.o_tmpmemctx = NULL;
 	op.o_tmpmfuncs = &ch_mfuncs;
 
-	filter2bv_x( &op, f, fstr );
+	filter2bv_undef_x( &op, f, noundef, fstr );
 }
 
 Filter *
@@ -886,7 +984,7 @@ get_simple_vrFilter(
 	int		err;
 	ValuesReturnFilter vrf;
 
-	Debug( LDAP_DEBUG_FILTER, "begin get_simple_vrFilter\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "begin get_simple_vrFilter\n" );
 
 	tag = ber_peek_tag( ber, &len );
 
@@ -902,7 +1000,7 @@ get_simple_vrFilter(
 
 	switch ( vrf.vrf_choice ) {
 	case LDAP_FILTER_EQUALITY:
-		Debug( LDAP_DEBUG_FILTER, "EQUALITY\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "EQUALITY\n" );
 		err = get_ava( op, ber, (Filter *)&vrf, SLAP_MR_EQUALITY, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
@@ -912,12 +1010,12 @@ get_simple_vrFilter(
 		break;
 
 	case LDAP_FILTER_SUBSTRINGS:
-		Debug( LDAP_DEBUG_FILTER, "SUBSTRINGS\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "SUBSTRINGS\n" );
 		err = get_ssa( op, ber, (Filter *)&vrf, text );
 		break;
 
 	case LDAP_FILTER_GE:
-		Debug( LDAP_DEBUG_FILTER, "GE\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "GE\n" );
 		err = get_ava( op, ber, (Filter *)&vrf, SLAP_MR_ORDERING, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
@@ -925,7 +1023,7 @@ get_simple_vrFilter(
 		break;
 
 	case LDAP_FILTER_LE:
-		Debug( LDAP_DEBUG_FILTER, "LE\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "LE\n" );
 		err = get_ava( op, ber, (Filter *)&vrf, SLAP_MR_ORDERING, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
@@ -935,7 +1033,7 @@ get_simple_vrFilter(
 	case LDAP_FILTER_PRESENT: {
 		struct berval type;
 
-		Debug( LDAP_DEBUG_FILTER, "PRESENT\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "PRESENT\n" );
 		if ( ber_scanf( ber, "m", &type ) == LBER_ERROR ) {
 			err = SLAPD_DISCONNECT;
 			*text = "error decoding filter";
@@ -966,7 +1064,7 @@ get_simple_vrFilter(
 		} break;
 
 	case LDAP_FILTER_APPROX:
-		Debug( LDAP_DEBUG_FILTER, "APPROX\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "APPROX\n" );
 		err = get_ava( op, ber, (Filter *)&vrf, SLAP_MR_EQUALITY_APPROX, text );
 		if ( err != LDAP_SUCCESS ) {
 			break;
@@ -974,7 +1072,7 @@ get_simple_vrFilter(
 		break;
 
 	case LDAP_FILTER_EXT:
-		Debug( LDAP_DEBUG_FILTER, "EXTENSIBLE\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "EXTENSIBLE\n" );
 
 		err = get_mra( op, ber, (Filter *)&vrf, text );
 		if ( err != LDAP_SUCCESS ) {
@@ -987,7 +1085,7 @@ get_simple_vrFilter(
 	default:
 		(void) ber_scanf( ber, "x" ); /* skip the element */
 		Debug( LDAP_DEBUG_ANY, "get_simple_vrFilter: unknown filter type=%lu\n",
-			vrf.vrf_choice, 0, 0 );
+			vrf.vrf_choice );
 		vrf.vrf_choice = SLAPD_FILTER_COMPUTED;
 		vrf.vrf_result = SLAPD_COMPARE_UNDEFINED;
 		break;
@@ -1005,7 +1103,7 @@ get_simple_vrFilter(
 		**filt = vrf;
 	}
 
-	Debug( LDAP_DEBUG_FILTER, "end get_simple_vrFilter %d\n", err, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "end get_simple_vrFilter %d\n", err );
 
 	return err;
 }
@@ -1049,7 +1147,7 @@ get_vrFilter( Operation *op, BerElement *ber,
 	ber_len_t	len;
 	char		*last;
 
-	Debug( LDAP_DEBUG_FILTER, "begin get_vrFilter\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "begin get_vrFilter\n" );
 
 	tag = ber_peek_tag( ber, &len );
 
@@ -1076,21 +1174,17 @@ get_vrFilter( Operation *op, BerElement *ber,
 	}
 	*n = NULL;
 
-	Debug( LDAP_DEBUG_FILTER, "end get_vrFilter\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "end get_vrFilter\n" );
 	return( LDAP_SUCCESS );
 }
 
 void
 vrFilter_free( Operation *op, ValuesReturnFilter *vrf )
 {
-	ValuesReturnFilter	*p, *next;
+	ValuesReturnFilter	*next;
 
-	if ( vrf == NULL ) {
-		return;
-	}
-
-	for ( p = vrf; p != NULL; p = next ) {
-		next = p->vrf_next;
+	for ( ; vrf != NULL; vrf = next ) {
+		next = vrf->vrf_next;
 
 		switch ( vrf->vrf_choice & SLAPD_FILTER_MASK ) {
 		case LDAP_FILTER_PRESENT:
@@ -1123,7 +1217,7 @@ vrFilter_free( Operation *op, ValuesReturnFilter *vrf )
 
 		default:
 			Debug( LDAP_DEBUG_ANY, "filter_free: unknown filter type=%lu\n",
-				vrf->vrf_choice, 0, 0 );
+				vrf->vrf_choice );
 			break;
 		}
 
