@@ -1,8 +1,8 @@
 /* filterentry.c - apply a filter to an entry */
-/* $OpenLDAP: pkg/ldap/servers/slapd/filterentry.c,v 1.104.2.1 2007/02/13 20:22:18 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2007 The OpenLDAP Foundation.
+ * Copyright 1998-2020 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -64,10 +64,10 @@ test_filter(
     Filter	*f )
 {
 	int	rc;
-	Debug( LDAP_DEBUG_FILTER, "=> test_filter\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "=> test_filter\n" );
 
 	if ( f->f_choice & SLAPD_FILTER_UNDEFINED ) {
-		Debug( LDAP_DEBUG_FILTER, "    UNDEFINED\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "    UNDEFINED\n" );
 		rc = SLAPD_COMPARE_UNDEFINED;
 		goto out;
 	}
@@ -78,53 +78,53 @@ test_filter(
 			f->f_result == LDAP_COMPARE_FALSE ? "false" :
 			f->f_result == LDAP_COMPARE_TRUE ? "true" :
 			f->f_result == SLAPD_COMPARE_UNDEFINED ? "undefined" : "error",
-			f->f_result, 0 );
+			f->f_result );
 
 		rc = f->f_result;
 		break;
 
 	case LDAP_FILTER_EQUALITY:
-		Debug( LDAP_DEBUG_FILTER, "    EQUALITY\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "    EQUALITY\n" );
 		rc = test_ava_filter( op, e, f->f_ava, LDAP_FILTER_EQUALITY );
 		break;
 
 	case LDAP_FILTER_SUBSTRINGS:
-		Debug( LDAP_DEBUG_FILTER, "    SUBSTRINGS\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "    SUBSTRINGS\n" );
 		rc = test_substrings_filter( op, e, f );
 		break;
 
 	case LDAP_FILTER_GE:
-		Debug( LDAP_DEBUG_FILTER, "    GE\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "    GE\n" );
 		rc = test_ava_filter( op, e, f->f_ava, LDAP_FILTER_GE );
 		break;
 
 	case LDAP_FILTER_LE:
-		Debug( LDAP_DEBUG_FILTER, "    LE\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "    LE\n" );
 		rc = test_ava_filter( op, e, f->f_ava, LDAP_FILTER_LE );
 		break;
 
 	case LDAP_FILTER_PRESENT:
-		Debug( LDAP_DEBUG_FILTER, "    PRESENT\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "    PRESENT\n" );
 		rc = test_presence_filter( op, e, f->f_desc );
 		break;
 
 	case LDAP_FILTER_APPROX:
-		Debug( LDAP_DEBUG_FILTER, "    APPROX\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "    APPROX\n" );
 		rc = test_ava_filter( op, e, f->f_ava, LDAP_FILTER_APPROX );
 		break;
 
 	case LDAP_FILTER_AND:
-		Debug( LDAP_DEBUG_FILTER, "    AND\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "    AND\n" );
 		rc = test_filter_and( op, e, f->f_and );
 		break;
 
 	case LDAP_FILTER_OR:
-		Debug( LDAP_DEBUG_FILTER, "    OR\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "    OR\n" );
 		rc = test_filter_or( op, e, f->f_or );
 		break;
 
 	case LDAP_FILTER_NOT:
-		Debug( LDAP_DEBUG_FILTER, "    NOT\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "    NOT\n" );
 		rc = test_filter( op, e, f->f_not );
 
 		/* Flip true to false and false to true
@@ -141,17 +141,17 @@ test_filter(
 		break;
 
 	case LDAP_FILTER_EXT:
-		Debug( LDAP_DEBUG_FILTER, "    EXT\n", 0, 0, 0 );
+		Debug( LDAP_DEBUG_FILTER, "    EXT\n" );
 		rc = test_mra_filter( op, e, f->f_mra );
 		break;
 
 	default:
 		Debug( LDAP_DEBUG_ANY, "    unknown filter type %lu\n",
-		    f->f_choice, 0, 0 );
+		    f->f_choice );
 		rc = LDAP_PROTOCOL_ERROR;
 	}
 out:
-	Debug( LDAP_DEBUG_FILTER, "<= test_filter %d\n", rc, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "<= test_filter %d\n", rc );
 	return( rc );
 }
 
@@ -169,7 +169,7 @@ static int test_mra_filter(
 
 	if ( op == NULL ) {
 		memctx = NULL;
-		memfree = slap_sl_free;
+		memfree = slap_sl_mfuncs.bmf_free;
 	} else {
 		memctx = op->o_tmpmemctx;
 		memfree = op->o_tmpfree;
@@ -212,12 +212,7 @@ static int test_mra_filter(
 			if ( mra->ma_cf && mra->ma_rule->smr_usage & SLAP_MR_COMPONENT ) {
 				num_attr_vals = 0;
 				if ( !a->a_comp_data ) {
-					for ( ;
-						!BER_BVISNULL( &a->a_vals[num_attr_vals] );
-						num_attr_vals++ )
-					{
-						/* empty */;
-					}
+					num_attr_vals = a->a_numvals;
 					if ( num_attr_vals <= 0 ) {
 						/* no attribute value */
 						return LDAP_INAPPROPRIATE_MATCHING;
@@ -225,7 +220,7 @@ static int test_mra_filter(
 					num_attr_vals++;
 
 					/* following malloced will be freed by comp_tree_free () */
-					a->a_comp_data = malloc( sizeof( ComponentData ) +
+					a->a_comp_data = SLAP_MALLOC( sizeof( ComponentData ) +
 						sizeof( ComponentSyntaxInfo* )*num_attr_vals );
 
 					if ( !a->a_comp_data ) return LDAP_NO_MEMORY;
@@ -551,7 +546,7 @@ test_ava_filter(
 	if ( ava->aa_desc == slap_schema.si_ad_hasSubordinates 
 		&& op && op->o_bd && op->o_bd->be_has_subordinates )
 	{
-		int	hasSubordinates;
+		int	hasSubordinates = 0;
 		struct berval hs;
 
 		if( type != LDAP_FILTER_EQUALITY &&
@@ -583,7 +578,7 @@ test_ava_filter(
 
 	if ( ava->aa_desc == slap_schema.si_ad_entryDN ) {
 		MatchingRule *mr;
-		int rc, match;
+		int match;
 		const char *text;
 
 		if( type != LDAP_FILTER_EQUALITY &&
@@ -662,6 +657,56 @@ test_ava_filter(
 			continue;
 		}
 
+		/* We have no Sort optimization for Approx matches */
+		if (( a->a_flags & SLAP_ATTR_SORTED_VALS ) && type != LDAP_FILTER_APPROX ) {
+			unsigned slot;
+			int ret;
+
+			/* For Ordering matches, we just need to do one comparison with
+			 * either the first (least) or last (greatest) value.
+			 */
+			if ( use == SLAP_MR_ORDERING ) {
+				const char *text;
+				int match, which;
+				which = (type == LDAP_FILTER_LE) ? 0 : a->a_numvals-1;
+				ret = value_match( &match, a->a_desc, mr, use,
+					&a->a_nvals[which], &ava->aa_value, &text );
+				if ( ret != LDAP_SUCCESS ) return ret;
+				if (( type == LDAP_FILTER_LE && match <= 0 ) ||
+					( type == LDAP_FILTER_GE && match >= 0 ))
+					return LDAP_COMPARE_TRUE;
+				continue;
+			}
+			/* Only Equality will get here */
+			ret = attr_valfind( a, use | SLAP_MR_ASSERTED_VALUE_NORMALIZED_MATCH |
+				SLAP_MR_ATTRIBUTE_VALUE_NORMALIZED_MATCH, 
+				&ava->aa_value, &slot, NULL );
+			if ( ret == LDAP_SUCCESS )
+				return LDAP_COMPARE_TRUE;
+			else if ( ret != LDAP_NO_SUCH_ATTRIBUTE )
+				return ret;
+#if 0
+			/* The following is useful if we want to know which values
+			 * matched an ordering test. But here we don't care, we just
+			 * want to know if any value did, and that is checked above.
+			 */
+			if ( ret == LDAP_NO_SUCH_ATTRIBUTE ) {
+				/* If insertion point is not the end of the list, there was
+				 * at least one value greater than the assertion.
+				 */
+				if ( type == LDAP_FILTER_GE && slot < a->a_numvals )
+					return LDAP_COMPARE_TRUE;
+				/* Likewise, if insertion point is not the head of the list,
+				 * there was at least one value less than the assertion.
+				 */
+				if ( type == LDAP_FILTER_LE && slot > 0 )
+					return LDAP_COMPARE_TRUE;
+				return LDAP_COMPARE_FALSE;
+			}
+#endif
+			continue;
+		}
+
 #ifdef LDAP_COMP_MATCH
 		if ( nibble_mem_allocator && ava->aa_cf && !a->a_comp_data ) {
 			/* Component Matching */
@@ -671,7 +716,7 @@ test_ava_filter(
 			num_attr_vals++;/* for NULL termination */
 
 			/* following malloced will be freed by comp_tree_free () */
-			a->a_comp_data = malloc( sizeof( ComponentData ) + sizeof( ComponentSyntaxInfo* )*num_attr_vals );
+			a->a_comp_data = SLAP_MALLOC( sizeof( ComponentData ) + sizeof( ComponentSyntaxInfo* )*num_attr_vals );
 
 			if ( !a->a_comp_data ) {
 				return LDAP_NO_MEMORY;
@@ -729,7 +774,7 @@ test_ava_filter(
 			} else 
 #endif
 			{
-				ret = value_match( &match, a->a_desc, mr, use,
+				ret = ordered_value_match( &match, a->a_desc, mr, use,
 					bv, &ava->aa_value, &text );
 			}
 
@@ -828,7 +873,7 @@ test_filter_and(
 	Filter	*f;
 	int rtn = LDAP_COMPARE_TRUE; /* True if empty */
 
-	Debug( LDAP_DEBUG_FILTER, "=> test_filter_and\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "=> test_filter_and\n" );
 
 	for ( f = flist; f != NULL; f = f->f_next ) {
 		int rc = test_filter( op, e, f );
@@ -845,7 +890,7 @@ test_filter_and(
 		}
 	}
 
-	Debug( LDAP_DEBUG_FILTER, "<= test_filter_and %d\n", rtn, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "<= test_filter_and %d\n", rtn );
 
 	return rtn;
 }
@@ -859,7 +904,7 @@ test_filter_or(
 	Filter	*f;
 	int rtn = LDAP_COMPARE_FALSE; /* False if empty */
 
-	Debug( LDAP_DEBUG_FILTER, "=> test_filter_or\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "=> test_filter_or\n" );
 
 	for ( f = flist; f != NULL; f = f->f_next ) {
 		int rc = test_filter( op, e, f );
@@ -876,7 +921,7 @@ test_filter_or(
 		}
 	}
 
-	Debug( LDAP_DEBUG_FILTER, "<= test_filter_or %d\n", rtn, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "<= test_filter_or %d\n", rtn );
 	return rtn;
 }
 
@@ -890,7 +935,7 @@ test_substrings_filter(
 	Attribute	*a;
 	int rc;
 
-	Debug( LDAP_DEBUG_FILTER, "begin test_substrings_filter\n", 0, 0, 0 );
+	Debug( LDAP_DEBUG_FILTER, "begin test_substrings_filter\n" );
 
 	if ( !access_allowed( op, e,
 		f->f_sub_desc, NULL, ACL_SEARCH, NULL ) )
@@ -936,6 +981,6 @@ test_substrings_filter(
 	}
 
 	Debug( LDAP_DEBUG_FILTER, "end test_substrings_filter %d\n",
-		rc, 0, 0 );
+		rc );
 	return rc;
 }
