@@ -1,8 +1,8 @@
 /* oidm.c - object identifier macro routines */
-/* $OpenLDAP: pkg/ldap/servers/slapd/oidm.c,v 1.21.2.1 2007/02/13 20:22:19 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2007 The OpenLDAP Foundation.
+ * Copyright 1998-2020 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -26,7 +26,7 @@
 #include "lutil.h"
 #include "config.h"
 
-static LDAP_STAILQ_HEAD(OidMacroList, slap_oid_macro) om_list
+static LDAP_STAILQ_HEAD(OidMacroList, OidMacro) om_list
 	= LDAP_STAILQ_HEAD_INITIALIZER(om_list);
 
 OidMacro *om_sys_tail;
@@ -61,7 +61,7 @@ oidm_find(char *oid)
 					+ suflen + 1);
 				if( tmp == NULL ) {
 					Debug( LDAP_DEBUG_ANY,
-						"oidm_find: SLAP_MALLOC failed", 0, 0, 0 );
+						"oidm_find: SLAP_MALLOC failed" );
 					return NULL;
 				}
 				strcpy(tmp, om->som_oid.bv_val);
@@ -105,22 +105,22 @@ parse_oidm(
 
 	oidv = oidm_find( c->argv[2] );
 	if( !oidv ) {
-		snprintf( c->msg, sizeof( c->msg ),
+		snprintf( c->cr_msg, sizeof( c->cr_msg ),
 			"%s: OID %s not recognized",
 			c->argv[0], c->argv[2] );
 		Debug( LDAP_DEBUG_CONFIG|LDAP_DEBUG_NONE,
-			"%s %s\n", c->log, c->msg, 0 );
+			"%s %s\n", c->log, c->cr_msg );
 		return 1;
 	}
 
 	oid = oidm_find( c->argv[1] );
 	if( oid != NULL ) {
 		int rc;
-		snprintf( c->msg, sizeof( c->msg ),
+		snprintf( c->cr_msg, sizeof( c->cr_msg ),
 			"%s: \"%s\" previously defined \"%s\"",
 			c->argv[0], c->argv[1], oid );
 		Debug( LDAP_DEBUG_CONFIG|LDAP_DEBUG_NONE,
-			"%s %s\n", c->log, c->msg, 0 );
+			"%s %s\n", c->log, c->cr_msg );
 		/* Allow duplicate if the definition is identical */
 		rc = strcmp( oid, oidv ) != 0;
 		SLAP_FREE( oid );
@@ -131,10 +131,10 @@ parse_oidm(
 
 	om = (OidMacro *) SLAP_CALLOC( sizeof(OidMacro), 1 );
 	if( om == NULL ) {
-		snprintf( c->msg, sizeof( c->msg ),
+		snprintf( c->cr_msg, sizeof( c->cr_msg ),
 			"%s: SLAP_CALLOC failed", c->argv[0] );
 		Debug( LDAP_DEBUG_ANY,
-			"%s %s\n", c->log, c->msg, 0 );
+			"%s %s\n", c->log, c->cr_msg );
 		if ( oidv != c->argv[2] )
 			SLAP_FREE( oidv );
 		return 1;
