@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/overlays/dds.c,v 1.7.2.3 2007/02/13 20:22:29 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2005-2007 The OpenLDAP Foundation.
+ * Copyright 2005-2020 The OpenLDAP Foundation.
  * Portions Copyright 2005-2006 SysNet s.n.c.
  * All rights reserved.
  *
@@ -123,7 +123,7 @@ dds_expire( void *ctx, dds_info_t *di )
 	Connection	conn = { 0 };
 	OperationBuffer opbuf;
 	Operation	*op;
-	slap_callback	sc = { 0 }, sc2 = { 0 };
+	slap_callback	sc = { 0 };
 	dds_cb_t	dc = { 0 };
 	dds_expire_t	*de = NULL, **dep;
 	SlapReply	rs = { REP_RESULT };
@@ -134,13 +134,16 @@ dds_expire( void *ctx, dds_info_t *di )
 
 	int		ndeletes, ntotdeletes;
 
-	op = (Operation *)&opbuf;
-	connection_fake_init( &conn, op, ctx );
+	int		rc;
+	char		*extra = "";
+
+	connection_fake_init2( &conn, &opbuf, ctx, 0 );
+	op = &opbuf.ob_op;
 
 	op->o_tag = LDAP_REQ_SEARCH;
 	memset( &op->oq_search, 0, sizeof( op->oq_search ) );
 
-	op->o_bd = select_backend( &di->di_nsuffix[ 0 ], 0, 0 );
+	op->o_bd = select_backend( &di->di_nsuffix[ 0 ], 0 );
 
 	op->o_req_dn = op->o_bd->be_suffix[ 0 ];
 	op->o_req_ndn = op->o_bd->be_nsuffix[ 0 ];
@@ -153,7 +156,7 @@ dds_expire( void *ctx, dds_info_t *di )
 	op->ors_slimit = SLAP_NO_LIMIT;
 	op->ors_attrs = slap_anlist_no_attrs;
 
-	expire = slap_get_time() + di->di_tolerance;
+	expire = slap_get_time() - di->di_tolerance;
 	ts.bv_val = tsbuf;
 	ts.bv_len = sizeof( tsbuf );
 	slap_timestamp( &expire, &ts );
@@ -182,21 +185,30 @@ dds_expire( void *ctx, dds_info_t *di )
 
 done_search:;
 	op->o_tmpfree( op->ors_filterstr.bv_val, op->o_tmpmemctx );
-	filter_free_x( op, op->ors_filter );
+	filter_free_x( op, op->ors_filter, 1 );
 
-	if ( rs.sr_err != LDAP_SUCCESS ) {
-		Log1( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
-			"DDS expired objects lookup failed err=%d\n",
-			rs.sr_err );
+	rc = rs.sr_err;
+	switch ( rs.sr_err ) {
+	case LDAP_SUCCESS:
+		break;
+
+	case LDAP_NO_SUCH_OBJECT:
+		/* (ITS#5267) database not created yet? */
+		rs.sr_err = LDAP_SUCCESS;
+		extra = " (ignored)";
+		/* fallthru */
+
+	default:
+		Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+			"DDS expired objects lookup failed err=%d%s\n",
+			rc, extra );
 		goto done;
 	}
 
 	op->o_tag = LDAP_REQ_DELETE;
 	op->o_callback = &sc;
-	sc.sc_response = slap_replog_cb;
+	sc.sc_response = slap_null_cb;
 	sc.sc_private = NULL;
-	sc.sc_next = &sc2;
-	sc2.sc_response = slap_null_cb;
 
 	for ( ntotdeletes = 0, ndeletes = 1; dc.dc_ndnlist != NULL  && ndeletes > 0; ) {
 		ndeletes = 0;
@@ -209,14 +221,14 @@ done_search:;
 			(void)op->o_bd->bd_info->bi_op_delete( op, &rs );
 			switch ( rs.sr_err ) {
 			case LDAP_SUCCESS:
-				Log1( LDAP_DEBUG_STATS, LDAP_LEVEL_INFO,
+				Log( LDAP_DEBUG_STATS, LDAP_LEVEL_INFO,
 					"DDS dn=\"%s\" expired.\n",
 					de->de_ndn.bv_val );
 				ndeletes++;
 				break;
 
 			case LDAP_NOT_ALLOWED_ON_NONLEAF:
-				Log1( LDAP_DEBUG_ANY, LDAP_LEVEL_NOTICE,
+				Log( LDAP_DEBUG_ANY, LDAP_LEVEL_NOTICE,
 					"DDS dn=\"%s\" is non-leaf; "
 					"deferring.\n",
 					de->de_ndn.bv_val );
@@ -225,16 +237,15 @@ done_search:;
 				break;
 	
 			default:
-				Log2( LDAP_DEBUG_ANY, LDAP_LEVEL_NOTICE,
+				Log( LDAP_DEBUG_ANY, LDAP_LEVEL_NOTICE,
 					"DDS dn=\"%s\" err=%d; "
 					"deferring.\n",
 					de->de_ndn.bv_val, rs.sr_err );
 				break;
 			}
-	
+
 			if ( de != NULL ) {
 				*dep = de->de_next;
-				dep = &de->de_next;
 				op->o_tmpfree( de, op->o_tmpmemctx );
 			}
 		}
@@ -244,7 +255,7 @@ done_search:;
 
 	rs.sr_err = LDAP_SUCCESS;
 
-	Log1( LDAP_DEBUG_STATS, LDAP_LEVEL_INFO,
+	Log( LDAP_DEBUG_STATS, LDAP_LEVEL_INFO,
 		"DDS expired=%d\n", ntotdeletes );
 
 done:;
@@ -383,7 +394,7 @@ dds_op_add( Operation *op, SlapReply *rs )
 	/* handle dynamic object operational attr(s) */
 	if ( is_dynamicObject ) {
 		time_t		ttl, expire;
-		char		ttlbuf[] = "31557600";
+		char		ttlbuf[STRLENOF("31557600") + 1];
 		char		tsbuf[ LDAP_LUTIL_GENTIME_BUFSIZE ];
 		struct berval	bv;
 
@@ -402,10 +413,12 @@ dds_op_add( Operation *op, SlapReply *rs )
 
 		ttl = DDS_DEFAULT_TTL( di );
 
+		/* assert because should be checked at configure */
 		assert( ttl <= DDS_RF2589_MAX_TTL );
 
 		bv.bv_val = ttlbuf;
 		bv.bv_len = snprintf( ttlbuf, sizeof( ttlbuf ), "%ld", ttl );
+		assert( bv.bv_len < sizeof( ttlbuf ) );
 
 		/* FIXME: apparently, values in op->ora_e are malloc'ed
 		 * on the thread's slab; works fine by chance,
@@ -429,6 +442,7 @@ dds_op_add( Operation *op, SlapReply *rs )
 			sc->sc_response = dds_counter_cb;
 			sc->sc_private = di;
 			sc->sc_next = op->o_callback;
+			sc->sc_writewait = 0;
 
 			op->o_callback = sc;
 		}
@@ -463,6 +477,7 @@ dds_op_delete( Operation *op, SlapReply *rs )
 			sc->sc_cleanup = dds_freeit_cb;
 			sc->sc_response = dds_counter_cb;
 			sc->sc_private = di;
+			sc->sc_writewait = 0;
 			sc->sc_next = op->o_callback;
 	
 			op->o_callback = sc;
@@ -568,11 +583,13 @@ dds_op_modify( Operation *op, SlapReply *rs )
 			}
 
 		} else if ( mod->sml_desc == slap_schema.si_ad_entryTtl ) {
-			unsigned long	ttl;
+			unsigned long	uttl;
+			time_t		ttl;
 			int		rc;
 
 			switch ( mod->sml_op ) {
 			case LDAP_MOD_DELETE:
+			case SLAP_MOD_SOFTDEL: /* FIXME? */
 				if ( mod->sml_values != NULL ) {
 					if ( BER_BVISEMPTY( &bv_entryTtl ) 
 						|| !bvmatch( &bv_entryTtl, &mod->sml_values[ 0 ] ) )
@@ -597,8 +614,9 @@ dds_op_modify( Operation *op, SlapReply *rs )
 				entryTtl = -1;
 				/* fallthru */
 
-			case SLAP_MOD_SOFTADD: /* FIXME? */
 			case LDAP_MOD_ADD:
+			case SLAP_MOD_SOFTADD: /* FIXME? */
+			case SLAP_MOD_ADD_IF_NOT_PRESENT: /* FIXME? */
 				assert( mod->sml_values != NULL );
 				assert( BER_BVISNULL( &mod->sml_values[ 1 ] ) );
 
@@ -615,7 +633,8 @@ dds_op_modify( Operation *op, SlapReply *rs )
 					goto done;
 				}
 
-				rc = lutil_atoul( &ttl, mod->sml_values[ 0 ].bv_val );
+				rc = lutil_atoul( &uttl, mod->sml_values[ 0 ].bv_val );
+				ttl = (time_t)uttl;
 				assert( rc == 0 );
 				if ( ttl > DDS_RF2589_MAX_TTL ) {
 					rs->sr_err = LDAP_PROTOCOL_ERROR;
@@ -632,7 +651,7 @@ dds_op_modify( Operation *op, SlapReply *rs )
 					goto done;
 				}
 
-				entryTtl = (time_t)ttl;
+				entryTtl = ttl;
 				bv_entryTtl.bv_len = mod->sml_values[ 0 ].bv_len;
 				AC_MEMCPY( bv_entryTtl.bv_val, mod->sml_values[ 0 ].bv_val, bv_entryTtl.bv_len );
 				bv_entryTtl.bv_val[ bv_entryTtl.bv_len ] = '\0';
@@ -790,6 +809,7 @@ done:;
 			tmpmod->sml_op = LDAP_MOD_REPLACE;
 			value_add_one( &tmpmod->sml_values, &bv );
 			value_add_one( &tmpmod->sml_nvalues, &bv );
+			tmpmod->sml_numvals = 1;
 		}
 	}
 
@@ -859,6 +879,72 @@ dds_op_rename( Operation *op, SlapReply *rs )
 	return SLAP_CB_CONTINUE;
 }
 
+/* entryTtl update for client */
+static int
+dds_response( Operation *op, SlapReply *rs )
+{
+	slap_overinst	*on = (slap_overinst *)op->o_bd->bd_info;
+	dds_info_t	*di = on->on_bi.bi_private;
+	int		rc;
+
+	if ( !DDS_OFF( di )
+	     && rs->sr_type == REP_SEARCH
+	     && attr_find( rs->sr_entry->e_attrs, slap_schema.si_ad_entryTtl ) )
+	{
+		BerVarray		vals = NULL;
+		struct lutil_tm		tm;
+		struct lutil_timet	tt;
+		char			ttlbuf[STRLENOF("31557600") + 1];
+		struct berval		ttlvalue;
+		time_t			ttl;
+		int			len;
+
+		/* User already has access to entryTtl, skip ACL checks on
+		 * entryExpireTimestamp */
+		rc = backend_attribute( op, NULL, &rs->sr_entry->e_nname,
+			ad_entryExpireTimestamp, &vals, ACL_NONE );
+		if ( rc != LDAP_SUCCESS ) {
+			return rc;
+		}
+
+		assert( vals[0].bv_val[vals[0].bv_len] == '\0' );
+		if ( lutil_parsetime( vals[0].bv_val, &tm ) ) {
+			goto done;
+		}
+
+		lutil_tm2time( &tm, &tt );
+		ttl = tt.tt_sec - op->o_time;
+		ttl = (ttl < 0) ? 0 : ttl;
+		assert( ttl <= DDS_RF2589_MAX_TTL );
+
+		len = snprintf( ttlbuf, sizeof(ttlbuf), "%ld", ttl );
+		if ( len < 0 )
+		{
+			goto done;
+		}
+		ttlvalue.bv_val = ttlbuf;
+		ttlvalue.bv_len = len;
+
+		rs_entry2modifiable( op, rs, on );
+
+		if ( attr_delete( &rs->sr_entry->e_attrs,
+				slap_schema.si_ad_entryTtl ) )
+		{
+			goto done;
+		}
+		if ( attr_merge_normalize_one( rs->sr_entry,
+				slap_schema.si_ad_entryTtl,
+				&ttlvalue, op->o_tmpmemctx ) )
+		{
+			goto done;
+		}
+
+done:;
+		ber_bvarray_free_x( vals, op->o_tmpmemctx );
+	}
+	return SLAP_CB_CONTINUE;
+}
+
 static int
 slap_parse_refresh(
 	struct berval	*in,
@@ -894,14 +980,14 @@ slap_parse_refresh(
 	tag = ber_scanf( ber, "{" /*}*/ );
 
 	if ( tag == LBER_ERROR ) {
-		Log0( LDAP_DEBUG_TRACE, LDAP_LEVEL_ERR,
+		Log( LDAP_DEBUG_TRACE, LDAP_LEVEL_ERR,
 			"slap_parse_refresh: decoding error.\n" );
 		goto decoding_error;
 	}
 
 	tag = ber_peek_tag( ber, &len );
 	if ( tag != LDAP_TAG_EXOP_REFRESH_REQ_DN ) {
-		Log0( LDAP_DEBUG_TRACE, LDAP_LEVEL_ERR,
+		Log( LDAP_DEBUG_TRACE, LDAP_LEVEL_ERR,
 			"slap_parse_refresh: decoding error.\n" );
 		goto decoding_error;
 	}
@@ -911,7 +997,7 @@ slap_parse_refresh(
 
 		tag = ber_scanf( ber, "m", &dn );
 		if ( tag == LBER_ERROR ) {
-			Log0( LDAP_DEBUG_TRACE, LDAP_LEVEL_ERR,
+			Log( LDAP_DEBUG_TRACE, LDAP_LEVEL_ERR,
 				"slap_parse_refresh: DN parse failed.\n" );
 			goto decoding_error;
 		}
@@ -932,14 +1018,14 @@ slap_parse_refresh(
 	tag = ber_peek_tag( ber, &len );
 
 	if ( tag != LDAP_TAG_EXOP_REFRESH_REQ_TTL ) {
-		Log0( LDAP_DEBUG_TRACE, LDAP_LEVEL_ERR,
+		Log( LDAP_DEBUG_TRACE, LDAP_LEVEL_ERR,
 			"slap_parse_refresh: decoding error.\n" );
 		goto decoding_error;
 	}
 
 	tag = ber_scanf( ber, "i", &tmp );
 	if ( tag == LBER_ERROR ) {
-		Log0( LDAP_DEBUG_TRACE, LDAP_LEVEL_ERR,
+		Log( LDAP_DEBUG_TRACE, LDAP_LEVEL_ERR,
 			"slap_parse_refresh: TTL parse failed.\n" );
 		goto decoding_error;
 	}
@@ -950,9 +1036,9 @@ slap_parse_refresh(
 
 	tag = ber_peek_tag( ber, &len );
 
-	if ( len != 0 ) {
+	if ( tag != LBER_DEFAULT || len != 0 ) {
 decoding_error:;
-		Log1( LDAP_DEBUG_TRACE, LDAP_LEVEL_ERR,
+		Log( LDAP_DEBUG_TRACE, LDAP_LEVEL_ERR,
 			"slap_parse_refresh: decoding error, len=%ld\n",
 			(long)len );
 		rc = LDAP_PROTOCOL_ERROR;
@@ -989,10 +1075,9 @@ dds_op_extended( Operation *op, SlapReply *rs )
 		SlapReply	rs2 = { REP_RESULT };
 		Operation	op2 = *op;
 		slap_callback	sc = { 0 };
-		slap_callback	sc2 = { 0 };
 		Modifications	ttlmod = { { 0 } };
 		struct berval	ttlvalues[ 2 ];
-		char		ttlbuf[] = "31557600";
+		char		ttlbuf[STRLENOF("31557600") + 1];
 
 		rs->sr_err = slap_parse_refresh( op->ore_reqdata, NULL, &ttl,
 			&rs->sr_text, NULL );
@@ -1017,7 +1102,7 @@ dds_op_extended( Operation *op, SlapReply *rs )
 			ttl = di->di_min_ttl;
 		}
 
-		/* This does not apply to multi-master case */
+		/* This does not apply to multi-provider case */
 		if ( !( !SLAP_SINGLE_SHADOW( op->o_bd ) || be_isupdate( op ) ) ) {
 			/* we SHOULD return a referral in this case */
 			BerVarray defref = op->o_bd->be_update_refs
@@ -1082,9 +1167,7 @@ dds_op_extended( Operation *op, SlapReply *rs )
 		op2.o_bd = &db;
 		db.bd_info = (BackendInfo *)on->on_info;
 		op2.o_callback = &sc;
-		sc.sc_response = slap_replog_cb;
-		sc.sc_next = &sc2;
-		sc2.sc_response = slap_null_cb;
+		sc.sc_response = slap_null_cb;
 		op2.o_relax = SLAP_CONTROL_CRITICAL;
 		op2.orm_modlist = &ttlmod;
 
@@ -1092,6 +1175,7 @@ dds_op_extended( Operation *op, SlapReply *rs )
 		ttlmod.sml_flags = SLAP_MOD_MANAGING;
 		ttlmod.sml_desc = slap_schema.si_ad_entryTtl;
 		ttlmod.sml_values = ttlvalues;
+		ttlmod.sml_numvals = 1;
 		ttlvalues[ 0 ].bv_val = ttlbuf;
 		ttlvalues[ 0 ].bv_len = snprintf( ttlbuf, sizeof( ttlbuf ), "%ld", ttl );
 		BER_BVZERO( &ttlvalues[ 1 ] );
@@ -1108,26 +1192,24 @@ dds_op_extended( Operation *op, SlapReply *rs )
 			BerElementBuffer	berbuf;
 			BerElement		*ber = (BerElement *)&berbuf;
 
-			if ( rs->sr_err == LDAP_SUCCESS ) {
-				ber_init_w_nullc( ber, LBER_USE_DER );
+			ber_init_w_nullc( ber, LBER_USE_DER );
 
-				rc = ber_printf( ber, "{tiN}", LDAP_TAG_EXOP_REFRESH_RES_TTL, (int)ttl );
+			rc = ber_printf( ber, "{tiN}", LDAP_TAG_EXOP_REFRESH_RES_TTL, (int)ttl );
 
-				if ( rc < 0 ) {
-					rs->sr_err = LDAP_OTHER;
-					rs->sr_text = "internal error";
+			if ( rc < 0 ) {
+				rs->sr_err = LDAP_OTHER;
+				rs->sr_text = "internal error";
 
-				} else {
-					(void)ber_flatten( ber, &rs->sr_rspdata );
-					rs->sr_rspoid = ch_strdup( slap_EXOP_REFRESH.bv_val );
+			} else {
+				(void)ber_flatten( ber, &rs->sr_rspdata );
+				rs->sr_rspoid = ch_strdup( slap_EXOP_REFRESH.bv_val );
 
-					Log3( LDAP_DEBUG_TRACE, LDAP_LEVEL_INFO,
-						"%s REFRESH dn=\"%s\" TTL=%ld\n",
-						op->o_log_prefix, op->o_req_ndn.bv_val, ttl );
-				}
-
-				ber_free_buf( ber );
+				Log( LDAP_DEBUG_TRACE, LDAP_LEVEL_INFO,
+					"%s REFRESH dn=\"%s\" TTL=%ld\n",
+					op->o_log_prefix, op->o_req_ndn.bv_val, ttl );
 			}
+
+			ber_free_buf( ber );
 		}
 
 		return rs->sr_err;
@@ -1159,24 +1241,28 @@ static ConfigTable dds_cfg[] = {
 		2, 2, 0, ARG_MAGIC|ARG_ON_OFF|DDS_STATE, dds_cfgen,
 		"( OLcfgOvAt:9.1 NAME 'olcDDSstate' "
 			"DESC 'RFC2589 Dynamic directory services state' "
+			"EQUALITY booleanMatch "
 			"SYNTAX OMsBoolean "
 			"SINGLE-VALUE )", NULL, NULL },
 	{ "dds-max-ttl", "ttl",
 		2, 2, 0, ARG_MAGIC|DDS_MAXTTL, dds_cfgen,
 		"( OLcfgOvAt:9.2 NAME 'olcDDSmaxTtl' "
 			"DESC 'RFC2589 Dynamic directory services max TTL' "
+			"EQUALITY caseIgnoreMatch "
 			"SYNTAX OMsDirectoryString "
 			"SINGLE-VALUE )", NULL, NULL },
 	{ "dds-min-ttl", "ttl",
 		2, 2, 0, ARG_MAGIC|DDS_MINTTL, dds_cfgen,
 		"( OLcfgOvAt:9.3 NAME 'olcDDSminTtl' "
 			"DESC 'RFC2589 Dynamic directory services min TTL' "
+			"EQUALITY caseIgnoreMatch "
 			"SYNTAX OMsDirectoryString "
 			"SINGLE-VALUE )", NULL, NULL },
 	{ "dds-default-ttl", "ttl",
 		2, 2, 0, ARG_MAGIC|DDS_DEFAULTTTL, dds_cfgen,
 		"( OLcfgOvAt:9.4 NAME 'olcDDSdefaultTtl' "
 			"DESC 'RFC2589 Dynamic directory services default TTL' "
+			"EQUALITY caseIgnoreMatch "
 			"SYNTAX OMsDirectoryString "
 			"SINGLE-VALUE )", NULL, NULL },
 	{ "dds-interval", "interval",
@@ -1184,6 +1270,7 @@ static ConfigTable dds_cfg[] = {
 		"( OLcfgOvAt:9.5 NAME 'olcDDSinterval' "
 			"DESC 'RFC2589 Dynamic directory services expiration "
 				"task run interval' "
+			"EQUALITY caseIgnoreMatch "
 			"SYNTAX OMsDirectoryString "
 			"SINGLE-VALUE )", NULL, NULL },
 	{ "dds-tolerance", "ttl",
@@ -1191,12 +1278,14 @@ static ConfigTable dds_cfg[] = {
 		"( OLcfgOvAt:9.6 NAME 'olcDDStolerance' "
 			"DESC 'RFC2589 Dynamic directory services additional "
 				"TTL in expiration scheduling' "
+			"EQUALITY caseIgnoreMatch "
 			"SYNTAX OMsDirectoryString "
 			"SINGLE-VALUE )", NULL, NULL },
 	{ "dds-max-dynamicObjects", "num",
 		2, 2, 0, ARG_MAGIC|ARG_INT|DDS_MAXDYNAMICOBJS, dds_cfgen,
 		"( OLcfgOvAt:9.7 NAME 'olcDDSmaxDynamicObjects' "
 			"DESC 'RFC2589 Dynamic directory services max number of dynamic objects' "
+			"EQUALITY integerMatch "
 			"SYNTAX OMsInteger "
 			"SINGLE-VALUE )", NULL, NULL },
 	{ NULL, NULL, 0, 0, 0, ARG_IGNORED }
@@ -1368,20 +1457,20 @@ dds_cfgen( ConfigArgs *c )
 
 	case DDS_MAXTTL:
 		if ( lutil_parse_time( c->argv[ 1 ], &t ) != 0 ) {
-			snprintf( c->msg, sizeof( c->msg),
+			snprintf( c->cr_msg, sizeof( c->cr_msg),
 				"DDS unable to parse dds-max-ttl \"%s\"",
 				c->argv[ 1 ] );
-			Log2( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
-				"%s: %s.\n", c->log, c->msg );
+			Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+				"%s: %s.\n", c->log, c->cr_msg );
 			return 1;
 		}
 
 		if ( t < DDS_RF2589_DEFAULT_TTL || t > DDS_RF2589_MAX_TTL ) {
-			snprintf( c->msg, sizeof( c->msg ),
-				"DDS invalid dds-max-ttl=%ld; must be between %d and %d",
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
+				"DDS invalid dds-max-ttl=%lu; must be between %d and %d",
 				t, DDS_RF2589_DEFAULT_TTL, DDS_RF2589_MAX_TTL );
-			Log2( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
-				"%s: %s.\n", c->log, c->msg );
+			Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+				"%s: %s.\n", c->log, c->cr_msg );
 			return 1;
 		}
 
@@ -1390,20 +1479,20 @@ dds_cfgen( ConfigArgs *c )
 
 	case DDS_MINTTL:
 		if ( lutil_parse_time( c->argv[ 1 ], &t ) != 0 ) {
-			snprintf( c->msg, sizeof( c->msg),
+			snprintf( c->cr_msg, sizeof( c->cr_msg),
 				"DDS unable to parse dds-min-ttl \"%s\"",
 				c->argv[ 1 ] );
-			Log2( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
-				"%s: %s.\n", c->log, c->msg );
+			Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+				"%s: %s.\n", c->log, c->cr_msg );
 			return 1;
 		}
 
-		if ( t < 0 || t > DDS_RF2589_MAX_TTL ) {
-			snprintf( c->msg, sizeof( c->msg ),
-				"DDS invalid dds-min-ttl=%ld",
+		if ( t > DDS_RF2589_MAX_TTL ) {
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
+				"DDS invalid dds-min-ttl=%lu",
 				t );
-			Log2( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
-				"%s: %s.\n", c->log, c->msg );
+			Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+				"%s: %s.\n", c->log, c->cr_msg );
 			return 1;
 		}
 
@@ -1417,20 +1506,20 @@ dds_cfgen( ConfigArgs *c )
 
 	case DDS_DEFAULTTTL:
 		if ( lutil_parse_time( c->argv[ 1 ], &t ) != 0 ) {
-			snprintf( c->msg, sizeof( c->msg),
+			snprintf( c->cr_msg, sizeof( c->cr_msg),
 				"DDS unable to parse dds-default-ttl \"%s\"",
 				c->argv[ 1 ] );
-			Log2( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
-				"%s: %s.\n", c->log, c->msg );
+			Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+				"%s: %s.\n", c->log, c->cr_msg );
 			return 1;
 		}
 
-		if ( t < 0 || t > DDS_RF2589_MAX_TTL ) {
-			snprintf( c->msg, sizeof( c->msg ),
-				"DDS invalid dds-default-ttl=%ld",
+		if ( t > DDS_RF2589_MAX_TTL ) {
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
+				"DDS invalid dds-default-ttl=%lu",
 				t );
-			Log2( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
-				"%s: %s.\n", c->log, c->msg );
+			Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+				"%s: %s.\n", c->log, c->cr_msg );
 			return 1;
 		}
 
@@ -1444,25 +1533,25 @@ dds_cfgen( ConfigArgs *c )
 
 	case DDS_INTERVAL:
 		if ( lutil_parse_time( c->argv[ 1 ], &t ) != 0 ) {
-			snprintf( c->msg, sizeof( c->msg),
+			snprintf( c->cr_msg, sizeof( c->cr_msg),
 				"DDS unable to parse dds-interval \"%s\"",
 				c->argv[ 1 ] );
-			Log2( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
-				"%s: %s.\n", c->log, c->msg );
+			Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+				"%s: %s.\n", c->log, c->cr_msg );
 			return 1;
 		}
 
 		if ( t <= 0 ) {
-			snprintf( c->msg, sizeof( c->msg ),
-				"DDS invalid dds-interval=%ld",
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
+				"DDS invalid dds-interval=%lu",
 				t );
-			Log2( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
-				"%s: %s.\n", c->log, c->msg );
+			Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+				"%s: %s.\n", c->log, c->cr_msg );
 			return 1;
 		}
 
 		if ( t < 60 ) {
-			Log2( LDAP_DEBUG_ANY, LDAP_LEVEL_NOTICE,
+			Log( LDAP_DEBUG_ANY, LDAP_LEVEL_NOTICE,
 				"%s: dds-interval=%lu may be too small.\n",
 				c->log, t );
 		}
@@ -1481,20 +1570,20 @@ dds_cfgen( ConfigArgs *c )
 
 	case DDS_TOLERANCE:
 		if ( lutil_parse_time( c->argv[ 1 ], &t ) != 0 ) {
-			snprintf( c->msg, sizeof( c->msg),
+			snprintf( c->cr_msg, sizeof( c->cr_msg),
 				"DDS unable to parse dds-tolerance \"%s\"",
 				c->argv[ 1 ] );
-			Log2( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
-				"%s: %s.\n", c->log, c->msg );
+			Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+				"%s: %s.\n", c->log, c->cr_msg );
 			return 1;
 		}
 
-		if ( t < 0 || t > DDS_RF2589_MAX_TTL ) {
-			snprintf( c->msg, sizeof( c->msg ),
-				"DDS invalid dds-tolerance=%ld",
+		if ( t > DDS_RF2589_MAX_TTL ) {
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
+				"DDS invalid dds-tolerance=%lu",
 				t );
-			Log2( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
-				"%s: %s.\n", c->log, c->msg );
+			Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+				"%s: %s.\n", c->log, c->cr_msg );
 			return 1;
 		}
 
@@ -1503,10 +1592,10 @@ dds_cfgen( ConfigArgs *c )
 
 	case DDS_MAXDYNAMICOBJS:
 		if ( c->value_int < 0 ) {
-			snprintf( c->msg, sizeof( c->msg ),
+			snprintf( c->cr_msg, sizeof( c->cr_msg ),
 				"DDS invalid dds-max-dynamicObjects=%d", c->value_int );
-			Log2( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
-				"%s: %s.\n", c->log, c->msg );
+			Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+				"%s: %s.\n", c->log, c->cr_msg );
 			return 1;
 		}
 		di->di_max_dynamicObjects = c->value_int;
@@ -1522,14 +1611,15 @@ dds_cfgen( ConfigArgs *c )
 
 static int
 dds_db_init(
-	BackendDB	*be )
+	BackendDB	*be,
+	ConfigReply	*cr)
 {
 	slap_overinst	*on = (slap_overinst *)be->bd_info;
 	dds_info_t	*di;
 	BackendInfo	*bi = on->on_info->oi_orig;
 
 	if ( SLAP_ISGLOBALOVERLAY( be ) ) {
-		Log0( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+		Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
 			"DDS cannot be used as global overlay.\n" );
 		return 1;
 	}
@@ -1542,7 +1632,7 @@ dds_db_init(
 		|| bi->bi_op_search == NULL		/* object expiration */
 		|| bi->bi_entry_get_rw == NULL )	/* object type/existence checking */
 	{
-		Log1( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+		Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
 			"DDS backend \"%s\" does not provide "
 			"required functionality.\n",
 			bi->bi_type );
@@ -1611,8 +1701,11 @@ dds_count( void *ctx, BackendDB *be )
 	slap_callback	sc = { 0 };
 	SlapReply	rs = { REP_RESULT };
 
-	op = (Operation *)&opbuf;
-	connection_fake_init( &conn, op, ctx );
+	int		rc;
+	char		*extra = "";
+
+	connection_fake_init2( &conn, &opbuf, ctx, 0 );
+	op = &opbuf.ob_op;
 
 	op->o_tag = LDAP_REQ_SEARCH;
 	memset( &op->oq_search, 0, sizeof( op->oq_search ) );
@@ -1646,6 +1739,7 @@ dds_count( void *ctx, BackendDB *be )
 	op->o_callback = &sc;
 	sc.sc_response = dds_count_cb;
 	sc.sc_private = &di->di_num_dynamicObjects;
+	di->di_num_dynamicObjects = 0;
 
 	op->o_bd->bd_info = (BackendInfo *)on->on_info;
 	(void)op->o_bd->bd_info->bi_op_search( op, &rs );
@@ -1653,17 +1747,27 @@ dds_count( void *ctx, BackendDB *be )
 
 done_search:;
 	op->o_tmpfree( op->ors_filterstr.bv_val, op->o_tmpmemctx );
-	filter_free_x( op, op->ors_filter );
+	filter_free_x( op, op->ors_filter, 1 );
 
-	if ( rs.sr_err == LDAP_SUCCESS ) {
-		Log1( LDAP_DEBUG_STATS, LDAP_LEVEL_INFO,
+	rc = rs.sr_err;
+	switch ( rs.sr_err ) {
+	case LDAP_SUCCESS:
+		Log( LDAP_DEBUG_STATS, LDAP_LEVEL_INFO,
 			"DDS non-expired=%d\n",
 			di->di_num_dynamicObjects );
+		break;
 
-	} else {
-		Log1( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
-			"DDS non-expired objects lookup failed err=%d\n",
-			rs.sr_err );
+	case LDAP_NO_SUCH_OBJECT:
+		/* (ITS#5267) database not created yet? */
+		rs.sr_err = LDAP_SUCCESS;
+		extra = " (ignored)";
+		/* fallthru */
+
+	default:
+		Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+			"DDS non-expired objects lookup failed err=%d%s\n",
+			rc, extra );
+		break;
 	}
 
 	return rs.sr_err;
@@ -1671,19 +1775,23 @@ done_search:;
 
 static int
 dds_db_open(
-	BackendDB	*be )
+	BackendDB	*be,
+	ConfigReply	*cr )
 {
 	slap_overinst	*on = (slap_overinst *)be->bd_info;
 	dds_info_t	*di = on->on_bi.bi_private;
 	int		rc = 0;
 	void		*thrctx = ldap_pvt_thread_pool_context();
 
+	if ( slapMode & SLAP_TOOL_MODE )
+		return 0;
+
 	if ( DDS_OFF( di ) ) {
 		goto done;
 	}
 
 	if ( SLAP_SINGLE_SHADOW( be ) ) {
-		Log1( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+		Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
 			"DDS incompatible with shadow database \"%s\".\n",
 			be->be_suffix[ 0 ].bv_val );
 		return 1;
@@ -1700,22 +1808,11 @@ dds_db_open(
 	di->di_suffix = be->be_suffix;
 	di->di_nsuffix = be->be_nsuffix;
 
-	/* ... so that count, if required, is accurate */
-	if ( di->di_max_dynamicObjects > 0 ) {
-		/* force deletion of expired entries... */
-		be->bd_info = (BackendInfo *)on->on_info;
-		rc = dds_expire( thrctx, di );
-		be->bd_info = (BackendInfo *)on;
-		if ( rc != LDAP_SUCCESS ) {
-			rc = 1;
-			goto done;
-		}
-
-		rc = dds_count( thrctx, be );
-		if ( rc != LDAP_SUCCESS ) {
-			rc = 1;
-			goto done;
-		}
+	/* count the dynamic objects first */
+	rc = dds_count( thrctx, be );
+	if ( rc != LDAP_SUCCESS ) {
+		rc = 1;
+		goto done;
 	}
 
 	/* start expire task */
@@ -1736,7 +1833,8 @@ done:;
 
 static int
 dds_db_close(
-	BackendDB	*be )
+	BackendDB	*be,
+	ConfigReply	*cr )
 {
 	slap_overinst	*on = (slap_overinst *)be->bd_info;
 	dds_info_t	*di = on->on_bi.bi_private;
@@ -1749,6 +1847,7 @@ dds_db_close(
 		}
 		ldap_pvt_runqueue_remove( &slapd_rq, di->di_expire_task );
 		ldap_pvt_thread_mutex_unlock( &slapd_rq.rq_mutex );
+		di->di_expire_task = NULL;
 	}
 
 	(void)entry_info_unregister( dds_entry_info, (void *)di );
@@ -1758,7 +1857,8 @@ dds_db_close(
 
 static int
 dds_db_destroy(
-	BackendDB	*be )
+	BackendDB	*be,
+	ConfigReply	*cr )
 {
 	slap_overinst	*on = (slap_overinst *)be->bd_info;
 	dds_info_t	*di = on->on_bi.bi_private;
@@ -1785,12 +1885,18 @@ slap_exop_refresh(
 		return rs->sr_err;
 	}
 
-	Log2( LDAP_DEBUG_STATS, LDAP_LEVEL_INFO,
+	Log( LDAP_DEBUG_STATS, LDAP_LEVEL_INFO,
 		"%s REFRESH dn=\"%s\"\n",
 		op->o_log_prefix, op->o_req_ndn.bv_val );
 	op->o_req_dn = op->o_req_ndn;
 
-	op->o_bd = select_backend( &op->o_req_ndn, 0, 0 );
+	op->o_bd = select_backend( &op->o_req_ndn, 0 );
+	if ( op->o_bd == NULL ) {
+		send_ldap_error( op, rs, LDAP_NO_SUCH_OBJECT,
+			"no global superior knowledge" );
+		goto done;
+	}
+
 	if ( !SLAP_DYNAMIC( op->o_bd ) ) {
 		send_ldap_error( op, rs, LDAP_UNAVAILABLE_CRITICAL_EXTENSION,
 			"backend does not support dynamic directory services" );
@@ -1843,6 +1949,7 @@ dds_initialize()
 	if ( !do_not_load_schema ) {
 		static struct {
 			char			*desc;
+			slap_mask_t		flags;
 			AttributeDescription	**ad;
 		}		s_at[] = {
 			{ "( 1.3.6.1.4.1.4203.666.1.57 "
@@ -1855,6 +1962,7 @@ dds_initialize()
 				"SINGLE-VALUE "
 				"NO-USER-MODIFICATION "
 				"USAGE dSAOperation )",
+				SLAP_AT_HIDE,
 				&ad_entryExpireTimestamp },
 			{ NULL }
 		};
@@ -1863,9 +1971,10 @@ dds_initialize()
 			code = register_at( s_at[ i ].desc, s_at[ i ].ad, 0 );
 			if ( code ) {
 				Debug( LDAP_DEBUG_ANY,
-					"dds_initialize: register_at failed\n", 0, 0, 0 );
+					"dds_initialize: register_at failed\n" );
 				return code;
 			}
+			(*s_at[ i ].ad)->ad_type->sat_flags |= SLAP_AT_HIDE;
 		}
 	}
 
@@ -1874,7 +1983,7 @@ dds_initialize()
 			SLAP_EXOP_WRITES|SLAP_EXOP_HIDE, slap_exop_refresh,
 			!do_not_replace_exop );
 		if ( rc != LDAP_SUCCESS ) {
-			Log1( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+			Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
 				"DDS unable to register refresh exop: %d.\n",
 				rc );
 			return rc;
@@ -1883,6 +1992,7 @@ dds_initialize()
 
 	dds.on_bi.bi_type = "dds";
 
+	dds.on_bi.bi_flags = SLAPO_BFLAG_SINGLE;
 	dds.on_bi.bi_db_init = dds_db_init;
 	dds.on_bi.bi_db_open = dds_db_open;
 	dds.on_bi.bi_db_close = dds_db_close;
@@ -1893,6 +2003,7 @@ dds_initialize()
 	dds.on_bi.bi_op_modify = dds_op_modify;
 	dds.on_bi.bi_op_modrdn = dds_op_rename;
 	dds.on_bi.bi_extended = dds_op_extended;
+	dds.on_response = dds_response;
 
 	dds.on_bi.bi_cf_ocs = dds_ocs;
 
@@ -1929,7 +2040,7 @@ init_module( int argc, char *argv[] )
 			do_not_load_schema = no;
 
 		} else {
-			Log2( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
+			Log( LDAP_DEBUG_ANY, LDAP_LEVEL_ERR,
 				"DDS unknown module arg[#%d]=\"%s\".\n",
 				i, argv[ i ] );
 			return 1;
