@@ -1,8 +1,8 @@
 /* overlays.c - Static overlay framework */
-/* $OpenLDAP: pkg/ldap/servers/slapd/overlays/overlays.c,v 1.24.2.1 2007/02/13 20:22:29 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2003-2007 The OpenLDAP Foundation.
+ * Copyright 2003-2020 The OpenLDAP Foundation.
  * Copyright 2003 by Howard Chu.
  * All rights reserved.
  *
@@ -35,7 +35,7 @@ overlay_init(void)
 		if ( rc ) {
 			Debug( LDAP_DEBUG_ANY,
 				"%s overlay setup failed, err %d\n",
-				slap_oinfo[i].ov_type, rc, 0 );
+				slap_oinfo[i].ov_type, rc );
 			break;
 		}
 	}
