@@ -1,8 +1,8 @@
 /* root_dse.c - Provides the Root DSA-Specific Entry */
-/* $OpenLDAP: pkg/ldap/servers/slapd/root_dse.c,v 1.113.2.6 2007/02/13 20:22:20 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2007 The OpenLDAP Foundation.
+ * Copyright 1999-2020 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -216,7 +216,7 @@ root_dse_info(
 	e = entry_alloc();
 	if( e == NULL ) {
 		Debug( LDAP_DEBUG_ANY,
-			"root_dse_info: entry_alloc failed", 0, 0, 0 );
+			"root_dse_info: entry_alloc failed" );
 		return LDAP_OTHER;
 	}
 
@@ -254,6 +254,9 @@ fail:
 			/* no suffix! */
 			continue;
 		}
+		if ( SLAP_DBHIDDEN( be )) {
+			continue;
+		}
 		if ( SLAP_MONITOR( be )) {
 			if( attr_merge_one( e, ad_monitorContext,
 					&be->be_suffix[0],
@@ -277,8 +280,7 @@ fail:
 		}
 		for ( j = 0; be->be_suffix[j].bv_val != NULL; j++ ) {
 			if( attr_merge_one( e, ad_namingContexts,
-					&be->be_suffix[j],
-					&be->be_nsuffix[0] ) )
+					&be->be_suffix[j], NULL ) )
 			{
 				goto fail;
 			}
@@ -402,13 +404,14 @@ int
 root_dse_read_file( const char *fname )
 {
 	struct LDIFFP	*fp;
-	int rc = 0, lineno = 0, lmax = 0;
+	int rc = 0, lmax = 0, ldifrc;
+	unsigned long lineno = 0;
 	char	*buf = NULL;
 
 	if ( (fp = ldif_open( fname, "r" )) == NULL ) {
 		Debug( LDAP_DEBUG_ANY,
 			"root_dse_read_file: could not open rootdse attr file \"%s\" - absolute path?\n",
-			fname, 0, 0 );
+			fname );
 		perror( fname );
 		return EXIT_FAILURE;
 	}
@@ -416,21 +419,21 @@ root_dse_read_file( const char *fname )
 	usr_attr = entry_alloc();
 	if( usr_attr == NULL ) {
 		Debug( LDAP_DEBUG_ANY,
-			"root_dse_read_file: entry_alloc failed", 0, 0, 0 );
+			"root_dse_read_file: entry_alloc failed" );
 		ldif_close( fp );
 		return LDAP_OTHER;
 	}
 	usr_attr->e_attrs = NULL;
 
-	while( ldif_read_record( fp, &lineno, &buf, &lmax ) ) {
+	while(( ldifrc = ldif_read_record( fp, &lineno, &buf, &lmax )) > 0 ) {
 		Entry *e = str2entry( buf );
 		Attribute *a;
 
 		if( e == NULL ) {
 			Debug( LDAP_DEBUG_ANY, "root_dse_read_file: "
-				"could not parse entry (file=\"%s\" line=%d)\n",
-				fname, lineno, 0 );
-			rc = EXIT_FAILURE;
+				"could not parse entry (file=\"%s\" line=%lu)\n",
+				fname, lineno );
+			rc = LDAP_OTHER;
 			break;
 		}
 
@@ -438,10 +441,10 @@ root_dse_read_file( const char *fname )
 		if( e->e_nname.bv_len ) {
 			Debug( LDAP_DEBUG_ANY,
 				"root_dse_read_file: invalid rootDSE "
-				"- dn=\"%s\" (file=\"%s\" line=%d)\n",
+				"- dn=\"%s\" (file=\"%s\" line=%lu)\n",
 				e->e_dn, fname, lineno );
 			entry_free( e );
-			rc = EXIT_FAILURE;
+			rc = LDAP_OTHER;
 			break;
 		}
 
@@ -464,6 +467,9 @@ root_dse_read_file( const char *fname )
 		if (rc) break;
 	}
 
+	if ( ldifrc < 0 )
+		rc = LDAP_OTHER;
+
 	if (rc) {
 		entry_free( usr_attr );
 		usr_attr = NULL;
@@ -473,7 +479,7 @@ root_dse_read_file( const char *fname )
 
 	ldif_close( fp );
 
-	Debug(LDAP_DEBUG_CONFIG, "rootDSE file=\"%s\" read.\n", fname, 0, 0);
+	Debug(LDAP_DEBUG_CONFIG, "rootDSE file=\"%s\" read.\n", fname );
 	return rc;
 }
 
