@@ -1,8 +1,8 @@
 /* schema_check.c - routines to enforce schema definitions */
-/* $OpenLDAP: pkg/ldap/servers/slapd/schema_check.c,v 1.103.2.1 2007/02/13 20:22:21 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2007 The OpenLDAP Foundation.
+ * Copyright 1998-2020 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -32,6 +32,7 @@ static char * oc_check_required(
 static int entry_naming_check(
 	Entry *e,
 	int manage,
+	int add_naming,
 	const char** text,
 	char *textbuf, size_t textlen );
 /*
@@ -47,7 +48,8 @@ entry_schema_check(
 	Entry *e,
 	Attribute *oldattrs,
 	int manage,
-	int add_soc,
+	int add,
+	Attribute **socp,
 	const char** text,
 	char *textbuf, size_t textlen )
 {
@@ -116,7 +118,7 @@ entry_schema_check(
 
 			Debug( LDAP_DEBUG_ANY,
 			    "Entry (%s), %s\n",
-			    e->e_dn, textbuf, 0 );
+			    e->e_dn, textbuf );
 
 			return LDAP_CONSTRAINT_VIOLATION;
 		}
@@ -125,7 +127,7 @@ entry_schema_check(
 	/* check the object class attribute */
 	if ( aoc == NULL ) {
 		Debug( LDAP_DEBUG_ANY, "No objectClass for entry (%s)\n",
-		    e->e_dn, 0, 0 );
+		    e->e_dn );
 
 		*text = "no objectClass attribute";
 		return LDAP_OBJECT_CLASS_VIOLATION;
@@ -135,10 +137,10 @@ entry_schema_check(
 	assert( aoc->a_vals[0].bv_val != NULL );
 
 	/* check the structural object class attribute */
-	if ( asc == NULL && !add_soc ) {
+	if ( asc == NULL && !add ) {
 		Debug( LDAP_DEBUG_ANY,
 			"No structuralObjectClass for entry (%s)\n",
-		    e->e_dn, 0, 0 );
+		    e->e_dn );
 
 		*text = "no structuralObjectClass operational attribute";
 		return LDAP_OTHER;
@@ -150,7 +152,7 @@ entry_schema_check(
 		return rc;
 	}
 
-	if ( asc == NULL && add_soc ) {
+	if ( asc == NULL && add ) {
 		attr_merge_one( e, ad_structuralObjectClass, &oc->soc_cname, NULL );
 		asc = attr_find( e->e_attrs, ad_structuralObjectClass );
 		sc = oc;
@@ -169,10 +171,10 @@ entry_schema_check(
 
 		Debug( LDAP_DEBUG_ANY,
 			"entry_check_schema(%s): %s\n",
-			e->e_dn, textbuf, 0 );
+			e->e_dn, textbuf );
 
 		rc = LDAP_OBJECT_CLASS_VIOLATION;
-		goto leave;
+		goto done;
 	}
 
 	if( sc->soc_kind != LDAP_SCHEMA_STRUCTURAL ) {
@@ -182,10 +184,10 @@ entry_schema_check(
 
 		Debug( LDAP_DEBUG_ANY,
 			"entry_check_schema(%s): %s\n",
-			e->e_dn, textbuf, 0 );
+			e->e_dn, textbuf );
 
 		rc = LDAP_OTHER;
-		goto leave;
+		goto done;
 	}
 
 got_soc:
@@ -196,10 +198,10 @@ got_soc:
 
 		Debug( LDAP_DEBUG_ANY,
 			"entry_check_schema(%s): %s\n",
-			e->e_dn, textbuf, 0 );
+			e->e_dn, textbuf );
 
 		rc = LDAP_OBJECT_CLASS_VIOLATION;
-		goto leave;
+		goto done;
 	}
 
 	*text = textbuf;
@@ -209,24 +211,37 @@ got_soc:
 			"unrecognized objectClass '%s'",
 			aoc->a_vals[0].bv_val );
 		rc = LDAP_OBJECT_CLASS_VIOLATION;
-		goto leave;
+		goto done;
 
-	} else if ( sc != slap_schema.si_oc_glue && sc != oc ) {
-		snprintf( textbuf, textlen, 
-			"structural object class modification "
-			"from '%s' to '%s' not allowed",
-			asc->a_vals[0].bv_val, oc->soc_cname.bv_val );
-		rc = LDAP_NO_OBJECT_CLASS_MODS;
-		goto leave;
-	} else if ( sc == slap_schema.si_oc_glue ) {
+	} else if ( sc != oc ) {
+		if ( !manage && sc != slap_schema.si_oc_glue ) {
+			snprintf( textbuf, textlen, 
+				"structural object class modification "
+				"from '%s' to '%s' not allowed",
+				asc->a_vals[0].bv_val, oc->soc_cname.bv_val );
+			rc = LDAP_NO_OBJECT_CLASS_MODS;
+			goto done;
+		}
+
+		assert( asc->a_vals != NULL );
+		assert( !BER_BVISNULL( &asc->a_vals[0] ) );
+		assert( BER_BVISNULL( &asc->a_vals[1] ) );
+		assert( asc->a_nvals == asc->a_vals );
+
+		/* draft-zeilenga-ldap-relax: automatically modify
+		 * structuralObjectClass if changed with relax */
 		sc = oc;
+		ber_bvreplace( &asc->a_vals[ 0 ], &sc->soc_cname );
+		if ( socp ) {
+			*socp = asc;
+		}
 	}
 
 	/* naming check */
 	if ( !is_entry_glue ( e ) ) {
-		rc = entry_naming_check( e, manage, text, textbuf, textlen );
+		rc = entry_naming_check( e, manage, add, text, textbuf, textlen );
 		if( rc != LDAP_SUCCESS ) {
-			goto leave;
+			goto done;
 		}
 	} else {
 		/* Glue Entry */
@@ -247,10 +262,10 @@ got_soc:
 
 			Debug( LDAP_DEBUG_ANY,
 				"Entry (%s): %s\n",
-				e->e_dn, textbuf, 0 );
+				e->e_dn, textbuf );
 
 			rc = LDAP_OBJECT_CLASS_VIOLATION;
-			goto leave;
+			goto done;
 		}
 
 		if( cr->scr_required ) for( i=0; cr->scr_required[i]; i++ ) {
@@ -271,10 +286,10 @@ got_soc:
 
 				Debug( LDAP_DEBUG_ANY,
 					"Entry (%s): %s\n",
-					e->e_dn, textbuf, 0 );
+					e->e_dn, textbuf );
 
 				rc = LDAP_OBJECT_CLASS_VIOLATION;
-				goto leave;
+				goto done;
 			}
 		}
 
@@ -296,10 +311,10 @@ got_soc:
 
 				Debug( LDAP_DEBUG_ANY,
 					"Entry (%s): %s\n",
-					e->e_dn, textbuf, 0 );
+					e->e_dn, textbuf );
 
 				rc = LDAP_OBJECT_CLASS_VIOLATION;
-				goto leave;
+				goto done;
 			}
 		}
 	}
@@ -315,17 +330,17 @@ got_soc:
 
 			Debug( LDAP_DEBUG_ANY,
 				"entry_check_schema(%s): %s\n",
-				e->e_dn, textbuf, 0 );
+				e->e_dn, textbuf );
 
 			rc = LDAP_OBJECT_CLASS_VIOLATION;
-			goto leave;
+			goto done;
 		}
 
 		if ( oc->soc_check ) {
 			rc = (oc->soc_check)( op->o_bd, e, oc,
 				text, textbuf, textlen );
 			if( rc != LDAP_SUCCESS ) {
-				goto leave;
+				goto done;
 			}
 		}
 
@@ -355,17 +370,17 @@ got_soc:
 					}
 				}
 
-				if( xc == NULL ) {
-					snprintf( textbuf, textlen, "instanstantiation of "
+				if( xc != NULL ) {
+					snprintf( textbuf, textlen, "instantiation of "
 						"abstract objectClass '%s' not allowed",
 						aoc->a_vals[i].bv_val );
 
 					Debug( LDAP_DEBUG_ANY,
 						"entry_check_schema(%s): %s\n",
-						e->e_dn, textbuf, 0 );
+						e->e_dn, textbuf );
 
 					rc = LDAP_OBJECT_CLASS_VIOLATION;
-					goto leave;
+					goto done;
 				}
 			}
 
@@ -405,10 +420,10 @@ got_soc:
 				if( k == -1 ) {
 					Debug( LDAP_DEBUG_ANY,
 						"Entry (%s): %s\n",
-						e->e_dn, textbuf, 0 );
+						e->e_dn, textbuf );
 
 					rc = LDAP_OBJECT_CLASS_VIOLATION;
-					goto leave;
+					goto done;
 				}
 			}
 
@@ -420,10 +435,10 @@ got_soc:
 
 				Debug( LDAP_DEBUG_ANY,
 					"Entry (%s): %s\n",
-					e->e_dn, textbuf, 0 );
+					e->e_dn, textbuf );
 
 				rc = LDAP_OBJECT_CLASS_VIOLATION;
-				goto leave;
+				goto done;
 			}
 
 			if( oc == slap_schema.si_oc_extensibleObject ) {
@@ -435,7 +450,7 @@ got_soc:
 	if( extensible ) {
 		*text = NULL;
 		rc = LDAP_SUCCESS;
-		goto leave;
+		goto done;
 	}
 
 	/* check that each attr in the entry is allowed by some oc */
@@ -474,14 +489,14 @@ got_soc:
 
 			Debug( LDAP_DEBUG_ANY,
 			    "Entry (%s), %s\n",
-			    e->e_dn, textbuf, 0 );
+			    e->e_dn, textbuf );
 
-			goto leave;
+			goto done;
 		}
 	}
 
 	*text = NULL;
-leave:
+done:
 	slap_sl_free( socs, op->o_tmpmemctx );
 	return rc;
 }
@@ -498,7 +513,7 @@ oc_check_required(
 
 	Debug( LDAP_DEBUG_TRACE,
 		"oc_check_required entry (%s), objectClass \"%s\"\n",
-		e->e_dn, ocname->bv_val, 0 );
+		e->e_dn?e->e_dn:"(null)", ocname->bv_val );
 
 
 	/* check for empty oc_required */
@@ -533,7 +548,7 @@ int oc_check_allowed(
 
 	Debug( LDAP_DEBUG_TRACE,
 		"oc_check_allowed type \"%s\"\n",
-		at->sat_cname.bv_val, 0, 0 );
+		at->sat_cname.bv_val );
 
 	/* always allow objectClass attribute */
 	if ( strcasecmp( at->sat_cname.bv_val, "objectClass" ) == 0 ) {
@@ -572,6 +587,10 @@ int oc_check_allowed(
 	for ( i = 0; socs[i]; i++ ) {
 		/* if we know about the oc */
 		ObjectClass	*oc = socs[i];
+		/* extensibleObject allows all */
+		if ( oc == slap_schema.si_oc_extensibleObject ) {
+			return LDAP_SUCCESS;
+		}
 		if ( oc != NULL && oc->soc_kind != LDAP_SCHEMA_ABSTRACT &&
 			( sc == NULL || oc->soc_kind == LDAP_SCHEMA_AUXILIARY ))
 		{
@@ -758,6 +777,7 @@ static int
 entry_naming_check(
 	Entry *e,
 	int manage,
+	int add_naming,
 	const char** text,
 	char *textbuf, size_t textlen )
 {
@@ -777,7 +797,7 @@ entry_naming_check(
 	if ( ldap_bv2rdn( &e->e_name, &rdn, (char **)&p,
 		LDAP_DN_FORMAT_LDAP ) )
 	{
-		*text = "unrecongized attribute type(s) in RDN";
+		*text = "unrecognized attribute type(s) in RDN";
 		return LDAP_INVALID_DN_SYNTAX;
 	}
 
@@ -788,12 +808,14 @@ entry_naming_check(
 		AttributeDescription *desc = NULL;
 		Attribute *attr;
 		const char *errtext;
+		int add = 0;
 
 		if( ava->la_flags & LDAP_AVA_BINARY ) {
 			snprintf( textbuf, textlen, 
 				"value of naming attribute '%s' in unsupported BER form",
 				ava->la_attr.bv_val );
 			rc = LDAP_NAMING_VIOLATION;
+			break;
 		}
 
 		rc = slap_bv2ad( &ava->la_attr, &desc, &errtext );
@@ -848,37 +870,64 @@ entry_naming_check(
 			snprintf( textbuf, textlen, 
 				"naming attribute '%s' is not present in entry",
 				ava->la_attr.bv_val );
-			rc = LDAP_NAMING_VIOLATION;
-			break;
+			if ( add_naming ) {
+				add = 1;
+
+			} else {
+				rc = LDAP_NAMING_VIOLATION;
+			}
+
+		} else {
+			rc = attr_valfind( attr, SLAP_MR_VALUE_OF_ASSERTION_SYNTAX|
+				SLAP_MR_ATTRIBUTE_VALUE_NORMALIZED_MATCH,
+				&ava->la_value, NULL, NULL );
+
+			if ( rc != 0 ) {
+				switch( rc ) {
+				case LDAP_INAPPROPRIATE_MATCHING:
+					snprintf( textbuf, textlen, 
+						"inappropriate matching for naming attribute '%s'",
+						ava->la_attr.bv_val );
+					break;
+				case LDAP_INVALID_SYNTAX:
+					snprintf( textbuf, textlen, 
+						"value of naming attribute '%s' is invalid",
+						ava->la_attr.bv_val );
+					break;
+				case LDAP_NO_SUCH_ATTRIBUTE:
+					if ( add_naming ) {
+						if ( is_at_single_value( desc->ad_type ) ) {
+							snprintf( textbuf, textlen, 
+								"value of single-valued naming attribute '%s' conflicts with value present in entry",
+								ava->la_attr.bv_val );
+
+						} else {
+							add = 1;
+							rc = LDAP_SUCCESS;
+						}
+
+					} else {
+						snprintf( textbuf, textlen, 
+							"value of naming attribute '%s' is not present in entry",
+							ava->la_attr.bv_val );
+					}
+					break;
+				default:
+					snprintf( textbuf, textlen, 
+						"naming attribute '%s' is inappropriate",
+						ava->la_attr.bv_val );
+				}
+
+				if ( !add ) {
+					rc = LDAP_NAMING_VIOLATION;
+				}
+			}
 		}
 
-		rc = value_find_ex( desc, SLAP_MR_VALUE_OF_ASSERTION_SYNTAX|
-			SLAP_MR_ATTRIBUTE_VALUE_NORMALIZED_MATCH,
-			attr->a_nvals, &ava->la_value, NULL );
+		if ( add ) {
+			attr_merge_normalize_one( e, desc, &ava->la_value, NULL );
 
-		if( rc != 0 ) {
-			switch( rc ) {
-			case LDAP_INAPPROPRIATE_MATCHING:
-				snprintf( textbuf, textlen, 
-					"inappropriate matching for naming attribute '%s'",
-					ava->la_attr.bv_val );
-				break;
-			case LDAP_INVALID_SYNTAX:
-				snprintf( textbuf, textlen, 
-					"value of naming attribute '%s' is invalid",
-					ava->la_attr.bv_val );
-				break;
-			case LDAP_NO_SUCH_ATTRIBUTE:
-				snprintf( textbuf, textlen, 
-					"value of naming attribute '%s' is not present in entry",
-					ava->la_attr.bv_val );
-				break;
-			default:
-				snprintf( textbuf, textlen, 
-					"naming attribute '%s' is inappropriate",
-					ava->la_attr.bv_val );
-			}
-			rc = LDAP_NAMING_VIOLATION;
+		} else if ( rc != LDAP_SUCCESS ) {
 			break;
 		}
 	}
