@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/search.c,v 1.181.2.2 2007/02/13 20:22:21 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1998-2007 The OpenLDAP Foundation.
+ * Copyright 1998-2020 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -41,8 +41,8 @@ do_search(
 	struct berval base = BER_BVNULL;
 	ber_len_t	siz, off, i;
 
-	Debug( LDAP_DEBUG_TRACE, "do_search\n", 0, 0, 0 );
-
+	Debug( LDAP_DEBUG_TRACE, "%s do_search\n",
+		op->o_log_prefix );
 	/*
 	 * Parse the search request.  It looks like this:
 	 *
@@ -112,8 +112,8 @@ do_search(
 
 	rs->sr_err = dnPrettyNormal( NULL, &base, &op->o_req_dn, &op->o_req_ndn, op->o_tmpmemctx );
 	if( rs->sr_err != LDAP_SUCCESS ) {
-		Debug( LDAP_DEBUG_ANY,
-			"do_search: invalid dn (%s)\n", base.bv_val, 0, 0 );
+		Debug( LDAP_DEBUG_ANY, "%s do_search: invalid dn: \"%s\"\n",
+			op->o_log_prefix, base.bv_val );
 		send_ldap_error( op, rs, LDAP_INVALID_DN_SYNTAX, "invalid DN" );
 		goto return_results;
 	}
@@ -138,7 +138,7 @@ do_search(
 	filter2bv_x( op, op->ors_filter, &op->ors_filterstr );
 	
 	Debug( LDAP_DEBUG_ARGS, "    filter: %s\n",
-		!BER_BVISEMPTY( &op->ors_filterstr ) ? op->ors_filterstr.bv_val : "empty", 0, 0 );
+		!BER_BVISEMPTY( &op->ors_filterstr ) ? op->ors_filterstr.bv_val : "empty" );
 
 	/* attributes */
 	siz = sizeof(AttributeName);
@@ -152,41 +152,71 @@ do_search(
 		const char *dummy;	/* ignore msgs from bv2ad */
 		op->ors_attrs[i].an_desc = NULL;
 		op->ors_attrs[i].an_oc = NULL;
-		op->ors_attrs[i].an_oc_exclude = 0;
+		op->ors_attrs[i].an_flags = 0;
 		if ( slap_bv2ad( &op->ors_attrs[i].an_name,
 			&op->ors_attrs[i].an_desc, &dummy ) != LDAP_SUCCESS )
 		{
-			slap_bv2undef_ad( &op->ors_attrs[i].an_name,
+			if ( slap_bv2undef_ad( &op->ors_attrs[i].an_name,
 				&op->ors_attrs[i].an_desc, &dummy,
-				SLAP_AD_PROXIED|SLAP_AD_NOINSERT );
-		};
-	}
+				SLAP_AD_PROXIED|SLAP_AD_NOINSERT ) )
+			{
+				struct berval *bv = &op->ors_attrs[i].an_name;
 
-	if( get_ctrls( op, rs, 1 ) != LDAP_SUCCESS ) {
-		Debug( LDAP_DEBUG_ANY, "do_search: get_ctrls failed\n", 0, 0, 0 );
+				/* RFC 4511 LDAPv3: All User Attributes */
+				if ( bvmatch( bv, slap_bv_all_user_attrs ) ) {
+					continue;
+				}
 
-		goto return_results;
-	}
+				/* RFC 3673 LDAPv3: All Operational Attributes */
+				if ( bvmatch( bv, slap_bv_all_operational_attrs ) ) {
+					continue;
+				}
 
-	Debug( LDAP_DEBUG_ARGS, "    attrs:", 0, 0, 0 );
+				/* RFC 4529 LDAP: Requesting Attributes by Object Class */
+				if ( bv->bv_len > 1 && bv->bv_val[0] == '@' ) {
+					/* FIXME: check if remaining is valid oc name? */
+					continue;
+				}
 
-	if ( siz != 0 ) {
-		for ( i = 0; i<siz; i++ ) {
-			Debug( LDAP_DEBUG_ARGS, " %s", op->ors_attrs[i].an_name.bv_val, 0, 0 );
+				/* add more "exceptions" to RFC 4511 4.5.1.8. */
+
+				/* invalid attribute description? remove */
+				if ( ad_keystring( bv ) ) {
+					/* NOTE: parsed in-place, don't modify;
+					 * rather add "1.1", which must be ignored */
+					BER_BVSTR( &op->ors_attrs[i].an_name, LDAP_NO_ATTRS );
+				}
+
+				/* otherwise leave in place... */
+			}
 		}
 	}
 
-	Debug( LDAP_DEBUG_ARGS, "\n", 0, 0, 0 );
+	if( get_ctrls( op, rs, 1 ) != LDAP_SUCCESS ) {
+		Debug( LDAP_DEBUG_ANY, "%s do_search: get_ctrls failed\n",
+			op->o_log_prefix );
+		goto return_results;
+	}
 
-	if ( StatslogTest( LDAP_DEBUG_STATS ) ) {
+	Debug( LDAP_DEBUG_ARGS, "    attrs:" );
+
+	if ( siz != 0 ) {
+		for ( i = 0; i<siz; i++ ) {
+			Debug( LDAP_DEBUG_ARGS, " %s", op->ors_attrs[i].an_name.bv_val );
+		}
+	}
+
+	Debug( LDAP_DEBUG_ARGS, "\n" );
+
+	if (LogTest( LDAP_DEBUG_STATS ) ) {
 		char abuf[BUFSIZ/2], *ptr = abuf;
-		int len = 0, alen;
+		unsigned len = 0, alen;
 
 		sprintf(abuf, "scope=%d deref=%d", op->ors_scope, op->ors_deref);
-		Statslog( LDAP_DEBUG_STATS,
+		Debug( LDAP_DEBUG_STATS,
 		        "%s SRCH base=\"%s\" %s filter=\"%s\"\n",
 		        op->o_log_prefix, op->o_req_dn.bv_val, abuf,
-		        op->ors_filterstr.bv_val, 0 );
+		        op->ors_filterstr.bv_val );
 
 		for ( i = 0; i<siz; i++ ) {
 			alen = op->ors_attrs[i].an_name.bv_len;
@@ -194,8 +224,8 @@ do_search(
 				alen = sizeof(abuf)-1;
 			}
 			if (len && (len + 1 + alen >= sizeof(abuf))) {
-				Statslog( LDAP_DEBUG_STATS, "%s SRCH attr=%s\n",
-				    op->o_log_prefix, abuf, 0, 0, 0 );
+				Debug( LDAP_DEBUG_STATS, "%s SRCH attr=%s\n",
+				    op->o_log_prefix, abuf );
 				len = 0;
 				ptr = abuf;
 			}
@@ -208,13 +238,17 @@ do_search(
 			*ptr = '\0';
 		}
 		if (len) {
-			Statslog( LDAP_DEBUG_STATS, "%s SRCH attr=%s\n",
-	    			op->o_log_prefix, abuf, 0, 0, 0 );
+			Debug( LDAP_DEBUG_STATS, "%s SRCH attr=%s\n",
+					op->o_log_prefix, abuf );
 		}
 	}
 
 	op->o_bd = frontendDB;
 	rs->sr_err = frontendDB->be_search( op, rs );
+	if ( rs->sr_err == SLAPD_ASYNCOP ) {
+		/* skip cleanup */
+		return rs->sr_err;
+	}
 
 return_results:;
 	if ( !BER_BVISNULL( &op->o_req_dn ) ) {
@@ -227,7 +261,7 @@ return_results:;
 		op->o_tmpfree( op->ors_filterstr.bv_val, op->o_tmpmemctx );
 	}
 	if ( op->ors_filter != NULL) {
-		filter_free_x( op, op->ors_filter );
+		filter_free_x( op, op->ors_filter, 1 );
 	}
 	if ( op->ors_attrs != NULL ) {
 		op->o_tmpfree( op->ors_attrs, op->o_tmpmemctx );
@@ -239,14 +273,9 @@ return_results:;
 int
 fe_op_search( Operation *op, SlapReply *rs )
 {
-	int			manageDSAit;
-	int			be_manageDSAit;
 	BackendDB		*bd = op->o_bd;
 
-	manageDSAit = get_manageDSAit( op );
-
-	/* fake while loop to allow breaking out */
-	while ( op->ors_scope == LDAP_SCOPE_BASE ) {
+	if ( op->ors_scope == LDAP_SCOPE_BASE ) {
 		Entry *entry = NULL;
 
 		if ( BER_BVISEMPTY( &op->o_req_ndn ) ) {
@@ -279,6 +308,12 @@ fe_op_search( Operation *op, SlapReply *rs )
 			goto return_results;
 
 		} else if ( entry != NULL ) {
+			if ( get_assert( op ) &&
+				( test_filter( op, entry, get_assertion( op )) != LDAP_COMPARE_TRUE )) {
+				rs->sr_err = LDAP_ASSERTION_FAILED;
+				goto fail1;
+			}
+
 			rs->sr_err = test_filter( op, entry, op->ors_filter );
 
 			if( rs->sr_err == LDAP_COMPARE_TRUE ) {
@@ -297,13 +332,12 @@ fe_op_search( Operation *op, SlapReply *rs )
 				rs->sr_entry = NULL;
 				rs->sr_operational_attrs = NULL;
 			}
-			entry_free( entry );
-
 			rs->sr_err = LDAP_SUCCESS;
+fail1:
+			entry_free( entry );
 			send_ldap_result( op, rs );
 			goto return_results;
 		}
-		break;
 	}
 
 	if( BER_BVISEMPTY( &op->o_req_ndn ) && !BER_BVISEMPTY( &default_search_nbase ) ) {
@@ -320,9 +354,7 @@ fe_op_search( Operation *op, SlapReply *rs )
 	 * if we don't hold it.
 	 */
 
-	be_manageDSAit = manageDSAit;
-
-	op->o_bd = select_backend( &op->o_req_ndn, be_manageDSAit, 1 );
+	op->o_bd = select_backend( &op->o_req_ndn, 1 );
 	if ( op->o_bd == NULL ) {
 		rs->sr_ref = referral_rewrite( default_referral,
 			NULL, &op->o_req_dn, op->ors_scope );
