@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/servers/slapd/slapi/plugin.c,v 1.43.2.1 2007/02/13 20:22:31 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2002-2007 The OpenLDAP Foundation.
+ * Copyright 2002-2020 The OpenLDAP Foundation.
  * Portions Copyright 1997,2002-2003 IBM Corporation.
  * All rights reserved.
  *
@@ -21,10 +21,11 @@
  */
 
 #include "portable.h"
-#include <ldap_pvt_thread.h>
-#include <slap.h>
-#include <slapi.h>
-#include <lutil.h>
+#include "ldap_pvt_thread.h"
+#include "slap.h"
+#include "config.h"
+#include "slapi.h"
+#include "lutil.h"
 
 /*
  * Note: if ltdl.h is not available, slapi should not be compiled
@@ -54,7 +55,7 @@ static ExtendedOp *pGExtendedOps = NULL;
  *
  * Output:             
  *
- * Return Values:      a pointer to a newly created Slapi_PBlock structrue or
+ * Return Values:      a pointer to a newly created Slapi_PBlock structure or
  *                     NULL - function failed 
  *
  * Messages:           None
@@ -128,7 +129,7 @@ done:
 /*********************************************************************
  * Function Name:      slapi_int_register_plugin
  *
- * Description:        insert the slapi_pblock structure to the end of the plugin
+ * Description:        insert the slapi_pblock structure to a given position the end of the plugin
  *                     list 
  *
  * Input:              a pointer to a plugin slapi_pblock structure to be added to 
@@ -142,21 +143,23 @@ done:
  * Messages:           None
  *********************************************************************/
 int 
-slapi_int_register_plugin(
+slapi_int_register_plugin_index(
 	Backend *be, 
-	Slapi_PBlock *pPB )
+	Slapi_PBlock *pPB,
+	int index )
 { 
 	Slapi_PBlock	*pTmpPB;
 	Slapi_PBlock	*pSavePB;
-	int   		 rc = LDAP_SUCCESS;
+	int		pos = 0, rc = LDAP_SUCCESS;
 
 	assert( be != NULL );
 
 	pTmpPB = SLAPI_BACKEND_PBLOCK( be );
-	if ( pTmpPB == NULL ) {
+	if ( pTmpPB == NULL || index == 0 ) {
 		SLAPI_BACKEND_PBLOCK( be ) = pPB;
 	} else {
-		while ( pTmpPB != NULL && rc == LDAP_SUCCESS ) {
+		while ( pTmpPB != NULL && rc == LDAP_SUCCESS &&
+				( index < 0 || pos++ < index ) ) {
 			pSavePB = pTmpPB;
 			rc = slapi_pblock_get( pTmpPB, SLAPI_IBM_PBLOCK, &pTmpPB );
 		}
@@ -165,10 +168,22 @@ slapi_int_register_plugin(
 			rc = slapi_pblock_set( pSavePB, SLAPI_IBM_PBLOCK, (void *)pPB ); 
 		}
 	}
+
+	if ( index >= 0 && rc == LDAP_SUCCESS ) {
+		rc = slapi_pblock_set( pPB, SLAPI_IBM_PBLOCK, (void *)pTmpPB );
+	}
      
 	return ( rc != LDAP_SUCCESS ) ? LDAP_OTHER : LDAP_SUCCESS;
 }
-       
+
+int
+slapi_int_register_plugin(
+	Backend *be,
+	Slapi_PBlock *pPB )
+{
+	return slapi_int_register_plugin_index( be, pPB, -1 );
+}
+
 /*********************************************************************
  * Function Name:      slapi_int_get_plugins
  *
@@ -449,7 +464,7 @@ error_return:
  *
  * Output:             pFuncAddr - the function address of the requested function name.
  *
- * Return Values:      a pointer to a newly created ExtendOp structrue or
+ * Return Values:      a pointer to a newly created ExtendOp structure or
  *                     NULL - function failed
  *
  * Messages:           None
@@ -598,7 +613,7 @@ slapi_int_call_plugins(
 	rc = slapi_int_get_plugins( be, funcType, &tmpPlugin );
 	if ( rc != LDAP_SUCCESS || tmpPlugin == NULL ) {
 		/* Nothing to do, front-end should ignore. */
-		return 1;
+		return rc;
 	}
 
 	for ( pGetPlugin = tmpPlugin ; *pGetPlugin != NULL; pGetPlugin++ ) {
@@ -629,7 +644,8 @@ slapi_int_read_config(
 	const char	*fname, 
 	int		lineno, 
 	int		argc, 
-	char		**argv )
+	char		**argv,
+	int		index )
 {
 	int		iType = -1;
 	int		numPluginArgc = 0;
@@ -645,8 +661,10 @@ slapi_int_read_config(
 
 	/* automatically instantiate overlay if necessary */
 	if ( !slapi_over_is_inst( be ) ) {
-		if ( slapi_over_config( be ) != 0 ) {
-			fprintf( stderr, "Failed to instantiate SLAPI overlay\n");
+		ConfigReply cr = { 0 };
+		if ( slapi_over_config( be, &cr ) != 0 ) {
+			fprintf( stderr, "Failed to instantiate SLAPI overlay: "
+				"err=%d msg=\"%s\"\n", cr.err, cr.msg );
 			return -1;
 		}
 	}
@@ -687,7 +705,7 @@ slapi_int_read_config(
 			}
 		}
 
-		rc = slapi_int_register_plugin( be, pPlugin );
+		rc = slapi_int_register_plugin_index( be, pPlugin, index );
 		if ( rc != LDAP_SUCCESS ) {
 			if ( iType == SLAPI_PLUGIN_EXTENDEDOP ) {
 				slapi_int_unregister_extop( be, &pGExtendedOps, pPlugin );
@@ -700,6 +718,74 @@ slapi_int_read_config(
 	return 0;
 }
 
+int
+slapi_int_unregister_plugin(
+	Backend *be,
+	Slapi_PBlock *pPlugin,
+	Slapi_PBlock *pPrev
+)
+{
+	int type;
+
+	assert( pPlugin != NULL );
+
+	slapi_pblock_get( pPlugin, SLAPI_PLUGIN_TYPE, (void *)&type );
+	if ( type == SLAPI_PLUGIN_EXTENDEDOP ) {
+		slapi_int_unregister_extop( be, &pGExtendedOps, pPlugin );
+	}
+
+	if ( pPrev != NULL ) {
+		Slapi_PBlock *pNext = NULL;
+
+		slapi_pblock_get( pPlugin, SLAPI_IBM_PBLOCK, &pNext );
+		slapi_pblock_set( pPrev, SLAPI_IBM_PBLOCK, &pNext );
+	}
+	slapi_pblock_destroy( pPlugin );
+
+	return LDAP_SUCCESS;
+}
+
+int
+slapi_int_unregister_plugins(
+	Backend *be,
+	int index
+)
+{
+	Slapi_PBlock	*pTmpPB = NULL;
+	Slapi_PBlock	*pSavePB = NULL;
+	int rc = LDAP_SUCCESS;
+
+	pTmpPB = SLAPI_BACKEND_PBLOCK( be );
+	if ( pTmpPB == NULL ) {
+		return ( index < 0 ) ? LDAP_SUCCESS : LDAP_OTHER;
+	}
+
+	if ( index < 0 ) {
+		/* All plugins must go */
+		while ( pTmpPB != NULL && rc == LDAP_SUCCESS ) {
+			pSavePB = pTmpPB;
+			rc = slapi_pblock_get( pTmpPB, SLAPI_IBM_PBLOCK, &pTmpPB );
+			if ( pSavePB != NULL ) {
+				slapi_int_unregister_plugin( be, pSavePB, NULL );
+			}
+		}
+	} else if ( index == 0 ) {
+		slapi_pblock_get( pTmpPB, SLAPI_IBM_PBLOCK, &pSavePB );
+		SLAPI_BACKEND_PBLOCK( be ) = pSavePB;
+		slapi_int_unregister_plugin( be, pTmpPB, NULL );
+	} else {
+		int pos = -1;
+		while ( pTmpPB != NULL && rc == LDAP_SUCCESS && ++pos < index ) {
+			pSavePB = pTmpPB;
+			rc = slapi_pblock_get( pTmpPB, SLAPI_IBM_PBLOCK, &pTmpPB );
+		}
+		if ( pos == index ) {
+			slapi_int_unregister_plugin( be, pTmpPB, pSavePB );
+		}
+	}
+	return rc;
+}
+
 void
 slapi_int_plugin_unparse(
 	Backend *be,
@@ -722,7 +808,11 @@ slapi_int_plugin_unparse(
 		slapi_pblock_get( pp, SLAPI_X_CONFIG_ARGV, &argv );
 		if ( argv == NULL ) /* could be dynamic plugin */
 			continue;
-		idx.bv_len = sprintf( idx.bv_val, "{%d}", i );
+		idx.bv_len = snprintf( idx.bv_val, sizeof( ibuf ), "{%d}", i );
+		if ( idx.bv_len >= sizeof( ibuf ) ) {
+			/* FIXME: just truncating by now */
+			idx.bv_len = sizeof( ibuf ) - 1;
+		}
 		bv.bv_len = idx.bv_len;
 		for (j=1; argv[j]; j++) {
 			bv.bv_len += strlen(argv[j]);
