@@ -1,8 +1,8 @@
 /* slapi_overlay.c - SLAPI overlay */
-/* $OpenLDAP: pkg/ldap/servers/slapd/slapi/slapi_overlay.c,v 1.40.2.4 2007/02/13 20:22:32 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2001-2007 The OpenLDAP Foundation.
+ * Copyright 2001-2020 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -27,6 +27,7 @@
 
 #include "slap.h"
 #include "slapi.h"
+#include "config.h"
 
 #ifdef LDAP_SLAPI
 
@@ -81,6 +82,7 @@ slapi_op_internal_p( Operation *op, SlapReply *rs, slap_callback *cb )
 		cb->sc_response = slapi_over_response;
 		cb->sc_cleanup = slapi_over_cleanup;
 		cb->sc_private = pb;
+		cb->sc_writewait = 0;
 		cb->sc_next = op->o_callback;
 		op->o_callback = cb;
 	}
@@ -275,13 +277,13 @@ slapi_op_bind_callback( Operation *op, SlapReply *rs, int prc )
 			ldap_pvt_thread_mutex_unlock( &op->o_conn->c_mutex );
 
 			/* log authorization identity */
-			Statslog( LDAP_DEBUG_STATS,
+			Debug( LDAP_DEBUG_STATS,
 				"%s BIND dn=\"%s\" mech=%s (SLAPI) ssf=0\n",
 				op->o_log_prefix,
 				BER_BVISNULL( &op->o_conn->c_dn )
 					? "<empty>" : op->o_conn->c_dn.bv_val,
-				BER_BVISNULL( &op->orb_tmp_mech )
-					? "<empty>" : op->orb_tmp_mech.bv_val, 0, 0 );
+				BER_BVISNULL( &op->orb_mech )
+					? "<empty>" : op->orb_mech.bv_val );
 
 			return -1;
 		}
@@ -295,6 +297,7 @@ static int
 slapi_op_search_callback( Operation *op, SlapReply *rs, int prc )
 {
 	Slapi_PBlock		*pb = SLAPI_OPERATION_PBLOCK( op );
+	Filter *f = op->ors_filter;
 
 	/* check preoperation result code */
 	if ( prc < 0 ) {
@@ -309,8 +312,10 @@ slapi_op_search_callback( Operation *op, SlapReply *rs, int prc )
 		 * The plugin can set the SLAPI_SEARCH_FILTER.
 		 * SLAPI_SEARCH_STRFILER is not normative.
 		 */
-		op->o_tmpfree( op->ors_filterstr.bv_val, op->o_tmpmemctx );
-		filter2bv_x( op, op->ors_filter, &op->ors_filterstr );
+		if (f != op->ors_filter) {
+			op->o_tmpfree( op->ors_filterstr.bv_val, op->o_tmpmemctx );
+			filter2bv_x( op, op->ors_filter, &op->ors_filterstr );
+		}
 	}
 
 	return LDAP_SUCCESS;
@@ -450,11 +455,11 @@ slapi_over_merge_controls( Operation *op, SlapReply *rs )
 	n_slapi_ctrls = slapi_int_count_controls( slapi_ctrls );
 	n_rs_ctrls = slapi_int_count_controls( rs->sr_ctrls );
 
-	slapi_pblock_set( pb, SLAPI_X_OLD_RESCONTROLS, (void *)rs->sr_ctrls );
-
 	if ( n_slapi_ctrls == 0 )
 		return LDAP_SUCCESS; /* no SLAPI controls */
 
+	slapi_pblock_set( pb, SLAPI_X_OLD_RESCONTROLS, (void *)rs->sr_ctrls );
+
 	ctrls = (LDAPControl **) op->o_tmpalloc(
 		( n_slapi_ctrls + n_rs_ctrls + 1 ) * sizeof(LDAPControl *),
 		op->o_tmpmemctx );
@@ -730,6 +735,7 @@ slapi_over_access_allowed(
 
 	cb.sc_response = NULL;
 	cb.sc_cleanup = NULL;
+	cb.sc_writewait = NULL;
 
 	pb = SLAPI_OPERATION_PBLOCK( op );
 
@@ -763,7 +769,7 @@ slapi_over_acl_group(
 	GroupAssertion		*g;
 	SlapReply		rs = { REP_RESULT };
 
-	op->o_bd = select_backend( gr_ndn, 0, 0 );
+	op->o_bd = select_backend( gr_ndn, 0 );
 
 	for ( g = op->o_groups; g; g = g->ga_next ) {
 		if ( g->ga_be != op->o_bd || g->ga_oc != group_oc ||
@@ -794,6 +800,7 @@ slapi_over_acl_group(
 
 		cb.sc_response = NULL;
 		cb.sc_cleanup = NULL;
+		cb.sc_writewait = NULL;
 
 		pb = SLAPI_OPERATION_PBLOCK( op );
 
@@ -848,7 +855,9 @@ done:
 }
 
 static int
-slapi_over_db_open( BackendDB *be )
+slapi_over_db_open(
+	BackendDB	*be,
+	ConfigReply	*cr )
 {
 	Slapi_PBlock		*pb;
 	int			rc;
@@ -863,7 +872,9 @@ slapi_over_db_open( BackendDB *be )
 }
 
 static int
-slapi_over_db_close( BackendDB *be )
+slapi_over_db_close(
+	BackendDB	*be,
+	ConfigReply	*cr )
 {
 	Slapi_PBlock		*pb;
 	int			rc;
@@ -911,12 +922,12 @@ int slapi_over_is_inst( BackendDB *be )
 	return overlay_is_inst( be, SLAPI_OVERLAY_NAME );
 }
 
-int slapi_over_config( BackendDB *be )
+int slapi_over_config( BackendDB *be, ConfigReply *cr )
 {
 	if ( slapi_over_initialized == 0 ) {
 		int rc;
 
-		/* do global initializaiton */
+		/* do global initialization */
 		ldap_pvt_thread_mutex_init( &slapi_hn_mutex );
 		ldap_pvt_thread_mutex_init( &slapi_time_mutex );
 		ldap_pvt_thread_mutex_init( &slapi_printmessage_mutex );
@@ -935,7 +946,7 @@ int slapi_over_config( BackendDB *be )
 		slapi_over_initialized = 1;
 	}
 
-	return overlay_config( be, SLAPI_OVERLAY_NAME, -1, NULL );
+	return overlay_config( be, SLAPI_OVERLAY_NAME, -1, NULL, cr );
 }
 
 #endif /* LDAP_SLAPI */
