@@ -1,6 +1,7 @@
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 2004-2007 The OpenLDAP Foundation.
+ * Copyright 2004-2020 The OpenLDAP Foundation.
  * Portions Copyright 2004 Pierangelo Masarati.
  * All rights reserved.
  *
@@ -37,10 +38,15 @@
 
 #include "slapcommon.h"
 
+#ifndef S_IWRITE
+#define S_IWRITE	S_IWUSR
+#endif
+
 static int
 test_file( const char *fname, const char *ftype )
 {
 	struct stat	st;
+	char ebuf[128];
 	int		save_errno;
 
 	switch ( stat( fname, &st ) ) {
@@ -48,7 +54,7 @@ test_file( const char *fname, const char *ftype )
 		if ( !( st.st_mode & S_IWRITE ) ) {
 			Debug( LDAP_DEBUG_ANY, "%s file "
 				"\"%s\" exists, but user does not have access\n",
-				ftype, fname, 0 );
+				ftype, fname );
 			return -1;
 		}
 		break;
@@ -65,7 +71,7 @@ test_file( const char *fname, const char *ftype )
 				Debug( LDAP_DEBUG_ANY, "unable to open file "
 					"\"%s\": %d (%s)\n",
 					fname,
-					save_errno, strerror( save_errno ) );
+					save_errno, AC_STRERROR_R( save_errno, ebuf, sizeof(ebuf) ) );
 
 				return -1;
 			}
@@ -77,7 +83,7 @@ test_file( const char *fname, const char *ftype )
 		Debug( LDAP_DEBUG_ANY, "unable to stat file "
 			"\"%s\": %d (%s)\n",
 			slapd_pid_file,
-			save_errno, strerror( save_errno ) );
+			save_errno, AC_STRERROR_R( save_errno, ebuf, sizeof(ebuf) ) );
 		return -1;
 	}
 
@@ -104,9 +110,12 @@ slaptest( int argc, char **argv )
 		}
 	}
 
-	fprintf( stderr, "config file testing succeeded\n");
+	if ( !quiet ) {
+		fprintf( stderr, "config file testing succeeded\n");
+	}
 
-	slap_tool_destroy();
+	if ( slap_tool_destroy())
+		rc = EXIT_FAILURE;
 
 	return rc;
 }
