@@ -1,7 +1,7 @@
-/* $OpenLDAP: pkg/ldap/tests/progs/slapd-tester.c,v 1.46.2.4 2007/02/13 20:22:35 kurt Exp $ */
+/* $OpenLDAP$ */
 /* This work is part of OpenLDAP Software <http://www.openldap.org/>.
  *
- * Copyright 1999-2007 The OpenLDAP Foundation.
+ * Copyright 1999-2020 The OpenLDAP Foundation.
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -21,15 +21,15 @@
 
 #include <stdio.h>
 
-#include <ac/stdlib.h>
+#include "ac/stdlib.h"
 
-#include <ac/ctype.h>
-#include <ac/dirent.h>
-#include <ac/param.h>
-#include <ac/socket.h>
-#include <ac/string.h>
-#include <ac/unistd.h>
-#include <ac/wait.h>
+#include "ac/ctype.h"
+#include "ac/dirent.h"
+#include "ac/param.h"
+#include "ac/socket.h"
+#include "ac/string.h"
+#include "ac/unistd.h"
+#include "ac/wait.h"
 
 
 #include "ldap_defaults.h"
@@ -40,12 +40,18 @@
 #include "lber_pvt.h"
 #include "slapd-common.h"
 
-#define SEARCHCMD		"slapd-search"
-#define READCMD			"slapd-read"
-#define ADDCMD			"slapd-addel"
-#define MODRDNCMD		"slapd-modrdn"
-#define MODIFYCMD		"slapd-modify"
-#define BINDCMD			"slapd-bind"
+#ifdef _WIN32
+#define EXE		".exe"
+#else
+#define EXE
+#endif
+
+#define SEARCHCMD		"slapd-search" EXE
+#define READCMD			"slapd-read" EXE
+#define ADDCMD			"slapd-addel" EXE
+#define MODRDNCMD		"slapd-modrdn" EXE
+#define MODIFYCMD		"slapd-modify" EXE
+#define BINDCMD			"slapd-bind" EXE
 #define MAXARGS      		100
 #define MAXREQS			5000
 #define LOOPS			100
@@ -60,7 +66,7 @@
 #define TBINDFILE		"do_bind.0"
 
 static char *get_file_name( char *dirname, char *filename );
-static int  get_search_filters( char *filename, char *filters[], char *attrs[], char *bases[] );
+static int  get_search_filters( char *filename, char *filters[], char *attrs[], char *bases[], LDAPURLDesc *luds[] );
 static int  get_read_entries( char *filename, char *entries[], char *filters[] );
 static void fork_child( char *prog, char **args );
 static void	wait4kids( int nkidval );
@@ -86,7 +92,7 @@ usage( char *name, char opt )
 
 	fprintf( stderr,
 		"usage: %s "
-		"-H <uri> | ([-h <host>] -p <port>) "
+		"-H <uri> "
 		"-D <manager> "
 		"-w <passwd> "
 		"-d <datadir> "
@@ -110,8 +116,6 @@ main( int argc, char **argv )
 {
 	int		i, j;
 	char		*uri = NULL;
-	char		*host = "localhost";
-	char		*port = NULL;
 	char		*manager = NULL;
 	char		*passwd = NULL;
 	char		*dirname = NULL;
@@ -133,12 +137,16 @@ main( int argc, char **argv )
 	char		*sreqs[MAXREQS];
 	char		*sattrs[MAXREQS];
 	char		*sbase[MAXREQS];
+	LDAPURLDesc	*slud[MAXREQS];
 	int		snum = 0;
 	char		*sargs[MAXARGS];
 	int		sanum;
+	int		sextra_args = 0;
 	char		scmd[MAXPATHLEN];
+	int		swamp = 0;
+	char		swampopt[sizeof("-SSS")];
 	/* static so that its address can be used in initializer below. */
-	static char	sloops[] = "18446744073709551615UL";
+	static char	sloops[LDAP_PVT_INTTYPE_CHARS(unsigned long)];
 	/* read */
 	char		*rfile = NULL;
 	char		*rreqs[MAXREQS];
@@ -146,15 +154,16 @@ main( int argc, char **argv )
 	char		*rargs[MAXARGS];
 	char		*rflts[MAXREQS];
 	int		ranum;
+	int		rextra_args = 0;
 	char		rcmd[MAXPATHLEN];
-	static char	rloops[] = "18446744073709551615UL";
+	static char	rloops[LDAP_PVT_INTTYPE_CHARS(unsigned long)];
 	/* addel */
 	char		*afiles[MAXREQS];
 	int		anum = 0;
 	char		*aargs[MAXARGS];
 	int		aanum;
 	char		acmd[MAXPATHLEN];
-	static char	aloops[] = "18446744073709551615UL";
+	static char	aloops[LDAP_PVT_INTTYPE_CHARS(unsigned long)];
 	/* modrdn */
 	char		*nfile = NULL;
 	char		*nreqs[MAXREQS];
@@ -162,7 +171,7 @@ main( int argc, char **argv )
 	char		*nargs[MAXARGS];
 	int		nanum;
 	char		ncmd[MAXPATHLEN];
-	static char	nloops[] = "18446744073709551615UL";
+	static char	nloops[LDAP_PVT_INTTYPE_CHARS(unsigned long)];
 	/* modify */
 	char		*mfile = NULL;
 	char		*mreqs[MAXREQS];
@@ -171,7 +180,7 @@ main( int argc, char **argv )
 	char		*margs[MAXARGS];
 	int		manum;
 	char		mcmd[MAXPATHLEN];
-	static char	mloops[] = "18446744073709551615UL";
+	static char	mloops[LDAP_PVT_INTTYPE_CHARS(unsigned long)];
 	/* bind */
 	char		*bfile = NULL;
 	char		*breqs[MAXREQS];
@@ -181,7 +190,7 @@ main( int argc, char **argv )
 	char		*bargs[MAXARGS];
 	int		banum;
 	char		bcmd[MAXPATHLEN];
-	static char	bloops[] = "18446744073709551615UL";
+	static char	bloops[LDAP_PVT_INTTYPE_CHARS(unsigned long)];
 	char		**bargs_extra = NULL;
 
 	char		*friendlyOpt = NULL;
@@ -206,7 +215,7 @@ main( int argc, char **argv )
 	mloops[0] = '\0';
 	bloops[0] = '\0';
 
-	while ( ( i = getopt( argc, argv, "AB:CD:d:FH:h:Ii:j:L:l:NP:p:r:t:Ww:y:" ) ) != EOF )
+	while ( ( i = getopt( argc, argv, "AB:CD:d:FH:h:Ii:j:L:l:NP:p:r:St:Ww:y:" ) ) != EOF )
 	{
 		switch ( i ) {
 		case 'A':
@@ -240,7 +249,7 @@ main( int argc, char **argv )
 			break;
 
 		case 'd':		/* data directory */
-			dirname = strdup( optarg );
+			dirname = optarg;
 			break;
 
 		case 'F':
@@ -248,11 +257,7 @@ main( int argc, char **argv )
 			break;
 
 		case 'H':		/* slapd uri */
-			uri = strdup( optarg );
-			break;
-
-		case 'h':		/* slapd host */
-			host = strdup( optarg );
+			uri = optarg;
 			break;
 
 		case 'I':
@@ -314,7 +319,7 @@ main( int argc, char **argv )
 			break;
 
 		case 'L':		/* the number of outerloops per client */
-			outerloops = strdup( optarg );
+			outerloops = optarg;
 			break;
 
 		case 'N':
@@ -322,19 +327,19 @@ main( int argc, char **argv )
 			break;
 
 		case 'P':		/* prog directory */
-			progdir = strdup( optarg );
-			break;
-
-		case 'p':		/* the servers port number */
-			port = strdup( optarg );
+			progdir = optarg;
 			break;
 
 		case 'r':		/* the number of retries in case of error */
-			retries = strdup( optarg );
+			retries = optarg;
+			break;
+
+		case 'S':
+			swamp++;
 			break;
 
 		case 't':		/* the delay in seconds between each retry */
-			delay = strdup( optarg );
+			delay = optarg;
 			break;
 
 		case 'w':		/* the managers passwd */
@@ -356,7 +361,7 @@ main( int argc, char **argv )
 		}
 	}
 
-	if (( dirname == NULL ) || ( port == NULL && uri == NULL ) ||
+	if (( dirname == NULL ) || ( uri == NULL ) ||
 			( manager == NULL ) || ( passwd == NULL ) || ( progdir == NULL ))
 	{
 		usage( argv[0], '\0' );
@@ -412,33 +417,67 @@ main( int argc, char **argv )
 		passwd = pw.bv_val;
 	}
 
+	if ( !sfile && !rfile && !nfile && !mfile && !bfile && !anum ) {
+		fprintf( stderr, "no data files found.\n" );
+		exit( EXIT_FAILURE );
+	}
+
 	/* look for search requests */
 	if ( sfile ) {
-		snum = get_search_filters( sfile, sreqs, sattrs, sbase );
+		snum = get_search_filters( sfile, sreqs, sattrs, sbase, slud );
+		if ( snum < 0 ) {
+			fprintf( stderr,
+				"unable to parse file \"%s\" line %d\n",
+				sfile, -2*(snum + 1));
+			exit( EXIT_FAILURE );
+		}
 	}
 
 	/* look for read requests */
 	if ( rfile ) {
 		rnum = get_read_entries( rfile, rreqs, rflts );
+		if ( rnum < 0 ) {
+			fprintf( stderr,
+				"unable to parse file \"%s\" line %d\n",
+				rfile, -2*(rnum + 1) );
+			exit( EXIT_FAILURE );
+		}
 	}
 
 	/* look for modrdn requests */
 	if ( nfile ) {
 		nnum = get_read_entries( nfile, nreqs, NULL );
+		if ( nnum < 0 ) {
+			fprintf( stderr,
+				"unable to parse file \"%s\" line %d\n",
+				nfile, -2*(nnum + 1) );
+			exit( EXIT_FAILURE );
+		}
 	}
 
 	/* look for modify requests */
 	if ( mfile ) {
-		mnum = get_search_filters( mfile, mreqs, NULL, mdn );
+		mnum = get_search_filters( mfile, mreqs, NULL, mdn, NULL );
+		if ( mnum < 0 ) {
+			fprintf( stderr,
+				"unable to parse file \"%s\" line %d\n",
+				mfile, -2*(mnum + 1) );
+			exit( EXIT_FAILURE );
+		}
 	}
 
 	/* look for bind requests */
 	if ( bfile ) {
-		bnum = get_search_filters( bfile, bcreds, battrs, breqs );
+		bnum = get_search_filters( bfile, bcreds, battrs, breqs, NULL );
+		if ( bnum < 0 ) {
+			fprintf( stderr,
+				"unable to parse file \"%s\" line %d\n",
+				bfile, -2*(bnum + 1) );
+			exit( EXIT_FAILURE );
+		}
 	}
 
 	/* setup friendly option */
-
 	switch ( friendly ) {
 	case 0:
 		break;
@@ -454,6 +493,15 @@ main( int argc, char **argv )
 		break;
 	}
 
+	/* setup swamp option */
+	if ( swamp ) {
+		swampopt[0] = '-';
+		if ( swamp > 3 ) swamp = 3;
+		swampopt[swamp + 1] = '\0';
+		for ( ; swamp-- > 0; ) swampopt[swamp + 1] = 'S';
+	}
+
+	/* setup loop options */
 	if ( sloops[0] == '\0' ) snprintf( sloops, sizeof( sloops ), "%d", 10 * loops );
 	if ( rloops[0] == '\0' ) snprintf( rloops, sizeof( rloops ), "%d", 20 * loops );
 	if ( aloops[0] == '\0' ) snprintf( aloops, sizeof( aloops ), "%d", loops );
@@ -469,15 +517,8 @@ main( int argc, char **argv )
 	snprintf( scmd, sizeof scmd, "%s" LDAP_DIRSEP SEARCHCMD,
 		progdir );
 	sargs[sanum++] = scmd;
-	if ( uri ) {
-		sargs[sanum++] = "-H";
-		sargs[sanum++] = uri;
-	} else {
-		sargs[sanum++] = "-h";
-		sargs[sanum++] = host;
-		sargs[sanum++] = "-p";
-		sargs[sanum++] = port;
-	}
+	sargs[sanum++] = "-H";
+	sargs[sanum++] = uri;
 	sargs[sanum++] = "-D";
 	sargs[sanum++] = manager;
 	sargs[sanum++] = "-w";
@@ -506,15 +547,21 @@ main( int argc, char **argv )
 		sargs[sanum++] = "-i";
 		sargs[sanum++] = ignore;
 	}
+	if ( swamp ) {
+		sargs[sanum++] = swampopt;
+	}
 	sargs[sanum++] = "-b";
 	sargs[sanum++] = NULL;		/* will hold the search base */
+	sargs[sanum++] = "-s";
+	sargs[sanum++] = NULL;		/* will hold the search scope */
 	sargs[sanum++] = "-f";
 	sargs[sanum++] = NULL;		/* will hold the search request */
 
 	sargs[sanum++] = NULL;
-	sargs[sanum] = NULL;		/* might hold the "attr" request */
+	sargs[sanum++] = NULL;		/* might hold the "attr" request */
+	sextra_args += 2;
 
-	sargs[sanum + 1] = NULL;
+	sargs[sanum] = NULL;
 
 	/*
 	 * generate the read clients
@@ -524,15 +571,8 @@ main( int argc, char **argv )
 	snprintf( rcmd, sizeof rcmd, "%s" LDAP_DIRSEP READCMD,
 		progdir );
 	rargs[ranum++] = rcmd;
-	if ( uri ) {
-		rargs[ranum++] = "-H";
-		rargs[ranum++] = uri;
-	} else {
-		rargs[ranum++] = "-h";
-		rargs[ranum++] = host;
-		rargs[ranum++] = "-p";
-		rargs[ranum++] = port;
-	}
+	rargs[ranum++] = "-H";
+	rargs[ranum++] = uri;
 	rargs[ranum++] = "-D";
 	rargs[ranum++] = manager;
 	rargs[ranum++] = "-w";
@@ -558,13 +598,17 @@ main( int argc, char **argv )
 		rargs[ranum++] = "-i";
 		rargs[ranum++] = ignore;
 	}
+	if ( swamp ) {
+		rargs[ranum++] = swampopt;
+	}
 	rargs[ranum++] = "-e";
 	rargs[ranum++] = NULL;		/* will hold the read entry */
 
 	rargs[ranum++] = NULL;
-	rargs[ranum] = NULL;		/* might hold the filter arg */
+	rargs[ranum++] = NULL;		/* might hold the filter arg */
+	rextra_args += 2;
 
-	rargs[ranum + 1] = NULL;
+	rargs[ranum] = NULL;
 
 	/*
 	 * generate the modrdn clients
@@ -574,15 +618,8 @@ main( int argc, char **argv )
 	snprintf( ncmd, sizeof ncmd, "%s" LDAP_DIRSEP MODRDNCMD,
 		progdir );
 	nargs[nanum++] = ncmd;
-	if ( uri ) {
-		nargs[nanum++] = "-H";
-		nargs[nanum++] = uri;
-	} else {
-		nargs[nanum++] = "-h";
-		nargs[nanum++] = host;
-		nargs[nanum++] = "-p";
-		nargs[nanum++] = port;
-	}
+	nargs[nanum++] = "-H";
+	nargs[nanum++] = uri;
 	nargs[nanum++] = "-D";
 	nargs[nanum++] = manager;
 	nargs[nanum++] = "-w";
@@ -607,7 +644,7 @@ main( int argc, char **argv )
 	}
 	nargs[nanum++] = "-e";
 	nargs[nanum++] = NULL;		/* will hold the modrdn entry */
-	nargs[nanum++] = NULL;
+	nargs[nanum] = NULL;
 	
 	/*
 	 * generate the modify clients
@@ -617,15 +654,8 @@ main( int argc, char **argv )
 	snprintf( mcmd, sizeof mcmd, "%s" LDAP_DIRSEP MODIFYCMD,
 		progdir );
 	margs[manum++] = mcmd;
-	if ( uri ) {
-		margs[manum++] = "-H";
-		margs[manum++] = uri;
-	} else {
-		margs[manum++] = "-h";
-		margs[manum++] = host;
-		margs[manum++] = "-p";
-		margs[manum++] = port;
-	}
+	margs[manum++] = "-H";
+	margs[manum++] = uri;
 	margs[manum++] = "-D";
 	margs[manum++] = manager;
 	margs[manum++] = "-w";
@@ -652,7 +682,7 @@ main( int argc, char **argv )
 	margs[manum++] = NULL;		/* will hold the modify entry */
 	margs[manum++] = "-a";;
 	margs[manum++] = NULL;		/* will hold the ava */
-	margs[manum++] = NULL;
+	margs[manum] = NULL;
 
 	/*
 	 * generate the add/delete clients
@@ -662,15 +692,8 @@ main( int argc, char **argv )
 	snprintf( acmd, sizeof acmd, "%s" LDAP_DIRSEP ADDCMD,
 		progdir );
 	aargs[aanum++] = acmd;
-	if ( uri ) {
-		aargs[aanum++] = "-H";
-		aargs[aanum++] = uri;
-	} else {
-		aargs[aanum++] = "-h";
-		aargs[aanum++] = host;
-		aargs[aanum++] = "-p";
-		aargs[aanum++] = port;
-	}
+	aargs[aanum++] = "-H";
+	aargs[aanum++] = uri;
 	aargs[aanum++] = "-D";
 	aargs[aanum++] = manager;
 	aargs[aanum++] = "-w";
@@ -695,7 +718,7 @@ main( int argc, char **argv )
 	}
 	aargs[aanum++] = "-f";
 	aargs[aanum++] = NULL;		/* will hold the add data file */
-	aargs[aanum++] = NULL;
+	aargs[aanum] = NULL;
 
 	/*
 	 * generate the bind clients
@@ -708,25 +731,16 @@ main( int argc, char **argv )
 	if ( !noinit ) {
 		bargs[banum++] = "-I";	/* init on each bind */
 	}
-	if ( uri ) {
-		bargs[banum++] = "-H";
-		bargs[banum++] = uri;
-	} else {
-		bargs[banum++] = "-h";
-		bargs[banum++] = host;
-		bargs[banum++] = "-p";
-		bargs[banum++] = port;
-	}
+	bargs[banum++] = "-H";
+	bargs[banum++] = uri;
 	bargs[banum++] = "-l";
 	bargs[banum++] = bloops;
 	bargs[banum++] = "-L";
 	bargs[banum++] = outerloops;
-#if 0
 	bargs[banum++] = "-r";
 	bargs[banum++] = retries;
 	bargs[banum++] = "-t";
 	bargs[banum++] = delay;
-#endif
 	if ( friendly ) {
 		bargs[banum++] = friendlyOpt;
 	}
@@ -745,56 +759,99 @@ main( int argc, char **argv )
 	bargs[banum++] = NULL;
 	bargs[banum++] = "-w";
 	bargs[banum++] = NULL;
-	bargs[banum++] = NULL;
+	bargs[banum] = NULL;
 
 #define	DOREQ(n,j) ((n) && ((maxkids > (n)) ? ((j) < maxkids ) : ((j) < (n))))
 
 	for ( j = 0; j < MAXREQS; j++ ) {
+		/* search */
 		if ( DOREQ( snum, j ) ) {
 			int	jj = j % snum;
+			int	x = sanum - sextra_args;
 
-			sargs[sanum - 2] = sreqs[jj];
-			sargs[sanum - 4] = sbase[jj];
-			if ( sattrs[jj] != NULL ) {
-				sargs[sanum - 1] = "-a";
-				sargs[sanum] = sattrs[jj];
+			/* base */
+			if ( sbase[jj] != NULL ) {
+				sargs[sanum - 7] = sbase[jj];
 
 			} else {
-				sargs[sanum - 1] = NULL;
+				sargs[sanum - 7] = slud[jj]->lud_dn;
 			}
+
+			/* scope */
+			if ( slud[jj] != NULL ) {
+				sargs[sanum - 5] = (char *)ldap_pvt_scope2str( slud[jj]->lud_scope );
+
+			} else {
+				sargs[sanum - 5] = "sub";
+			}
+
+			/* filter */
+			if ( sreqs[jj] != NULL ) {
+				sargs[sanum - 3] = sreqs[jj];
+
+			} else if ( slud[jj]->lud_filter != NULL ) {
+				sargs[sanum - 3] = slud[jj]->lud_filter;
+
+			} else {
+				sargs[sanum - 3] = "(objectClass=*)";
+			}
+
+			/* extras */
+			sargs[x] = NULL;
+
+			/* attr */
+			if ( sattrs[jj] != NULL ) {
+				sargs[x++] = "-a";
+				sargs[x++] = sattrs[jj];
+			}
+
+			/* attrs */
+			if ( slud[jj] != NULL && slud[jj]->lud_attrs != NULL ) {
+				int	i;
+
+				for ( i = 0; slud[jj]->lud_attrs[ i ] != NULL && x + i < MAXARGS - 1; i++ ) {
+					sargs[x + i] = slud[jj]->lud_attrs[ i ];
+				}
+				sargs[x + i] = NULL;
+			}
+
 			fork_child( scmd, sargs );
 		}
 
+		/* read */
 		if ( DOREQ( rnum, j ) ) {
 			int	jj = j % rnum;
+			int	x = ranum - rextra_args;
 
-			rargs[ranum - 2] = rreqs[jj];
+			rargs[ranum - 3] = rreqs[jj];
 			if ( rflts[jj] != NULL ) {
-				rargs[ranum - 1] = "-f";
-				rargs[ranum] = rflts[jj];
-
-			} else {
-				rargs[ranum - 1] = NULL;
+				rargs[x++] = "-f";
+				rargs[x++] = rflts[jj];
 			}
+			rargs[x] = NULL;
 			fork_child( rcmd, rargs );
 		}
 
+		/* rename */
 		if ( j < nnum ) {
-			nargs[nanum - 2] = nreqs[j];
+			nargs[nanum - 1] = nreqs[j];
 			fork_child( ncmd, nargs );
 		}
 
+		/* modify */
 		if ( j < mnum ) {
-			margs[manum - 4] = mdn[j];
-			margs[manum - 2] = mreqs[j];
+			margs[manum - 3] = mdn[j];
+			margs[manum - 1] = mreqs[j];
 			fork_child( mcmd, margs );
 		}
 
+		/* add/delete */
 		if ( j < anum ) {
-			aargs[aanum - 2] = afiles[j];
+			aargs[aanum - 1] = afiles[j];
 			fork_child( acmd, aargs );
 		}
 
+		/* bind */
 		if ( DOREQ( bnum, j ) ) {
 			int	jj = j % bnum;
 
@@ -808,23 +865,25 @@ main( int argc, char **argv )
 			}
 
 			if ( battrs[jj] != NULL ) {
-				bargs[banum - 4] = manager ? manager : "";
-				bargs[banum - 2] = passwd ? passwd : "";
+				bargs[banum - 3] = manager ? manager : "";
+				bargs[banum - 1] = passwd ? passwd : "";
+
+				bargs[banum + 0] = "-b";
+				bargs[banum + 1] = breqs[jj];
+				bargs[banum + 2] = "-f";
+				bargs[banum + 3] = bcreds[jj];
+				bargs[banum + 4] = "-a";
+				bargs[banum + 5] = battrs[jj];
+				bargs[banum + 6] = NULL;
 
-				bargs[banum - 1] = "-b";
-				bargs[banum] = breqs[jj];
-				bargs[banum + 1] = "-f";
-				bargs[banum + 2] = bcreds[jj];
-				bargs[banum + 3] = "-a";
-				bargs[banum + 4] = battrs[jj];
 			} else {
-				bargs[banum - 4] = breqs[jj];
-				bargs[banum - 2] = bcreds[jj];
-				bargs[banum - 1] = NULL;
+				bargs[banum - 3] = breqs[jj];
+				bargs[banum - 1] = bcreds[jj];
+				bargs[banum] = NULL;
 			}
 
 			fork_child( bcmd, bargs );
-			bargs[banum - 1] = NULL;
+			bargs[banum] = NULL;
 		}
 	}
 
@@ -845,7 +904,7 @@ get_file_name( char *dirname, char *filename )
 
 
 static int
-get_search_filters( char *filename, char *filters[], char *attrs[], char *bases[] )
+get_search_filters( char *filename, char *filters[], char *attrs[], char *bases[], LDAPURLDesc *luds[] )
 {
 	FILE    *fp;
 	int     filter = 0;
@@ -854,12 +913,37 @@ get_search_filters( char *filename, char *filters[], char *attrs[], char *bases[
 		char  line[BUFSIZ];
 
 		while (( filter < MAXREQS ) && ( fgets( line, BUFSIZ, fp ))) {
-			char *nl;
+			char	*nl;
+			int	got_URL = 0;
 
 			if (( nl = strchr( line, '\r' )) || ( nl = strchr( line, '\n' )))
 				*nl = '\0';
-			bases[filter] = ArgDup( line );
-			fgets( line, BUFSIZ, fp );
+
+			if ( luds ) luds[filter] = NULL;
+
+			if ( luds && strncmp( line, "ldap:///", STRLENOF( "ldap:///" ) ) == 0 ) {
+				LDAPURLDesc	*lud;
+
+				got_URL = 1;
+				bases[filter] = NULL;
+				if ( ldap_url_parse( line, &lud ) != LDAP_URL_SUCCESS ) {
+					filter = -filter - 1;
+					break;
+				}
+
+				if ( lud->lud_dn == NULL || lud->lud_exts != NULL ) {
+					filter = -filter - 1;
+					ldap_free_urldesc( lud );
+					break;
+				}
+
+				luds[filter] = lud;
+
+			} else {
+				bases[filter] = ArgDup( line );
+			}
+			if ( fgets( line, BUFSIZ, fp ) == NULL )
+				*line = '\0';
 			if (( nl = strchr( line, '\r' )) || ( nl = strchr( line, '\n' )))
 				*nl = '\0';
 
@@ -868,15 +952,15 @@ get_search_filters( char *filename, char *filters[], char *attrs[], char *bases[
 				if ( filters[filter][0] == '+') {
 					char	*sep = strchr( filters[filter], ':' );
 
+					attrs[ filter ] = &filters[ filter ][ 1 ];
 					if ( sep != NULL ) {
-						attrs[ filter ] = &filters[ filter ][ 1 ];
 						sep[ 0 ] = '\0';
 						/* NOTE: don't free this! */
 						filters[ filter ] = &sep[ 1 ];
 					}
 
 				} else {
-					attrs[ filter] = NULL;
+					attrs[ filter ] = NULL;
 				}
 			}
 			filter++;
@@ -885,7 +969,7 @@ get_search_filters( char *filename, char *filters[], char *attrs[], char *bases[
 		fclose( fp );
 	}
 
-	return( filter );
+	return filter;
 }
 
 
@@ -907,13 +991,13 @@ get_read_entries( char *filename, char *entries[], char *filters[] )
 				LDAPURLDesc	*lud;
 
 				if ( ldap_url_parse( &line[1], &lud ) != LDAP_URL_SUCCESS ) {
-					entry = -1;
+					entry = -entry - 1;
 					break;
 				}
 
 				if ( lud->lud_dn == NULL || lud->lud_dn[ 0 ] == '\0' ) {
 					ldap_free_urldesc( lud );
-					entry = -1;
+					entry = -entry - 1;
 					break;
 				}
 
@@ -928,6 +1012,9 @@ get_read_entries( char *filename, char *entries[], char *filters[] )
 				ldap_free_urldesc( lud );
 
 			} else {
+				if ( filters != NULL )
+					filters[entry] = NULL;
+
 				entries[entry] = ArgDup( line );
 			}
 
@@ -967,6 +1054,13 @@ fork_child( char *prog, char **args )
 #endif
 		execvp( prog, args );
 		tester_perror( "execvp", NULL );
+		{ int i;
+			for (i=0; args[i]; i++);
+			fprintf(stderr,"%d args\n", i);
+			for (i=0; args[i]; i++)
+				fprintf(stderr,"%d %s\n", i, args[i]);
+		}
+
 		exit( EXIT_FAILURE );
 		break;
 
@@ -986,17 +1080,17 @@ wait4kids( int nkidval )
 	int		status;
 
 	while ( nkids >= nkidval ) {
-		wait( &status );
+		pid_t pid = wait( &status );
 
 		if ( WIFSTOPPED(status) ) {
 			fprintf( stderr,
-			    "stopping: child stopped with signal %d\n",
-			    (int) WSTOPSIG(status) );
+			    "stopping: child PID=%ld stopped with signal %d\n",
+			    (long) pid, (int) WSTOPSIG(status) );
 
 		} else if ( WIFSIGNALED(status) ) {
 			fprintf( stderr, 
-			    "stopping: child terminated with signal %d%s\n",
-			    (int) WTERMSIG(status),
+			    "stopping: child PID=%ld terminated with signal %d%s\n",
+			    (long) pid, (int) WTERMSIG(status),
 #ifdef WCOREDUMP
 				WCOREDUMP(status) ? ", core dumped" : ""
 #else
@@ -1007,8 +1101,8 @@ wait4kids( int nkidval )
 
 		} else if ( WEXITSTATUS(status) != 0 ) {
 			fprintf( stderr, 
-			    "stopping: child exited with status %d\n",
-			    (int) WEXITSTATUS(status) );
+			    "stopping: child PID=%ld exited with status %d\n",
+			    (long) pid, (int) WEXITSTATUS(status) );
 			exit( WEXITSTATUS(status) );
 
 		} else {
