@@ -1,8 +1,8 @@
 #! /bin/sh
-# $OpenLDAP: pkg/ldap/tests/scripts/conf.sh,v 1.49.2.3 2007/02/13 20:22:35 kurt Exp $
+# $OpenLDAP$
 ## This work is part of OpenLDAP Software <http://www.openldap.org/>.
 ##
-## Copyright 1998-2007 The OpenLDAP Foundation.
+## Copyright 1998-2020 The OpenLDAP Foundation.
 ## All rights reserved.
 ##
 ## Redistribution and use in source and binary forms, with or without
@@ -12,15 +12,6 @@
 ## A copy of this license is available in the file LICENSE in the
 ## top-level directory of the distribution or, alternatively, at
 ## <http://www.OpenLDAP.org/license.html>.
-MONMOD=nomod
-if [ x"$MONITORDB" = xyes -o x"$MONITORDB" = xmod ] ; then
-	MON=monitor
-	if [ $MONITORDB = mod ] ; then
-		MONMOD=monitormod
-	fi
-else
-	MON=nomonitor
-fi
 if [ x"$WITH_SASL" = x"yes" -a x"$USE_SASL" != x"no" ] ; then
 	SASL="sasl"
 	if [ x"$USE_SASL" = x"yes" ] ; then
@@ -33,17 +24,22 @@ else
 fi
 sed -e "s/@BACKEND@/${BACKEND}/"			\
 	-e "s/^#${BACKEND}#//"				\
+	-e "/^#~/s/^#[^#]*~${BACKEND}~[^#]*#/#omit: /"	\
+		-e "s/^#~[^#]*~#//"			\
 	-e "s/@RELAY@/${RELAY}/"			\
 	-e "s/^#relay-${RELAY}#//"			\
 	-e "s/^#${BACKENDTYPE}#//"			\
 	-e "s/^#${AC_ldap}#//"				\
 	-e "s/^#${AC_meta}#//"				\
+	-e "s/^#${AC_asyncmeta}#//"			\
 	-e "s/^#${AC_relay}#//"				\
 	-e "s/^#${AC_sql}#//"				\
 		-e "s/^#${RDBMS}#//"			\
 	-e "s/^#${AC_accesslog}#//"			\
+	-e "s/^#${AC_autoca}#//"			\
 	-e "s/^#${AC_dds}#//"				\
 	-e "s/^#${AC_dynlist}#//"			\
+	-e "s/^#${AC_memberof}#//"			\
 	-e "s/^#${AC_pcache}#//"			\
 	-e "s/^#${AC_ppolicy}#//"			\
 	-e "s/^#${AC_refint}#//"			\
@@ -53,8 +49,8 @@ sed -e "s/@BACKEND@/${BACKEND}/"			\
 	-e "s/^#${AC_translucent}#//"			\
 	-e "s/^#${AC_unique}#//"			\
 	-e "s/^#${AC_valsort}#//"			\
-	-e "s/^#${MON}#//"				\
-	-e "s/^#${MONMOD}#//"				\
+	-e "s/^#${INDEXDB}#//"				\
+	-e "s/^#${MAINDB}#//"				\
 	-e "s/^#${SASL}#//"				\
 	-e "s/^#${ACI}#//"				\
 	-e "s;@URI1@;${URI1};"				\
@@ -69,9 +65,30 @@ sed -e "s/@BACKEND@/${BACKEND}/"			\
 	-e "s;@PORT4@;${PORT4};"			\
 	-e "s;@PORT5@;${PORT5};"			\
 	-e "s;@PORT6@;${PORT6};"			\
+	-e "s;@SURI1@;${SURI1};"			\
+	-e "s;@SURI2@;${SURI2};"			\
+	-e "s;@SURI3@;${SURI3};"			\
+	-e "s;@SURI4@;${SURI4};"			\
+	-e "s;@SURI5@;${SURI5};"			\
+	-e "s;@SURI6@;${SURI6};"			\
+	-e "s;@URIP1@;${URIP1};"			\
+	-e "s;@URIP2@;${URIP2};"			\
+	-e "s;@URIP3@;${URIP3};"			\
+	-e "s;@URIP4@;${URIP4};"			\
+	-e "s;@URIP5@;${URIP5};"			\
+	-e "s;@URIP6@;${URIP6};"			\
+	-e "s;@SURIP1@;${SURIP1};"			\
+	-e "s;@SURIP2@;${SURIP2};"			\
+	-e "s;@SURIP3@;${SURIP3};"			\
+	-e "s;@SURIP4@;${SURIP4};"			\
+	-e "s;@SURIP5@;${SURIP5};"			\
+	-e "s;@SURIP6@;${SURIP6};"			\
 	-e "s/@SASL_MECH@/${SASL_MECH}/"		\
-	-e "s/@CACHETTL@/${CACHETTL}/"			\
-	-e "s/@ENTRY_LIMIT@/${CACHE_ENTRY_LIMIT}/"	\
 	-e "s;@TESTDIR@;${TESTDIR};"			\
+	-e "s;@TESTWD@;${TESTWD};"			\
 	-e "s;@DATADIR@;${DATADIR};"			\
-	-e "s;@SCHEMADIR@;${SCHEMADIR};"
+	-e "s;@SCHEMADIR@;${SCHEMADIR};"		\
+	-e "s;@KRB5REALM@;${KRB5REALM};"		\
+	-e "s;@KDCHOST@;${KDCHOST};"			\
+	-e "s;@KDCPORT@;${KDCPORT};"			\
+	-e "/^#/d"
